/***************************************************************************
                          renderer.cpp  -  description
                             -------------------
    begin                : Sun Jul 22 2001
    copyright            : (C) 2001 by T Kontkanen
    email                : tkontkanen@mail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "renderer.h"
#include "font.h"
#include "game.h"
#include <algorithm>

using namespace TGL;
using namespace msg;

Renderer::Renderer( Game *parent, int w, int h, int depth, bool fullscr ) {
	setFuncName( "Renderer::Renderer" );

	int i;
	
	game=parent;
	screen_w=w;
	screen_h=h;
	screen_depth=depth;
	fullscreen=fullscr;
	cameraAngle=cameraPitch=0;
	
	// init SDL
	SDL_GL_SetAttribute(SDL_GL_BUFFER_SIZE,getScreenDepth());
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,1);
	setVideoMode( w, h, depth, fullscr );
	SDL_WM_SetCaption( "Executor", "executor" );
	
	// enable glError handling
	msg::glReady(true);
	
	//gl stuff
	glClearColor(0.5f,0.75f,1.0f,0.0f);
	glDisable(GL_NORMALIZE);
	glEnable(GL_DEPTH_TEST);
	glClearDepth( 1.0 );
	glDepthFunc( GL_LEQUAL );
	glShadeModel(GL_SMOOTH);
	maxViewDistance=4500;
	
	// multitexture
	glActiveTextureARB_ptr = (GL_ActiveTextureARB_Func) SDL_GL_GetProcAddress("glActiveTextureARB");
	glMultiTexCoord2fARB_ptr = (GL_MultiTexCoord2fARB_Func) SDL_GL_GetProcAddress("glMultiTexCoord2fARB");
	if( glActiveTextureARB_ptr!=NULL && glMultiTexCoord2fARB_ptr!=NULL ) multitexturing=true;
	else multitexturing=false;
	msgOut( MSG_INFO, "Multitexturing %s\n", multitexturing?"enabled":"disabled" );
	wireframe=false;

	// fog
	setFog( 1, 0, maxViewDistance, 0.0005 );
	glFogfv(GL_FOG_COLOR,fogColor.RGBA());

	//TGL init
	if( getFullScreen() == true )
		tgl_setmode( getScreenW(), getScreenH(), TGL_FULLSCREEN, getScreenDepth() );
	else
		tgl_setmode( getScreenW(), getScreenH(), TGL_WINDOWED, getScreenDepth() );
		
	tgl_init( MAX_TGL_TEXTURES ); //max number of textures
	tgl_object_rendermode=TGL_ELEMENTS;
	
	// init tgl objects
	for(i=0; i<MAX_TGL_OBJECTS; i++ ) {
		objects[i]=NULL;
		object_clients[i]=0;
		bzero( object_file_names[i], MAX_FILE_NAME_LEN ); //clear
		strcpy( object_file_names[i], "" ); //may be obsolete
	}
		
	//init textures
	for(i=0; i<MAX_TEXTURES; i++ ) {
		textures[i]=NULL;
	}
	
	//misc
	currentfont=NULL;
	setParticleMode( FUNC_ADD, 10 );
	setParticleAmount( 2 ); // 50%
	render_lines=true;
	particle_jobs.clear();
	particle_effects.clear();
	
	//sun
	sunPos.set(100000,100000,100000);
	sunDiffuse.set(1,1,1,1);
	sunAmbient.set(.1,.1,.1,1);
	setSun( sunPos, sunDiffuse, sunAmbient );

	readConfig();
	checkError("Renderer::Renderer" );
	exitFunc();
}


Renderer::~Renderer(){
	setFuncName( "Renderer::~Renderer" );
	msg::glReady(false);
	SDL_FreeSurface( screen );
	exitFunc();
}

void Renderer::readConfig()
{
	game->config.setGroup( "renderer" );
	setParticleAmount( game->config.getInt( "particleReduction", 2 ) ); // 50%
	tgl_debug=game->config.getInt( "TGLDebugLevel",1 );
	tgl_disable_lighting=game->config.getBool( "disableLighting", false );
	render_lines=game->config.getBool( "renderLines", true );
	
	particle_sorting=game->config.getBool( "particleSorting", true );
	particle_depthbuffer=game->config.getBool( "particleDepthBuffering",false );
	particle_post_rendering=game->config.getBool( "particlePostRendering", true );
	mouseCursor=loadTexture( game->dataRoot()+IMAGEPATH+game->config.getString( "mouseCursor", "mousecurs.png" ) );
}


//camera methods:
void Renderer::initCamera()
{
	setFuncName( "Renderer::initCamera" );
	glViewport( 0, 0, getScreenW(), getScreenH() );
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective( 60.0f,(float)getScreenW()/(float)getScreenH(),0.65,maxViewDistance );
//	gluPerspective( 45.0f,(float)getScreenW()/(float)getScreenH(),1,maxViewDistance );
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	camera_changed=true;
	camera_moved=true;
	cameraMode=CAMERA_FREE;
		
	lookUp.set( 0, 0, 1 );
	cameraPosition.set( 0, 0, 0 );
	lookAt.set( 0, 1, 0 );
	lookAt+=cameraPosition;
	setCameraAngle( 0, 0 );
	cameraDistance=1.0;
	
	smoothLookAt=lookAt;
	smoothLookUp=lookUp;
	smoothCameraPosition=cameraPosition;
}

void Renderer::setVideoMode( int w, int h, int depth, bool fullscr )
{
	setFuncName( "Renderer::setVideoMode" );
	msgOut( MSG_DEBUG_3, "Set new video mode" );
	
	Uint32 screenFlags = SDL_OPENGL|SDL_GL_DOUBLEBUFFER;

	fullscreen=fullscr;
	screen_w=w;
	screen_h=h;
	
	screenFlags |=SDL_RESIZABLE;
	if( fullscreen )
		screenFlags |= SDL_FULLSCREEN;
	
//	screen = SDL_SetVideoMode( 300,300, 24, SDL_OPENGL|SDL_GL_DOUBLEBUFFER );

	// set new mode
	screen = SDL_SetVideoMode( w, h, depth, screenFlags );
	if ( !screen )
		msgOut( MSG_FATAL,  "Unable to set video mode: %s", SDL_GetError() );

	glViewport( 0, 0, getScreenW(), getScreenH() );
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective( 60.0f,(float)getScreenW()/(float)getScreenH(),0.65,maxViewDistance );
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	// apply camera
	camera_changed=true;
	camera_moved=true;
//	applyCamera();
}


void Renderer::setCameraPosition( float x, float y, float z,  bool updateLookat )
{
	Vector lastpos;
	lastpos = cameraPosition;
	cameraPosition.set( x, y, z );
	if( updateLookat )
		lookAt += (cameraPosition-lastpos);
	camera_changed=true;
}

void Renderer::setCameraPosition( Vector pos, bool updateLookat )
{
	setCameraPosition( pos.x(),pos.y(), pos.z(), updateLookat );
}

void Renderer::setCameraLookAt( Vector pos, bool updatePosition )
{
	Vector lastpos;
	lastpos = lookAt;
	lookAt=pos;
	if( updatePosition )
		cameraPosition += (lookAt-lastpos);
	camera_changed=true;
}


void Renderer::setCameraAltitude( float z )
{
	Vector lastpos;
	lastpos = cameraPosition;
	cameraPosition.setZ( z );
	lookAt += (cameraPosition-lastpos);
	camera_changed=true;
}
	
void Renderer::moveCameraForward( float forward )
{
	Vector direction=lookAt-cameraPosition;
	direction.normalize();
	cameraPosition+=direction*forward;
	lookAt+=direction*forward;
	
	camera_changed=true;
}

void Renderer::moveCameraXYZ( float x, float y, float z, bool updateLookat )
{
	Vector addpos;
	
	addpos.set(x,y,z);
	cameraPosition+=addpos;
	if( updateLookat )
		lookAt+=addpos;
	
	camera_changed=true;
}

void Renderer::setCameraAngle( float angle, float pitch )
{
	//clamp pitch
	if( pitch > M_PI/2 ) pitch=M_PI/2.0;
	if( pitch < -M_PI/2 ) pitch=-M_PI/2.0;

	if( cameraMode==CAMERA_FREE )
	{
		cameraAngle=angle;
		cameraPitch=pitch;
		lookAt.set( 0, 1, 0 );
		lookAt.rotateX( pitch );
		lookAt.rotateZ( angle );
		lookAt += cameraPosition;
	
		//calculate up vector
		lookUp.set( 0, 0, 1 );
		lookUp.rotateX( pitch );
		lookUp.rotateZ( angle );
	} else
	{
		cameraAngle=angle;
		cameraPitch=pitch;
		
		cameraPosition.set( 0, -cameraDistance, 0 );
		cameraPosition.rotateX( pitch );
		cameraPosition.rotateZ( angle );
		cameraPosition += lookAt;
	
		//calculate up vector
		lookUp.set( 0, 0, 1 );
		lookUp.rotateX( pitch );
		lookUp.rotateZ( angle );
	}
	
	camera_changed=true;
}

void Renderer::rotateCameraAngle( float angle, float pitch )
{
	cameraAngle+=angle;
	cameraPitch+=pitch;
	
	setCameraAngle( cameraAngle, cameraPitch );
}

void Renderer::getCameraAngle( float &angle, float &pitch )
{
	angle=cameraAngle;
	pitch=cameraPitch;
}


void Renderer::moveCamera2DRight( float x )
{
	Vector right2d;
	right2d.set( 1, 0, 0 );
	right2d.rotateZ( cameraAngle );
	cameraPosition += right2d*x;
	lookAt += right2d*x;
	camera_changed=true;
}

void Renderer::moveCamera2DUp( float y )
{
	Vector up2d;
	up2d.set( 0, 1, 0 );
	up2d.rotateZ( cameraAngle );
	cameraPosition += up2d*y;
	lookAt += up2d*y;
	camera_changed=true;
}

void Renderer::setCameraMode( enum CameraModes mode )
{
	cameraMode=mode;
	rotateCameraAngle( 0,0 ); //update mode
}

void Renderer::setCameraLookAtDistance( float dist )
{
	if( dist < 0.001 ) dist=0.001;
	cameraDistance=dist;
	rotateCameraAngle(0,0);// update
}

void Renderer::adjustCameraLookAtDistance( float adjdist )
{
	cameraDistance+=adjdist;
	if( cameraDistance < 0.001 ) cameraDistance=0.001;
	rotateCameraAngle(0,0);// update
}

void Renderer::set3DModelView()
{
	setFuncName( "Renderer::set3DModelView" );
	
//	smoothLookAt=smoothLookAt*0.8+lookAt*0.2;
//	smoothLookUp=smoothLookUp*0.8+lookUp*0.2;
//	smoothCameraPosition=smoothCameraPosition*0.8+cameraPosition*0.2;
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective( 60.0f,(float)getScreenW()/(float)getScreenH(),0.65,maxViewDistance );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	
/*	gluLookAt( cameraPosition.x(), cameraPosition.y(), cameraPosition.z(),
		lookAt.i(), lookAt.j(), lookAt.k(),
		lookUp.i(), lookUp.j(), lookUp.k() );*/

	gluLookAt( smoothCameraPosition.x(), smoothCameraPosition.y(), smoothCameraPosition.z(),
		smoothLookAt.i(), smoothLookAt.j(), smoothLookAt.k(),
		smoothLookUp.i(), smoothLookUp.j(), smoothLookUp.k() );
}

void Renderer::applyCamera( float prevPosWeight )
{
	float m;
	
	setFuncName( "Renderer::applyCamera" );
	
	if( prevPosWeight<0.0 ) prevPosWeight=0.0;
	if( prevPosWeight>1.0 ) prevPosWeight=1.0;
	m=1-prevPosWeight;
	
	smoothLookAt=smoothLookAt*prevPosWeight+lookAt*m;
	smoothLookUp=smoothLookUp*prevPosWeight+lookUp*m;
	smoothCameraPosition=smoothCameraPosition*prevPosWeight+cameraPosition*m;
	
	if( lastLookAt==smoothLookAt && lastLookUp==smoothLookUp && lastCameraPosition==smoothCameraPosition )
	{
		camera_changed=false;
		camera_moved=false;
	}
	else
	{
		camera_changed=true;
		camera_moved=true;
	}
	
	lastLookAt=smoothLookAt;
	lastLookUp=smoothLookUp;
	lastCameraPosition=smoothCameraPosition;
	
	if( camera_changed );
	{
		setRenderingMode( RENDER_3D );
		set3DModelView();
		extractFrustum();
	}
	//if changed...
/*	if( camera_changed ) camera_moved=true;
	else camera_moved=false;
	camera_changed=false;*/
}


void Renderer::extractFrustum()
{
	// this piece of code is from http://www.markmorley.com/opengl/frustumculling.html
	setFuncName( "Renderer::extractFrustum" );
	float proj[16];
	float modl[16];
	float clip[16];
	float t;

	 /* Get the current PROJECTION matrix from OpenGL */
	glGetFloatv( GL_PROJECTION_MATRIX, proj );
//	glGetFloatv( GL_MODELVIEW_MATRIX, proj );

	 /* Get the current MODELVIEW matrix from OpenGL */
	glGetFloatv( GL_MODELVIEW_MATRIX, modl );
//	glGetFloatv( GL_PROJECTION_MATRIX, modl );

	/* Combine the two matrices (multiply projection by modelview) */
	clip[ 0] = modl[ 0] * proj[ 0] + modl[ 1] * proj[ 4] + modl[ 2] * proj[ 8] + modl[ 3] * proj[12];
	clip[ 1] = modl[ 0] * proj[ 1] + modl[ 1] * proj[ 5] + modl[ 2] * proj[ 9] + modl[ 3] * proj[13];
	clip[ 2] = modl[ 0] * proj[ 2] + modl[ 1] * proj[ 6] + modl[ 2] * proj[10] + modl[ 3] * proj[14];
	clip[ 3] = modl[ 0] * proj[ 3] + modl[ 1] * proj[ 7] + modl[ 2] * proj[11] + modl[ 3] * proj[15];

	clip[ 4] = modl[ 4] * proj[ 0] + modl[ 5] * proj[ 4] + modl[ 6] * proj[ 8] + modl[ 7] * proj[12];
	clip[ 5] = modl[ 4] * proj[ 1] + modl[ 5] * proj[ 5] + modl[ 6] * proj[ 9] + modl[ 7] * proj[13];
	clip[ 6] = modl[ 4] * proj[ 2] + modl[ 5] * proj[ 6] + modl[ 6] * proj[10] + modl[ 7] * proj[14];
	clip[ 7] = modl[ 4] * proj[ 3] + modl[ 5] * proj[ 7] + modl[ 6] * proj[11] + modl[ 7] * proj[15];

	clip[ 8] = modl[ 8] * proj[ 0] + modl[ 9] * proj[ 4] + modl[10] * proj[ 8] + modl[11] * proj[12];
	clip[ 9] = modl[ 8] * proj[ 1] + modl[ 9] * proj[ 5] + modl[10] * proj[ 9] + modl[11] * proj[13];
	clip[10] = modl[ 8] * proj[ 2] + modl[ 9] * proj[ 6] + modl[10] * proj[10] + modl[11] * proj[14];
	clip[11] = modl[ 8] * proj[ 3] + modl[ 9] * proj[ 7] + modl[10] * proj[11] + modl[11] * proj[15];
	
	clip[12] = modl[12] * proj[ 0] + modl[13] * proj[ 4] + modl[14] * proj[ 8] + modl[15] * proj[12];
	clip[13] = modl[12] * proj[ 1] + modl[13] * proj[ 5] + modl[14] * proj[ 9] + modl[15] * proj[13];
	clip[14] = modl[12] * proj[ 2] + modl[13] * proj[ 6] + modl[14] * proj[10] + modl[15] * proj[14];
	clip[15] = modl[12] * proj[ 3] + modl[13] * proj[ 7] + modl[14] * proj[11] + modl[15] * proj[15];

	/* Extract the numbers for the RIGHT plane */
	frustum[0][0] = clip[ 3] - clip[ 0];
	frustum[0][1] = clip[ 7] - clip[ 4];
	frustum[0][2] = clip[11] - clip[ 8];
	frustum[0][3] = clip[15] - clip[12];

	/* Normalize the result */
	t = sqrt( frustum[0][0] * frustum[0][0] + frustum[0][1] * frustum[0][1] + frustum[0][2] * frustum[0][2] );
	frustum[0][0] /= t;
	frustum[0][1] /= t;
	frustum[0][2] /= t;
	frustum[0][3] /= t;

	/* Extract the numbers for the LEFT plane */
	frustum[1][0] = clip[ 3] + clip[ 0];
	frustum[1][1] = clip[ 7] + clip[ 4];
	frustum[1][2] = clip[11] + clip[ 8];
	frustum[1][3] = clip[15] + clip[12];

	/* Normalize the result */
	t = sqrt( frustum[1][0] * frustum[1][0] + frustum[1][1] * frustum[1][1] + frustum[1][2] * frustum[1][2] );
	frustum[1][0] /= t;
	frustum[1][1] /= t;
	frustum[1][2] /= t;
	frustum[1][3] /= t;

	/* Extract the BOTTOM plane */
	frustum[2][0] = clip[ 3] + clip[ 1];
	frustum[2][1] = clip[ 7] + clip[ 5];
	frustum[2][2] = clip[11] + clip[ 9];
	frustum[2][3] = clip[15] + clip[13];

	/* Normalize the result */
	t = sqrt( frustum[2][0] * frustum[2][0] + frustum[2][1] * frustum[2][1] + frustum[2][2] * frustum[2][2] );
	frustum[2][0] /= t;
	frustum[2][1] /= t;
	frustum[2][2] /= t;
	frustum[2][3] /= t;

	/* Extract the TOP plane */
	frustum[3][0] = clip[ 3] - clip[ 1];
	frustum[3][1] = clip[ 7] - clip[ 5];
	frustum[3][2] = clip[11] - clip[ 9];
	frustum[3][3] = clip[15] - clip[13];

	/* Normalize the result */
	t = sqrt( frustum[3][0] * frustum[3][0] + frustum[3][1] * frustum[3][1] + frustum[3][2] * frustum[3][2] );
	frustum[3][0] /= t;
	frustum[3][1] /= t;
	frustum[3][2] /= t;
	frustum[3][3] /= t;

	/* Extract the FAR plane */
	frustum[4][0] = clip[ 3] - clip[ 2];
	frustum[4][1] = clip[ 7] - clip[ 6];
	frustum[4][2] = clip[11] - clip[10];
	frustum[4][3] = clip[15] - clip[14];

	/* Normalize the result */
	t = sqrt( frustum[4][0] * frustum[4][0] + frustum[4][1] * frustum[4][1] + frustum[4][2] * frustum[4][2] );
	frustum[4][0] /= t;
	frustum[4][1] /= t;
	frustum[4][2] /= t;
	frustum[4][3] /= t;

	/* Extract the NEAR plane */
	frustum[5][0] = clip[ 3] + clip[ 2];
	frustum[5][1] = clip[ 7] + clip[ 6];
	frustum[5][2] = clip[11] + clip[10];
	frustum[5][3] = clip[15] + clip[14];

	/* Normalize the result */
	t = sqrt( frustum[5][0] * frustum[5][0] + frustum[5][1] * frustum[5][1] + frustum[5][2] * frustum[5][2] );
	frustum[5][0] /= t;
	frustum[5][1] /= t;
	frustum[5][2] /= t;
	frustum[5][3] /= t;
	
}


/****************************************************** end of camera methods ********************************************************/



Vector Renderer::screenToWorld( int x, int y, float z ) // 2d to 3d transoform
{
	setFuncName( "Renderer::screenToWorld" );
	Vector position;
	float depth[1];
	GLdouble modelm[16], projm[16], pos[3];
	int view[4];

	//flip y axis
	y=getScreenH()-y-1;
	
	// get matrices
	glGetDoublev( GL_MODELVIEW_MATRIX, modelm );
	glGetDoublev( GL_PROJECTION_MATRIX, projm );
	glGetIntegerv( GL_VIEWPORT, view );

	// read depth component
	if( z == -1 )
		glReadPixels( x, y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, depth );
	else depth[0]=z;

	//unproject
	gluUnProject( float(x), float(y), depth[0], modelm, projm, view, &pos[0], &pos[1], &pos[2] );

	position.set( pos[0], pos[1], pos[2] );
	return position;
}

Vector Renderer::worldToScreen( Vector objpos ) //3d to 2d transoform
{
	setFuncName( "Renderer::worldToScreen" );
	Vector position;
	GLdouble modelm[16], projm[16], pos[3];
	int view[4];

	// get matrices
	glGetDoublev( GL_MODELVIEW_MATRIX, modelm );
	glGetDoublev( GL_PROJECTION_MATRIX, projm );
	glGetIntegerv( GL_VIEWPORT, view );

	//unproject
	gluProject( objpos.x(), objpos.y(), objpos.z(), modelm, projm, view, &pos[0], &pos[1], &pos[2] );

	//flip y-axis
	pos[1] = getScreenH()-pos[1];
	position.set( pos[0], pos[1], pos[2] );
	return position;
}

tgl_object *Renderer::loadTGLObject( const char *fname, float pre_scale ) // load and  return newly created TGL object
{
	setFuncName( "Renderer::loadTGLObject" );
	int i;
	// search already loaded models
	for( i=0; i<MAX_TGL_OBJECTS; i++ )
		if( !strcmp( fname, object_file_names[i] ) ) {
			object_clients[i]++; // new client for obj
			msgOut( MSG_DEBUG_4, "TGL Model %s found. Total clients %d", fname, object_clients[i] );
			return objects[i];
		}
	
	// if not found find emty slot for new object
	for( i=0; i<MAX_TGL_OBJECTS; i++ )
		if( !strcmp( "", object_file_names[i] ) ) break;
	if( i>MAX_TGL_OBJECTS ) msgOut( MSG_FAILURE, "TGL Object limit exceeded (%d)", MAX_TGL_OBJECTS );
	
	if( objects[i]==NULL ) {
		objects[i]=new tgl_object( fname );
		objects[i]->pre_scale(pre_scale,pre_scale,pre_scale);
		object_clients[i]++;
		strcpy( object_file_names[i], fname );
		msgOut( MSG_INIT, "TGL Model %s loaded", fname );
	}
	else {
		msgOut( MSG_WARN, "Bug in Renderer::loadTGLObject, please make a bug report" );
	}
	
	return objects[i];
}

tgl_object *Renderer::getTGLObject( const char *fname ) // return existing TGL object
{
	int i;
	// search already loaded models
	for( i=0; i<MAX_TGL_OBJECTS; i++ )
	{
		if( !strcmp( fname, object_file_names[i] ) )
		{
			msgOut( MSG_DEBUG_3, "TGL Model %s found", fname );
			 return objects[i];
		}
	}
	
	msgOut( MSG_FAILURE, "TGL Model %s not loaded", fname );
	return NULL;
}

void Renderer::freeTGLObject( const char *fname ) // unload TGL object
{
	setFuncName( "Renderer::freeTGLObject" );
	int i;
	
	// search already loaded models
	
//	msgOut( MSG_DEBUG_4, "Freeing TGL object %s", fname );
	
	for( i=0; i<MAX_TGL_OBJECTS; i++ )
		if( !strcmp( fname, object_file_names[i] ) )
		{
			if( objects[i] != NULL )
			{
				object_clients[i]--;
				msgOut( MSG_DEBUG_4, "Freeing TGL object, %s has %d clients left", fname, object_clients[i] );
				if( object_clients[i] <= 0 ) { // if obj doesn't have clients, delete it
					delete objects[i];
					objects[i]=NULL;
					strcpy( object_file_names[i], "" );
					msgOut( MSG_UNINIT, "TGL Model %s deleted", fname );
					return;
				}
			} else { msgOut( MSG_WARN, "Bug in Renderer::freeTGLObject, please make a bug report" ); }
		}
		
	if( i>MAX_TGL_OBJECTS ) //if not found
		msgOut( MSG_WARN, "Tried to delete already deleted or not loaded TGL object %s", fname );
}

void Renderer::freeTGLObject( tgl_object *obj ) // unload TGL object
{
	setFuncName( "Renderer::freeTGLObject" );
	int i;
	string fname;
	
	if( obj==NULL ) {
		msgOut( MSG_WARN, "Tried to delete NULL TGL object" );
		return;
	}
	
	// search already loaded models
	for( i=0; i<MAX_TGL_OBJECTS; i++ )
		if( objects[i]==obj )
		{
			object_clients[i]--;
			fname=object_file_names[i];
			msgOut( MSG_DEBUG_4, "Freeing TGL object, %s has %d clients left", object_file_names[i], object_clients[i] );
			if( object_clients[i] <= 0 ) { // if obj doesn't have clients, delete it
				delete objects[i];
				objects[i]=NULL;
				strcpy( object_file_names[i], "" );
				msgOut( MSG_UNINIT, "TGL Model %s deleted", fname.c_str() );
				return;
			}
		}
		
	if( i>MAX_TGL_OBJECTS ) //if not found
		msgOut( MSG_WARN, "Tried to delete already deleted or not loaded TGL object" );
}



void Renderer::renderSelectionBox( Vector p1, Vector p2 )  //render 2d box, p1 and p2 are opposing corners
{
	setFuncName( "Renderer::renderSelectionBox" );
	
	setRenderingMode( RENDER_INTERFACE ); // switch to 2d projection
	
	glBegin( GL_POLYGON );
		glNormal3f( 1,0,0 );
		glTexCoord2f( 0,0 );
		
		glColor4f( 0, 1, 0, 0.3 );
		glVertex3f( p1.x(), p1.y(), 0 );
	
		glColor4f( 0.5, 0.75, 0.25, 0.3 );
		glVertex3f( p2.x(), p1.y(), 0 );
	
		glColor4f( 0, 0.5, 0.5, 0.3 );
		glVertex3f( p2.x(), p2.y(), 0 );
		
		glColor4f( 0.5, 0.75, 0.25, 0.3 );
		glVertex3f( p1.x(), p2.y(), 0 );
	glEnd();
}

void Renderer::renderBox( Vector p1, Vector p2, Vector textureoff1, Vector textureoff2, float r, float g, float b, float a, GLint textureid )  //render 2d box, p1 and p2 are opposing corners
{
	setFuncName( "Renderer::renderTextureBox" );
	
	setRenderingMode( RENDER_INTERFACE ); // switch to 2d projection
	
	if( textureid ){
		glEnable( GL_TEXTURE_2D );
		glBindTexture( GL_TEXTURE_2D, textureid );
	}
	
	glBegin( GL_POLYGON );
		glNormal3f( 1,0,0 );
		glColor4f( r, g, b, a );
		
		glTexCoord2f( textureoff1.x(), textureoff1.y() );
		glVertex3f( p1.x(), p1.y(), 0 );
		glTexCoord2f( textureoff2.x(), textureoff1.y() );
		glVertex3f( p2.x(), p1.y(), 0 );
		glTexCoord2f( textureoff2.x(), textureoff2.y() );
		glVertex3f( p2.x(), p2.y(), 0 );
		glTexCoord2f( textureoff1.x(), textureoff2.y() );
		glVertex3f( p1.x(), p2.y(), 0 );
	glEnd();
	
	glDisable( GL_TEXTURE_2D );
}

void Renderer::renderBox( Vector p1, Vector p2, float r, float g, float b, float a, GLint textureid )  //render 2d box, p1 and p2 are opposing corners
{
	setFuncName( "Renderer::renderBox" );
	
	setRenderingMode( RENDER_INTERFACE ); // switch to 2d projection
	
	if( textureid ){
		glEnable( GL_TEXTURE_2D );
		glBindTexture( GL_TEXTURE_2D, textureid );
	}
	
	glBegin( GL_POLYGON );
		glNormal3f( 1,0,0 );
		glColor4f( r, g, b, a );
		
		glTexCoord2f( 0, 0 );
		glVertex3f( p1.x(), p1.y(), 0 );
		glTexCoord2f( 1, 0 );
		glVertex3f( p2.x(), p1.y(), 0 );
		glTexCoord2f( 1, 1 );
		glVertex3f( p2.x(), p2.y(), 0 );
		glTexCoord2f( 0, 1 );
		glVertex3f( p1.x(), p2.y(), 0 );
	glEnd();
	
	if( textureid ) glDisable( GL_TEXTURE_2D );
}

void Renderer::renderMouseCursor( Vector p )
{
	setFuncName( "Renderer::renderMouseCursor" );

	if( mouseCursor!=NULL ) { //render textured cursor if avaiable
		renderMouseCursor( p, mouseCursor );
		return;
	}
	
	setRenderingMode( RENDER_INTERFACE );
	
	glBegin( GL_TRIANGLES );
		//highlighted
		glColor4f( 0, 1, 0, 1 );
		glVertex3f( p.x(), p.y(), 0 );
	
		glColor4f( 0.5, 0.75, 0.25, 0.5 );
		glVertex3f( p.x()+20, p.y()+10, 0 );
	
		glColor4f( 0, 0.5, 0.5, 0.5 );
		glVertex3f( p.x()+10, p.y()+10, 0 );
		
		//shadow
		glColor4f( 0, 0.7, 0, 1 );
		glVertex3f( p.x(), p.y(), 0 );
	
		glColor4f( 0.5*.7, 0.75*.7, 0.25*.7, 0.5 );
		glVertex3f( p.x()+10, p.y()+10, 0 );
	
		glColor4f( 0, 0.5, 0.5, 0.5 );
		glVertex3f( p.x()+10, p.y()+20, 0 );
		
	glEnd();
}

void Renderer::renderMouseCursor( Vector pos, Texture *curs )
{
//	setRenderingMode( RENDER_INTERFACE );

	renderBox( pos, pos+Vector(curs->width(),curs->height()), 1, 1, 1, 1, curs->getTextureID() );
}

void Renderer::setRenderingMode( enum RenderingModes mode )
{
	setFuncName( "Renderer::setRenderingMode" );
	
	if( mode == currentRenderingMode ) return; //avoid unnecessary changes
	
	switch( mode )
	{
		case RENDER_TGL_OBJECT:
			glLightfv(GL_LIGHT0, GL_POSITION, sunPos.xyz() ); //light pos shouldn't move when camera moves...
			msgOut( MSG_DEBUG_5, "Setting TGL Object rendering mode" );
			// set 3d projection
//			if( currentRenderingMode=RENDER_TERRAIN|| currentRenderingMode!=RENDER_TGL_OBJECT||currentRenderingMode!=RENDER_TERRAIN||currentRenderingMode!=RENDER_3D )
			if( currentRenderingMode==RENDER_INTERFACE||currentRenderingMode==RENDER_2D )
			{
//				glViewport( 0, 0, getScreenW(), getScreenH() );
				set3DModelView();
			}
			applyLights();
			if( multitexturing ) { //disable multitexturing stuff
				glActiveTextureARB_ptr(GL_TEXTURE0_ARB);
				glDisable(GL_TEXTURE_2D);
				glActiveTextureARB_ptr(GL_TEXTURE1_ARB);
				glDisable(GL_TEXTURE_2D);
				glActiveTextureARB_ptr(GL_TEXTURE0_ARB);
			}
//			glEnable(GL_TEXTURE_2D);
			glColor4f( 1,1,1,1 );
			glDisable(GL_NORMALIZE);
//			glEnable(GL_NORMALIZE);
			glEnable(GL_DEPTH_TEST);
			setFog( fog_type, fog_start, fog_end, fog_density );
/*			glEnable(GL_FOG);
			glFogf(GL_FOG_MODE,GL_LINEAR);
			glFogfv(GL_FOG_COLOR,fogColor.RGBA());
			glFogf(GL_FOG_START,100.0f);
			glFogf(GL_FOG_END,maxViewDistance - 100.0f);*/
			currentRenderingMode=mode;
			break;
			
		case RENDER_TERRAIN:
			msgOut( MSG_DEBUG_5, "Setting Terrain rendering mode" );
			
			// set 3d projection
//			if( currentRenderingMode!=RENDER_TGL_OBJECT||currentRenderingMode!=RENDER_TERRAIN||currentRenderingMode!=RENDER_3D )
			if( currentRenderingMode==RENDER_INTERFACE||currentRenderingMode==RENDER_2D )
//			if( currentRenderingMode != RENDER_TERRAIN )
			{
//				glViewport( 0, 0, getScreenW(), getScreenH() );
//				glMatrixMode(GL_PROJECTION);
//				glLoadIdentity();
//				gluPerspective( 60.0f,(float)getScreenW()/(float)getScreenH(),0.65,maxViewDistance );
				set3DModelView();
			}
			glDisable( GL_TEXTURE_2D );
			glDisable( GL_CULL_FACE );
			glDisable(GL_NORMALIZE);
			glEnable(GL_DEPTH_TEST);
			glColor4f( 1,1,1,1 );
			glEnable(GL_FOG);
			setBlendFunc( FUNC_SOLID );
			glDisable(GL_LIGHTING);
			setFog( fog_type, fog_start, fog_end, fog_density );
/*			glFogf(GL_FOG_MODE,GL_LINEAR);
			glFogfv(GL_FOG_COLOR,fogColor.RGBA());
			glFogf(GL_FOG_START,100.0f);
			glFogf(GL_FOG_END,maxViewDistance - 100.0f);*/
			currentRenderingMode=mode;
			break;
			
		case RENDER_INTERFACE:
			msgOut( MSG_DEBUG_5, "Setting Interface rendering mode" );
			
			//set a 1:1 pixel mapping, with (0,0,0) in the top-left of the viewport and (width,height,0) in the bottom-right:
			if( currentRenderingMode!=RENDER_INTERFACE || currentRenderingMode!=RENDER_2D )
			{
//				glViewport( 0, 0, getScreenW(), getScreenH() );
				glMatrixMode(GL_PROJECTION);
				glLoadIdentity();
				glMatrixMode(GL_MODELVIEW);
				glLoadIdentity();
				glScalef(2.0f / (float)getScreenW(), -2.0f / (float)getScreenH(), 1.0f);
				glTranslatef(-((float)getScreenW() / 2.0f), -((float)getScreenH() / 2.0f), 0.0f);
			}
			if( multitexturing ) { //disable multitexturing stuff
				glActiveTextureARB_ptr(GL_TEXTURE1_ARB);
				glDisable(GL_TEXTURE_2D);
				glActiveTextureARB_ptr(GL_TEXTURE0_ARB);
				glDisable(GL_TEXTURE_2D);
			} else glDisable(GL_TEXTURE_2D);
//			glDisable(GL_TEXTURE_2D);
			
			glDisable( GL_CULL_FACE );
			glDisable(GL_NORMALIZE);
			glColor4f( 1,1,1,1 );
			glDisable(GL_DEPTH_TEST);
			glShadeModel(GL_SMOOTH);
			glDisable(GL_FOG);
			setBlendFunc( FUNC_ALPHA );
			glDisable(GL_LIGHTING);
			currentRenderingMode=mode;
			break;
			
		case RENDER_3D:
			// set 3d projection
			if( currentRenderingMode==RENDER_INTERFACE || currentRenderingMode==RENDER_2D )
			{
//				glViewport( 0, 0, getScreenW(), getScreenH() );
//				glMatrixMode(GL_PROJECTION);
//				glLoadIdentity();
//				gluPerspective( 60.0f,(float)getScreenW()/(float)getScreenH(),0.65,maxViewDistance );
				set3DModelView();
				currentRenderingMode=mode;
			}
			break;
			
		case RENDER_2D:
			// set 2d projection
			if( currentRenderingMode!=RENDER_INTERFACE || currentRenderingMode!=RENDER_2D )
			{
//				glViewport( 0, 0, getScreenW(), getScreenH() );
				glMatrixMode(GL_PROJECTION);
				glLoadIdentity();
				glMatrixMode(GL_MODELVIEW);
				glLoadIdentity();
				glScalef(2.0f / (float)getScreenW(), -2.0f / (float)getScreenH(), 1.0f);
				glTranslatef(-((float)getScreenW() / 2.0f), -((float)getScreenH() / 2.0f), 0.0f);
				currentRenderingMode=mode;
			}
			break;
			
		default:
			break;
	}
	
//				currentRenderingMode=mode;
}


bool Renderer::checkError( const char *str )
{
	GLenum error = msg::getLastGlError();
	char *sdl_error;

	if( error ) {
		char *err = (char *) gluErrorString( error );
		if( str!=NULL )
			msgOut( MSG_ERROR, "%s (error after function %s): GL error: %s\n", str, getLastFuncName(), err );
		else
			msgOut( MSG_ERROR, "GL error (error after function %s): %s\n",getLastFuncName(), err );
		return true;
	}

	sdl_error = SDL_GetError( );

	if( sdl_error[0] != '\0' ) {
		if( str!=NULL )
			msgOut( MSG_ERROR, "%s (error after function %s): SDL error: %s\n", str, getLastFuncName(), sdl_error );
		else
			msgOut( MSG_ERROR, "SDL error (error after function %s): %s\n", getLastFuncName(), sdl_error );
		SDL_ClearError();
		return true;
	}

	return false;
}

Texture *Renderer::loadTexture( const char *texturefile, enum AlphaModes alpha, bool mipmap, int minfilter, int magfilter, int wrapmode )
{
	setFuncName( "Renderer::loadTexture" );
	
	int i, slot=0;
	Texture *result;

	msgOut( MSG_DEBUG_4, "Loading/searching texture %s", texturefile );
	
	//check if texture is already loaded
	if( (result=getTexture( texturefile, false )) )
	{
		result->clients++;
		msgOut( MSG_DEBUG_4, "Texture %s has now %d clients", texturefile, result->clients );
		return result;
	}

	//find unused slot
	for( i=0; i<MAX_TEXTURES; i++ )
		if( textures[i]==NULL ) {
			slot=i;
			break;
		}
	if( i==MAX_TEXTURES ) {
		msgOut( MSG_ERROR, "Maximum number of textures reached (%d)!\n", MAX_TEXTURES );
		return NULL;
	}

	//load
	textures[slot]=new Texture( texturefile, alpha, mipmap, minfilter, magfilter, wrapmode );

	if( !textures[slot]->getTextureID() ) //if failed
	{
		msgOut( MSG_ERROR, "Failed loading texture %s", texturefile );
		textures[slot]=NULL;
		return NULL;
	} else {
		return textures[slot];
	}
}


GLint Renderer::getTextureID( const char *texturefile, bool warn )
{
	int i;
	
	msgOut( MSG_DEBUG_5, "Seeking texture %s", texturefile );
		
	for( i=0; i<MAX_TEXTURES; i++ ) //search
		if( textures[i]!=NULL )
			if( textures[i]->isSameFile( texturefile ) ) return textures[i]->getTextureID();

	if( warn ) msgOut( MSG_WARN, "Texture %s not loaded", texturefile );
		
	return 0;
}

Texture* Renderer::getTexture( const char *texturefile, bool warn )
{
	int i;
	
	msgOut( MSG_DEBUG_5, "Seeking texture %s", texturefile );
		
	for( i=0; i<MAX_TEXTURES; i++ ) //search
		if( textures[i]!=NULL )
			if( textures[i]->isSameFile( texturefile ) ) return textures[i];

	if( warn ) msgOut( MSG_WARN, "Texture %s not loaded", texturefile );
		
	return NULL;
}

void Renderer::freeTexture( string texturefile )
{
	int i;
	msgOut( MSG_DEBUG_5, "Seeking texture %s", texturefile.c_str() );
	
	for( i=0; i<MAX_TEXTURES; i++ ) //search
		if( textures[i]!=NULL )
			if( textures[i]->isSameFile( texturefile.c_str() ) ) {
				textures[i]->clients--;
				msgOut( MSG_DEBUG_4, "Freeing texture. %s has now %d clients left", texturefile.c_str(), textures[i]->clients );
				if( textures[i]->clients <= 0 ) {
					delete textures[i];
					textures[i]=NULL;
					msgOut( MSG_DEBUG_3, "Texture %s unloaded", texturefile.c_str() );
				}
				return;
			}
			
	msgOut( MSG_ERROR, "%s not loaded, unable to delete", texturefile.c_str() );
}


void Renderer::renderText( int x, int y, const char *str )
{
	setFuncName( "Renderer::renederText" );
	
	//vars
	if( currentfont==NULL ) return;
	BFont *font=currentfont;
	float r,g,b,a, hr, hg, hb, ha;
	int hotkey=0;
	int i, len, xp=0, h;
	Vector p1, p2, t1, t2;
	float texw=256, texh=256;
	unsigned char c;
	
	//gather information
	len = strlen( str );
	if( len <= 0 ) return;
	h=font->getLineHeight();
	font->getRGBA( r, g, b, a );
	font->getHotkeyRGBA( hr, hg, hb, ha );
	texw=font->getTexture()->width();
	texh=font->getTexture()->height();
	
	//loop all chars
	for( i=0, xp=0; i<len; i++ )
	{
		//check hotkey:
		c = str[i];
		if( hotkey > 0 ) hotkey--;
		if( c=='&' && hotkey==0 ) //use hotkey color for next char
		{
			hotkey++;
			i++; //skip one char
			if( i >= len ) return; // no char after &
			c = str[i];
		}
		
		// calculate coordinates:
		p1.set( x+xp, y, 0 );
//		p2.set( x+xp+font->getCharWidth(c), y+h, 0 );
		p2.set( x+xp+font->getBlockWidth(), y+h, 0 );
		t1.set( (float)font->getCharXPos(c)/texw, (float)font->getCharYPos(c)/texh, 0 );
//		t2.set( t1.x()+((float)font->getCharWidth(c)/texw), t1.y()+((float)h/texh), 0 );
		t2.set( t1.x()+((float)font->getBlockWidth()/texw), t1.y()+((float)h/texh), 0 );
		
		// render
		if( hotkey== 1 && c!='&' )
			renderBox( p1, p2, t1, t2, hr, hg, hb, ha, font->getTextureID() );
		else
			renderBox( p1, p2, t1, t2, r, g, b, a, font->getTextureID() );
		
		xp+=font->getCharWidth( str[i] ); // add char width to x offset
	}
}

float Renderer::textLength( const char *str )
{
	setFuncName( "Renderer::renederText" );
	
	//vars
	if( currentfont==NULL ) return 0;
	BFont *font=currentfont;
	int hotkey=0;
	int i, len, xp=0;
	unsigned char c;
	
	//gather information
	len = strlen( str );
	if( len <= 0 ) return 0;
	
	//loop all chars
	for( i=0, xp=0; i<len; i++ )
	{
		//check hotkey:
		c = str[i];
		if( hotkey > 0 ) hotkey--;
		if( c=='&' && hotkey==0 ) //use hotkey color for next char
		{
			hotkey++;
			i++; //skip one char
			if( i >= len ) return xp; // no char after &
			c = str[i];
		}
		
		xp+=font->getCharWidth( str[i] ); // add char width to x offset
	}
	
	return xp;
}

void Renderer::renderPoint( Vector pos, Color col, float size )
{
	setFuncName( "Renderer::renderPoint" );
	
	setRenderingMode( RENDER_INTERFACE );
	
	glPointSize(size);
	glBegin(GL_POINTS);
		glColor4fv( col.RGBA() );
		glVertex3fv( pos.xyz() );
	glEnd();


	exitFunc();
}

void Renderer::renderInstantParticle( Vector pos, Vector size, Texture *tex, Color col, bool enable_decimation  )
{
	int amount;
	setFuncName("Renderer::renderParticle");
	
	if( particle_amount==-1 ) return; //particles disabled
	
	if( enable_decimation ) {
		// choose greater visibility
		if( particle_amount>particle_minvis ) amount=particle_minvis;
		else amount=particle_amount;
		if( amount == 0 ) amount=1; // 0, 1=100%
	
		if( !(particlecounter%amount) )
		{
			particlecounter++;
		}
		else
		{
			particlecounter++;
			return; //don't render this
		}
	}
	
	setRenderingMode( RENDER_TGL_OBJECT );
	
	glEnable( GL_TEXTURE_2D );
	glBindTexture( GL_TEXTURE_2D, tex->getTextureID() );
	glDisable( GL_LIGHTING );
	glEnable( GL_COLOR_MATERIAL );
	if( !particle_depthbuffer ) glDisable( GL_DEPTH_TEST );
	
	glPushMatrix();
	glTranslatef( pos.x(), pos.y(), pos.z() );
	glRotatef( RAD_TO_DEG(cameraAngle), 0, 0, 1 );
	glRotatef( 90-RAD_TO_DEG(cameraPitch), -1, 0, 0 );
	setBlendFunc( particle_blend );
	size.scale(0.5);
	
	glBegin( GL_POLYGON );
		glColor4fv( col.RGBA() );
		
		glTexCoord2f( 0, 0 );
		glVertex3f( -size.x(), -size.y(), 0 );
		glTexCoord2f( 1, 0 );
		glVertex3f( size.x(), -size.y(), 0 );
		glTexCoord2f( 1, 1 );
		glVertex3f( size.x(), size.y(), 0 );
		glTexCoord2f( 0, 1 );
		glVertex3f( -size.x(), size.y(), 0 );
	glEnd();
	glPopMatrix();
	
	if( !particle_depthbuffer ) glEnable( GL_DEPTH_TEST );
	glDisable( GL_TEXTURE_2D );
}

void Renderer::renderParticle( Vector pos, Vector size, Texture *tex, Color col, bool enable_decimation  )
{
	int amount;
	ParticleJob job;
	if( particle_amount==-1 ) return; //particles disabled
	
	if( enable_decimation ) {
		// choose greater visibility
		if( particle_amount>particle_minvis ) amount=particle_minvis;
		else amount=particle_amount;
		if( amount == 0 ) amount=1; // 0, 1=100%
	
		if( !(particlecounter%amount) )
		{
			particlecounter++;
		}
		else
		{
			particlecounter++;
			return; //don't render this
		}
	}
	setFuncName("Renderer::renderParticle");
	
	//create new job
	if( particle_post_rendering ) {
		job.size=size;
		job.pos=pos;
		job.col=col;
		job.tex=tex;
		job.blend=particle_blend;
		particle_jobs.push_back(job);
	} else //render instantly
		renderInstantParticle( pos, size, tex, col, false );

}

void Renderer::renderParticleQueue()
{
	ParticleJob job;
	Vector ppos;
	int i, c;
	
	GLdouble modelm[16], projm[16], pos[3];
	int view[4];
	
	if( !particle_post_rendering ) return; //queue disabled
	
	c=particle_jobs.size();
	
	// do z-sorting:
	if( particle_sorting )
	{
		// get matrices
		glGetDoublev( GL_MODELVIEW_MATRIX, modelm );
		glGetDoublev( GL_PROJECTION_MATRIX, projm );
		glGetIntegerv( GL_VIEWPORT, view );

		for( i=0; i<c; i++ )
		{
			ppos=particle_jobs[i].pos;
			// calculate z-pos
			gluProject( ppos.x(), ppos.y(), ppos.z(), modelm, projm, view, &pos[0], &pos[1], &pos[2] );
			particle_jobs[i].depthpos=pos[2];
/*			job=particle_jobs[i];
			ppos=job.pos;
			// calculate z-pos
			gluProject( ppos.x(), ppos.y(), ppos.z(), modelm, projm, view, &pos[0], &pos[1], &pos[2] );
			job.depthpos=pos[2];
			particle_jobs[i]=job;*/
		}

		// sort by z-pos
		sort( particle_jobs.begin(), particle_jobs.end() );
	}
	
	//render
	for( i=0; i<c; i++ )
	{
		job=particle_jobs[i];
//		setParticleMode( job.blend, 0 );
		particle_blend=job.blend; // DIRECT
		renderInstantParticle( job.pos, job.size, job.tex, job.col, false );
/*		job=particle_jobs[0];
		setParticleMode( job.blend, 0 );
		renderInstantParticle( job.pos, job.size, job.tex, job.col );
		particle_jobs.erase(particle_jobs.begin());*/
	}

	particle_jobs.clear(); //done
}

void Renderer::setParticleMode( enum BlendFunc blending, int minvisibility )
{
	particle_blend=blending;
	if( particle_minvis!=minvisibility )
		particlecounter=0;
	particle_minvis=minvisibility;
}

void Renderer::addParticleEffect( Vector pos, Vector size, Vector endsize,  Color col, Color endcol, float lifetime, Texture *tex )
{
	ParticleJob job;
	int amount;
	if( particle_amount==-1 ) return; //particles disabled
	
	// choose greater visibility
	if( particle_amount>particle_minvis ) amount=particle_minvis;
	else amount=particle_amount;
	if( amount == 0 ) amount=1; // 0, 1=100%
	
	if( !(particlecounter%amount) )
	{
		particlecounter++;
	}
	else
	{
		particlecounter++;
		return; //don't render this
	}
	
	if( lifetime < 0.01 ) return; //too short time
	
	//new job
	job.pos=pos;
	job.blend=particle_blend;
	
	job.size=size;
	job.diffsize=(endsize-size)/lifetime; //how much size changes in one second
	
	job.col=col;
	job.diffcol=endcol-col;
	job.diffcol.scaleValues( 1/lifetime ); //how much color changes in one second

	job.tex=tex;
	job.lifetime=lifetime;
	
	particle_effects.push_back( job );
}

void Renderer::processParticleEffects( float timediff )
{
	int i, c;
	ParticleJob job;
	
	timediff/=1000.0;
	
	c=particle_effects.size();
//printf("%d ", c );	
	for( i=c-1; i>=0; i-- ) //loop backwards
	{
		job=particle_effects[i];
		job.lifetime-=timediff;

		job.size+=(job.diffsize*timediff);
//		job.col+=(job.diffcol*timediff); // FIXME
		job.col+=(Color(job.diffcol)*timediff);
		
		particle_effects[i]=job;

		if( job.lifetime <= 0 ) //delete
		{
			particle_effects.erase( particle_effects.begin()+i );
		}
	}
}

void Renderer::renderParticleEffects()
{
	int i, c;
	ParticleJob job;

	c=particle_effects.size();
	for( i=0; i<c; i++ )
	{
		job=particle_effects[i];
//		setParticleMode( job.blend, 0 );
		particle_blend=job.blend; // DIRECT
//		particle_jobs.push_back( particle_effects[i] ); //DIRECT
//		renderInstantParticle( job.pos, job.size, job.tex, job.col, false );
		renderParticle( job.pos, job.size, job.tex, job.col, false );
	}
		
//	particle_effects.clear();
}

void Renderer::setBlendFunc( enum BlendFunc func )
{
	if( currentBlendFunc==func ) return;
	
	switch( func )
	{
		case FUNC_SOLID:
			glDisable( GL_BLEND );
			break;
		case FUNC_ADD:
			glBlendFunc( GL_ONE, GL_ONE );
			glEnable( GL_BLEND );
			break;
		case FUNC_COLORADD:
			glBlendFunc( GL_ONE, GL_SRC_COLOR );
			glEnable( GL_BLEND );
			break;
		case FUNC_MULTIPLY:
			glBlendFunc( GL_DST_COLOR, GL_ZERO );
			glEnable( GL_BLEND );
			break;
		case FUNC_SCREEN:
			glBlendFunc( GL_ONE, GL_ONE_MINUS_SRC_COLOR );
			glEnable( GL_BLEND );
			break;
		case FUNC_ALPHA:
			glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
			glEnable( GL_BLEND );
			break;
		default:
			glDisable( GL_BLEND );
			break;
	}
}

void Renderer::setSkyColor( Color col )
{
	fogColor=col;
	glClearColor(col.r(),col.g(),col.b(),col.a());
	glFogfv(GL_FOG_COLOR,fogColor.RGBA());
}

void Renderer::setSun( Vector pos, Color diffuse, Color ambient )
{
//ambient.set(1,1,1,1);
	sunDiffuse=diffuse;
	sunAmbient=ambient;
	sunPos=pos;
//	pos.set(0,0,0);
	
	applyLights();
}

void Renderer::applyLights()
{
	//light0 is sun
	glLightModelfv( GL_LIGHT_MODEL_AMBIENT, sunAmbient.RGBA() );
	glLightfv(GL_LIGHT0, GL_DIFFUSE, sunDiffuse.RGBA() );
	glLightfv(GL_LIGHT0, GL_AMBIENT, sunAmbient.RGBA() );
//	glLightfv(GL_LIGHT0, GL_SPECULAR, sunDiffuse.RGBA() );
	glLightfv(GL_LIGHT0, GL_POSITION, sunPos.xyz() );
	glEnable(GL_LIGHT0);
}

void Renderer::renderLine( Vector pos1, Vector pos2, Color col, enum BlendFunc blend )
{
	const float *c;
	if( !render_lines ) return; //lines disabled
	
	c=col.RGBA();
	if( blend==FUNC_ADD && c[0]==0 && c[1]==0 && c[2]==0 ) return; //nothing to draw
	if( blend==FUNC_ALPHA && c[3]==0) return; //nothing to draw
	
	setFuncName("Renderer::renderLine");
	
	
	// choose greater visibility
	setRenderingMode( RENDER_TGL_OBJECT );
//	glDisable( GL_DEPTH_TEST );
	glDisable( GL_LIGHTING );
	glDisable( GL_TEXTURE_2D );
	
	setBlendFunc( blend );
	
	glBegin( GL_LINES );
		glColor4fv( col.RGBA() );
		
		glVertex3fv( pos1.xyz() );
		glVertex3fv( pos2.xyz() );
	glEnd();
	
//	glEnable( GL_DEPTH_TEST );
}

void Renderer::renderInterfaceLine( Vector pos1, Vector pos2, Color col, enum BlendFunc blend )
{
	const float *c;
//	if( !render_lines ) return; //lines disabled
	
	c=col.RGBA();
	if( blend==FUNC_ADD && c[0]==0 && c[1]==0 && c[2]==0 ) return; //nothing to draw
	if( blend==FUNC_ALPHA && c[3]==0) return; //nothing to draw
	
	setFuncName("Renderer::renderInterfaceLine");
	
	
	// choose greater visibility
	setRenderingMode( RENDER_INTERFACE );
	
	setBlendFunc( blend );
	
	glBegin( GL_LINES );
		glColor4fv( col.RGBA() );
		
		glVertex3fv( pos1.xyz() );
		glVertex3fv( pos2.xyz() );
	glEnd();
}

float Renderer::isVisible( Vector pos, float radius )
{
	// this piece of code is mostly from http://www.markmorley.com/opengl/frustumculling.html
	int p;
	float d, x=pos.x(), y=pos.y(), z=pos.z();

	for( p = 0; p < 6; p++ )
	{
		d = (frustum[p][0] * x) + (frustum[p][1] * y) +( frustum[p][2] * z) + frustum[p][3];
		if( d <= (-radius) )
		{
			return 0;
		}
	}
	return d + radius;
}

/*bool Renderer::isVisible( Vector pos, float radius )
{
	Vector scr;
	setRenderingMode( RENDER_3D );
	scr=worldToScreen( pos );
	if( scr > Vector(0,0,-999999) && scr < Vector(screen_w,screen_h,999999) ) //DIRECT
		return true;
	return false;
}
  */

void Renderer::setFog( int fogtype, float start, float end, float density )
{
	// fog
	if( fogtype!=0 ) glEnable(GL_FOG);
	
	switch(fogtype)
	{
		case 1: glFogf(GL_FOG_MODE,GL_LINEAR); break;
		case 2: glFogf(GL_FOG_MODE,GL_EXP); break;
		case 3: glFogf(GL_FOG_MODE,GL_EXP2); break;
		default: glDisable(GL_FOG); break;
	}
	glFogf(GL_FOG_DENSITY,density);
	glFogfv(GL_FOG_COLOR,fogColor.RGBA());
	glFogf(GL_FOG_START,start);
	glFogf(GL_FOG_END,end);
	
	//store settings
	fog_density=density;
	fog_start=start;
	fog_end=end;
	fog_type=fogtype;
}

void Renderer::setViewDistance( float dist )
{
	maxViewDistance=dist;
}

