/***************************************************************************
                          renderer.h  -  description
                             -------------------
    begin                : Sun Jul 22 2001
    copyright            : (C) 2001 by T Kontkanen
    email                : tkontkanen@mail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef RENDERER_H
#define RENDERER_H


/**Sets up OpenGL and handles rendering stuff
  *@author T Kontkanen
  */
#include "SDL/SDL.h"
#include "GL/gl.h"
#include "GL/glu.h"
#include "GL/glpng.h"

#include "tgl.h"
#include <vector>
#include <slist>
#include "allincludes.h"
#include "string.h"
#include "gamedefs.h"
#include "vector.h"
#include "texture.h"
#include "color.h"

using namespace TGL;

enum RenderingModes { RENDER_TGL_OBJECT, RENDER_INTERFACE, RENDER_TERRAIN, RENDER_3D, RENDER_2D };
enum CameraModes { CAMERA_ORBIT, CAMERA_FREE };
enum BlendFunc { FUNC_SOLID, FUNC_ADD, FUNC_SCREEN, FUNC_MULTIPLY, FUNC_COLORADD, FUNC_ALPHA };

class ParticleJob {
public:
	Vector pos, size, diffsize;
	Color col, diffcol;
	Texture *tex;
	float lifetime; // 0 = render once
	float depthpos;
	enum BlendFunc blend;
	
	/// Helpter for z-sorting. Returns true if this should be rendered before other particle.
	bool operator<(const ParticleJob& compare) const
	{
		if( depthpos<=compare.depthpos ) return false;
		return true;
	}
};


class BFont;
class Game;

class Renderer {
public:
	
	/// Constructor. w=screen width (pixels), h=screen height (pixels), depth=screen BPP (0=current depth), fullscr=use fullscreen mode if true
	Renderer( Game *parent, int w, int h, int depth, bool fullscr );
	/// Destructor
	~Renderer();
	
	//screen
	/// Update GL viewport when resizing rendering window
	void setVideoMode( int w, int , int depth, bool fullscr );
	/// Returns screen width in pixels
	int getScreenW() { return screen_w; }
	/// Returns screen height in pixels
	int getScreenH() { return screen_h; }
	/// Returns screen depth (bits per pixel). 0 means default depth
	int getScreenDepth() { return screen_depth; }
	/// Returns true if fullscreen mode is active
	bool getFullScreen() { return fullscreen; }
	/// Toggle between fullscreen and windowed mode
	void toggleFullScreen() {
		setFuncName( "Renderer::toggleFullScreen" );
		if( SDL_WM_ToggleFullScreen( screen ) )
			if( fullscreen ) fullscreen=false; //invert
			else fullscreen=true;
		else msg::msgOut( msg::MSG_FAILURE, "Unable to toggle fullscreen/windowed mode" );
	}
	/// Toggle wireframe/solid rendering
	void toggleWireframe() {
		setFuncName( "Renderer::toggleWireframe" );
		if( !wireframe ) { glPolygonMode(GL_FRONT_AND_BACK,GL_LINE); wireframe=true; }
		else { glPolygonMode(GL_FRONT_AND_BACK,GL_FILL); wireframe=false; }
		msgOut( MSG_DEBUG_3, "Wireframe mode toggled" );
	}
	/// Read values from config file. Constructor calls this at creation time.
	void readConfig();
	/// Set sky/fog color as col.
	void setSkyColor( Color col );
	/// Set sun position and color
	void setSun( Vector pos, Color diffuse, Color ambient );
	void setFog( int fogtype, float start, float end, float density );
	void setViewDistance( float dist );

	
	// rendering methods
	/// Swap GL doublebuffers
	void swapBuffers() { SDL_GL_SwapBuffers(); checkError(); }
	/// Clear GL doublebuffer
	void clearBuffer() { glClear( GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT ); }
	/// Render a 2d selection box
	void renderSelectionBox( Vector p1, Vector p2 );  //render 2d box, p1 and p2 are opposing corners
	/// Render a 2d box with custom color (RGBA), used in GUI rendering. Texture is rendered if textureid is some GL texture ID.
	void renderBox( Vector p1, Vector p2, float r, float g, float b, float a, GLint textureid=0 );
	/// Render a 2d box with custom Color, used in GUI rendering. Texture is rendered if textureid is some GL texture ID.
	void renderBox( Vector p1, Vector p2, Color col, GLint textureid=0 ) { renderBox( p1, p2, col.r(),col.g(),col.b(),col.a(), textureid ); }
	/// Render a 2d box with custom color (RGBA) and texture, used in text rendering. texoff1 and texoff2 specifies texture rendering offsets.
	void renderBox( Vector p1, Vector p2, Vector textureoff1, Vector textureoff2, float r, float g, float b, float a, GLint textureid );
	/// Render a 2d box with custom Color and texture, used in text rendering. texoff1 and texoff2 specifies texture rendering offsets.
	void renderBox( Vector p1, Vector p2, Vector textureoff1, Vector textureoff2, Color col, GLint textureid ) { renderBox( p1, p2, textureoff1, textureoff2,col.r(),col.g(),col.b(),col.a(), textureid ); }
	/// Render 2d-particle in 3d-space. Particle will be rendered instantly.
	void renderInstantParticle( Vector pos, Vector size, Texture *tex, Color col, bool enable_decimation=true );
	/// Render 2d-particle in 3d-space. Particle rendering job will be put into particle rendering queue.
	void renderParticle( Vector pos, Vector size, Texture *tex, Color col, bool enable_decimation=true );
	/// Renders all particles which are in the particle job queue.
	void renderParticleQueue();
	/// Set particle rendering parameters. Parameter minvisibility means amount of particles need to be rendered despite of setParticleAmount() value. Values:  1=100%, 2=50%, 3=33%, 4=25%, 5=20%, ..., 1/n.
	void setParticleMode( enum BlendFunc blending, int minvisibility=10000 );
	/// Set amount of rendered particles. Values:  1=100%, 2=50%, 3=33%, 4=25%, 5=20%, ..., 1/n.
	void setParticleAmount( int lod ) { if( lod<-1 ) lod=-1; particle_amount=lod; }
	/// Add particle effect that will be displayed and whose lifetime can be declared. Particle size and color are morphed from sizet to endsize and from colto endcol.
	void addParticleEffect( Vector pos, Vector size, Vector endsize,  Color col, Color endcol, float lifetime, Texture *tex );
	/// Process particle effects. Timediff is time differential (in seconds) between two calls.
	void processParticleEffects( float timediff );
	/// Render particle effects or add them to the particle queue.
	void renderParticleEffects();
	
	/// Set OpenGL state for different operations, enums: RENDER_TGL_OBJECT, RENDER_INTERFACE, RENDER_TERRAIN, RENDER_3D (view matrices only), RENDER_2D (view matrices only)
	void setRenderingMode( enum RenderingModes mode );
	/// Render 2d mouse cursor at p.
	void renderMouseCursor( Vector p );
	/// Render 2d mouse cursor at 'pos' using custom texture. Alpha blending is used.
	void renderMouseCursor( Vector pos, Texture *curs );
	/// Render 2d point at position (pos) wit custom color (col). Size is point radius in pixels.
	void renderPoint( Vector pos, Color col, float size=1 );
	/// Render a line in 3d space from pos1 to pos2.
	void renderLine( Vector pos1, Vector pos2, Color col, enum BlendFunc blend=FUNC_SOLID );
	/// Render a line in 2d space from pos1 to pos2 in RENDER_INTERFACE mode.
	void Renderer::renderInterfaceLine( Vector pos1, Vector pos2, Color col, enum BlendFunc blend=FUNC_SOLID );
	
	
	//camera methods
	/// Initialize camera, this must be called before any other camera operations
	void initCamera();
	/// Set camera movement modes: CAMERA_ORBIT or CAMERA_FREE. In orbit mode camera rotates around lookat point but in free mode it rotates around camera position.
	void setCameraMode( enum CameraModes mode );
	/// Set distance from camera to lookat point. Distance must be >= 0.001.
	void setCameraLookAtDistance( float dist );
	/// Adjust distance between camera and lookat point.
	void adjustCameraLookAtDistance( float adjdist );
	/// Return distance between camera and lookat point.
	float getCameraLookAtDistance() const { return cameraDistance; }
	/// Set camera 3d position to (x, y, z). To leave camera lookat position untouched set updateLookat as false.
	void setCameraPosition( float x, float y, float z, bool updateLookat=true );
	/// Set camera 3d position to pos, To leave camera lookat position untouched set updateLookat as false.
	void setCameraPosition( Vector pos, bool updateLookat=true );
	/// Set camera 3d lookat position to pos. To leave camera eye position position untouched set updateLookat as false.
	void setCameraLookAt( Vector pos, bool updatePosition=true );
	/// Set camera z (altitude)
	void setCameraAltitude( float z );
	/// Get camera position
	Vector getCameraPosition() { return cameraPosition; }
	/// Get camera lookat position
	Vector getCameraLookAt() { return lookAt; }
	/// Move camera towards lookat. forward=distance to move
	void moveCameraForward( float forward );
	/// Move camera in x,y,z-space. To leave camera lookat position untouched set updateLookat as false.
	void moveCameraXYZ( float x, float y, float z=0, bool updateLookat=true );
	/// Move camera right and left (strafe). x=distance to move
	void moveCamera2DRight( float x );
	/// Move camera up and down. y=distance to move
	void moveCamera2DUp( float y );
	/// Set angle. angle=rotation around the z-axis, pitch=rotation around the x-axis
	void setCameraAngle( float angle, float pitch );
	/// Adjust camera angle. angle=rotation around the z-axis, pitch=rotation around the x-axis
	void rotateCameraAngle( float angle, float pitch );
	/// Apply camera trasformations. Variable prevPosWeight can be float between 0.0 and 1.0 which determines how fast the actual camera parameters changes (cool smooth movement effect).
	void applyCamera( float prevPosWeight=0.0 );
	/// Returns true if camera has changed/moved by last applyCamera call
	bool cameraMoved() { return camera_moved; }
	/// Variables angle and pitch are replaced with camera angle and pitch values.
	void getCameraAngle( float &angle, float &pitch );
	

	//other
	/// 2d to 3d transoform, if z=-1, the depth value is read from depth buffer
	Vector screenToWorld( int x, int y, float z=-1 );
	/// 2d to 3d transoform
	Vector screenToWorld( float x, float y, float z=-1 ) { return screenToWorld( int(x), int(y), z ); }
	/// 3d to 2d transoform, return objpos in window coordinates
	Vector worldToScreen( Vector objpos );
	/// Check GL/SDL errors. Returns true if error occurred. str=name of error source or NULL
	bool checkError( const char *str=NULL );
	/// Returns zero if spere with specified radius and position is visible in current frustum. If value is non-zero, return value tells distance from near clipping plane to pos (useful with LOD).
	float isVisible( Vector pos, float radius );
	
	//tgl objects
	/// Load and return newly created TGL object. All loaded objects must be deleted with freeTGLObject. Every model is loaded only once and after first call it returns already loaded model pointer. Objects can be scaled at loading time by pre_scale factor.
	tgl_object* loadTGLObject( const char *fname, float pre_scale=1.0 );
	tgl_object* loadTGLObject( string fname, float pre_scale=1.0 ) { return loadTGLObject( fname.c_str(), pre_scale ); }
	/// Returns existing TGL object
	tgl_object* getTGLObject( const char *fname );
	tgl_object* getTGLObject( string fname ) { return getTGLObject( fname.c_str() ); }
	/// Unload TGL object. This unloads object only if it doesn't have any clients. Any rendering method should NOT delete any TGL objects.
	void freeTGLObject( const char *fname );
	/// Unload TGL object. This unloads object only if it doesn't have any clients. Any rendering method should NOT delete any TGL objects.
	void freeTGLObject( string fname ) { return freeTGLObject( fname.c_str() ); }
	/// Unload TGL object according to pointer. This unloads object only if it doesn't have any clients. Any rendering method should NOT delete any TGL objects.
	void freeTGLObject( tgl_object* obj );

	//textures
	/// Load and return Texture for texture file. AlphaModes are BLEND_ALPHA, BLEND_SOLID, BLEND_RGB ((r+g+b)/3), BLEND_RGBDIST (sqrt(r2+g2+b2)) or BLEND_STENCIL (1-bit alpha, pure white color is transparent). Currently supports PNG-files. freeTexture should be called when texture is not used any more.
	Texture *loadTexture( const char *texturefile, enum AlphaModes alpha=BLEND_ALPHA, bool mipmap=false, int minfilter=GL_NEAREST, int magfilter=GL_NEAREST, int wrapmode=GL_CLAMP );
	Texture *loadTexture( string texturefile, enum AlphaModes alpha=BLEND_ALPHA, bool mipmap=false, int minfilter=GL_NEAREST, int magfilter=GL_NEAREST, int wrapmode=GL_CLAMP ) {
		return loadTexture( texturefile.c_str(), alpha, mipmap, minfilter,magfilter, wrapmode );
	}
	/// Return GL texture ID for texture file, if error occurred return value is 0. Note: You have to load textures separately. Parameter warn is only used by loadTexture.
	GLint getTextureID( const char *texturefile, bool warn=true );
	GLint getTextureID( string texturefile, bool warn=true ) { return getTextureID( texturefile.c_str() ); }
	/// Retrurn a pointer to Texture class for specified file, Returns NULL if texture not loaded. Parameter warn is only used by loadTexture.
	Texture *getTexture( const char *texturefile, bool warn=true );
	Texture *getTexture( string texturefile, bool warn=true ) { return getTexture( texturefile.c_str() ); }
	/// Attempt to delete texture. Texture remains in memory unit it has zero clients. Any rendering method should NOT delete textures.
	void freeTexture( string texturefile );
	/// Attempt to delete texture. Texture remains in memory unit it has zero clients. Any rendering method should NOT delete textures.
	void freeTexture( Texture *texture ) { freeTexture( texture->getFilename() ); }
	
	//text&fonts
	/// Set font. Font must be set before other text methods are usable.
	void setFont( BFont *font ) { currentfont=font; };
	/// Render str at position (x,y) with custom font.
	void renderText( int x, int y, const char *str );
	void renderText( float x, float y, const char *str ) { renderText( int(x), int(y), str ); }
	/// Return a lenghth in pixels of a string with curren font
	float textLength( const char *str );
	
	/// Update sun&lights.
	void applyLights();
	Vector sunPos;
private:
	/// Set camera matrices
	void set3DModelView();
	/// Set GL blend function
	void setBlendFunc( enum BlendFunc func );
	/// Extracts view frusrtum, called in applyCamera
	void extractFrustum();

	//some variables
	typedef void (*GL_ActiveTextureARB_Func)(GLenum);
	typedef void (*GL_MultiTexCoord2fARB_Func)(GLenum, GLfloat, GLfloat);
	GL_ActiveTextureARB_Func glActiveTextureARB_ptr;
	GL_MultiTexCoord2fARB_Func glMultiTexCoord2fARB_ptr;
	Color fogColor, sunDiffuse, sunAmbient;
	enum RenderingModes currentRenderingMode; //current mode
	enum BlendFunc currentBlendFunc;
	Texture *mouseCursor;
	
	// screen
	int screen_h, screen_w, screen_depth;
	bool fullscreen;
	SDL_Surface* screen;
	Game *game; //parent

	//gl
	bool multitexturing, wireframe, render_lines;

	//fog	
	float fog_density, fog_start, fog_end;
	int fog_type;
	
	//camera
	bool camera_moved, camera_changed;
	float cameraAngle, cameraPitch, cameraDistance;
	Vector lookAt; // cam parameters
	Vector lookUp;
	Vector cameraPosition;
	Vector smoothLookAt, smoothLookUp, smoothCameraPosition; // camera parameters with smooth movement
	Vector lastCameraPosition, lastLookUp, lastLookAt; // used to determine if camera has changed
	float maxViewDistance;
	enum CameraModes cameraMode;
	float frustum[6][4];

	//tgl objects
	tgl_object *objects[MAX_TGL_OBJECTS];
	int object_clients[MAX_TGL_OBJECTS]; //determines how many units are using each model
	char object_file_names[MAX_TGL_OBJECTS][MAX_FILE_NAME_LEN];
	
	//textures
	Texture *textures[MAX_TEXTURES];
	
	//particlemode
	enum BlendFunc particle_blend;
	int particle_amount; // -1=0% (disabled), 0,1=100%, 2=50%, 3=33%, 4=25%, 5=20%, ..., 1/n
	int particle_minvis, particlecounter;
	vector<ParticleJob> particle_jobs;
	vector<ParticleJob> particle_effects;
	bool particle_sorting, particle_depthbuffer, particle_post_rendering;

	//font
	BFont *currentfont;
};




#endif
