/***************************************************************************
                          unit.h  -  description
                             -------------------
    begin                : Wed Jul 25 2001
    copyright            : (C) 2001 by T Kontkanen
    email                : tkontkanen@mail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef UNIT_H
#define UNIT_H

class Game;
class Unit;
class UnitID;

#include "game.h"
#include "renderer.h"
//#include "player.h"
//#include "ground.h"
#include "allincludes.h"
#include "gamedefs.h"
#include "vector.h"
#include "unitcommand.h"
#include "unitid.h"

/**Base class of all units
  *@author T Kontkanen
  */


enum UnitTypes { STATIC, BUILDING, BULLET, GROUND_UNIT, AIR_UNIT };

class Unit {
public:
	Unit( Game *pgame );
	virtual ~Unit();

	//controls
	/// Give orders for unit,
//	virtual int action( int actionnum, Vector destpos )=0;
	virtual int action( UnitCommand action );
	/// set primary action (right-click)
//	virtual void setPrimaryAction( int actionnum ) { primaryActionNum=actionnum; }
	virtual void setPrimaryAction( UnitCommand actiontype ) { primaryActionType=actiontype; }
	/// execute primary action
//	virtual int primaryAction( Vector destpos ) { action( primaryActionNum, destpos ); }
	virtual int primaryAction( UnitCommand target ) {
		target.setActionID( primaryActionType.getActionID() ); // set primary action type
		action( target );
	}
	/// set position to pos, implementation required in all units
	virtual void setUnitPosition( Vector pos )=0;
	/// get unit position, implementation required in all units
	virtual Vector getUnitPosition( void )=0;
	/// Return true if unit is selected
	bool isSelected() { return bIsSelected; }
	/// Select if unit is in selection area, sel1 and sel2 are 2d coordinates of selection rectangle corners
	bool screenSelect( Vector sel1, Vector sel2 );
	/// Returns a number of possible actions of this unit
	virtual int getNumberOfActions();
	/// Returns a action type where "number" is something between 0 and getNumberOfActions().
	virtual UnitCommand getActionType( int number );
	
	/// Get enemy to which this Unit is attacking.
	UnitID getTarget();
//	Unit *getTarget();
	/// Set enemy for this Unit. This enemy will be primary target.
//	void setTarget( Unit *enemy );
	/// Set enemy for this Unit. This enemy will be primary target.
//	void setTarget( int player, int unitnum );
	void setTarget( UnitID enemy );
	/// Returns true if unit is moving
	virtual bool isMoving() { return false; }
	
	/// Returns type of unit
	enum UnitTypes getUnitType() { return unittype; }
	/// Returns name of unit. For e.g. "Marine", "Command Center" etc.
	string getName() { return name; }
	/// Returns player number
	int getPlayer() { return myID.getPlayernum(); }
	/// Retuns a maximum distance between unit origin and surface.
	float getMaxSize() { return maxSize; }
	/// Returns an average distance between unit origin and surface.
	float getSize() { return avgSize; }
	
	/// Return number of hitpoints
	float getHitpoints() { return hitpoints; }
	/// Return number of maximum hitpoints
	float getMaxHitpoints() { return maxhitpoints; }
	/// Set amount of hitpoints. Percent if float between 0.0 and 1.0 where 1.0 is maximum amount of hitpoints and 0.0 kills unit.
	void setHitpoints( float percent ) { hitpoints=percent*maxhitpoints; }
	/// Reduce amount of hitpoints by "amount". Weapons must define variable hitpos when soohting at this unit.
	virtual void decreaseHitpoints( float amount, Vector hitposition ) { hitpoints-=amount; }
	/// Increase amount of hitpoints by "amount"
	virtual void increaseHitpoints( float amount ) { hitpoints+=amount; }
	/// Retruns true if unit is not dead
	virtual bool isAlive() { if( hitpoints>0 ) return true; else return false; }

	/// Returns UnitID of this Unit.
	UnitID getID() const { return myID; }
	/// Sets player side to specified players
	void setUnitSide( int playernum ) { mySide=playernum; }
	/// Returns player side (player number)
	int getUnitSide() { return mySide; }
	/// Returns our target unit.
	UnitID unitShootingAt() { return myShootingAt; }
	/// Set average attack damage per second.
	void setShootingAt( UnitID target ) { myShootingAt=target; }
	/// Returns our average attack damage per second.
	float getAverageDamage() { return avgDamage; }
	/// Returns a per Unit random number. Each Unit has it's own random number which can should be used istead of game->rand(). Note: this returns every time same value!
	int myRand() { return myRandnum; }
	/// Returns true if current job (UnitCommand, action)  is finished.
	virtual bool jobFinished() { return true; }
	
	/// Render this unit, implementation required in all units
	virtual int render()=0;
	/// make a move (called once per frame)
	virtual int step()=0;

protected:
	/// Set amount of maximum hitpoints.
	void setMaxHitpoints( float maxpoints ) { maxhitpoints=maxpoints; }
	/// Set unit type. Can be STATIC, BUILDING, BULLET, GROUND_UNIT or AIR_UNIT.
	void setUnitType( enum UnitTypes type ) { unittype=type; }
	/// Replaces current UnitID with new one. Only instertUnit is allowed to use this.
	void setUnitID( UnitID newid ) { myID=newid; }
	/// Set unit's name.
	void setName( string unitname ) { name=unitname; }
	/// Set average attack damage per second.
	void setAverageDamage( float damage ) { avgDamage=damage; }
	/// Set maximum size. See getMaxSIze().
	void setMaxSize( float size ) { maxSize=size; }
	/// Set average size. See getSIze().
	void setSize( float size ) { avgSize=size; }
	
	UnitID myID;
	float maxhitpoints, hitpoints;
private:
	//needed information
	Renderer *uRenderer;
	Game *uGame;
	Ground *uGround;
	Player **uPlayers; //the other players including this
	
	// unit stuff
	int myPlayernum, mySide;
	int primaryActionNum;
	bool bIsSelected;
	UnitID target;
	enum UnitTypes unittype;
	UnitCommand primaryActionType;
	string name;
	int myRandnum;
	float avgDamage;
	UnitID myShootingAt;
	float maxSize, avgSize;
	
	friend Player;
};

#endif
