// Demeter Terrain Visualization Library by Clay Fowler
// Copyright (C) 2001 Clay Fowler

/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.
*/

#include "Terrain.h"
#include "BitArray.h"
#include <iostream>
#ifdef _USE_SDL_
	#include "SDL_image.h"
#endif
#ifdef _WIN32
    #include <crtdbg.h>
#endif
#include "assert.h"
#include <string>

#define GL_CLAMP_TO_EDGE_EXT                0x812F
#define GL_TEXTURE0_ARB                     0x84C0
#define GL_TEXTURE1_ARB                     0x84C1
#define GL_COMBINE_RGB_EXT                  0x8571
#define GL_ARRAY_ELEMENT_LOCK_FIRST_EXT     0x81A8
#define GL_ARRAY_ELEMENT_LOCK_COUNT_EXT     0x81A9
#define COMPRESSED_RGB_S3TC_DXT1_EXT        0x83F0

PFNGLMULTITEXCOORD2FARBPROC glMultiTexCoord2fARB_ptr = NULL;
PFNGLACTIVETEXTUREARBPROC glActiveTextureARB_ptr = NULL;
#ifdef _USE_VERTEX_ARRAYS_
    PFNGLLOCKARRAYSEXTPROC glLockArraysEXT_ptr = NULL;
    PFNGLUNLOCKARRAYSEXTPROC glUnlockArraysEXT_ptr = NULL;
#endif

using namespace Demeter;

Settings* pSettingsInstance = NULL; // Singleton instance of the global settings

const int MAX_FILENAME_LENGTH = 2048;

float numBlocks = 0.0f;
float numLevels = 0.0f;
float hashDelta = 0.0f;
bool bMultiTextureSupported = true;

GLuint CreateTexture(Uint8* pTexels,int width,int height,int rowLength,int border,int internalFormat,bool bClamp,bool bColorKey = false);
#ifdef _USE_SDL_
void LoadImage(char* szFilename,int* pWidth,int* pHeight,Uint8** pBuffer,bool bColorKey = false);
#endif
int RayPlaneIntersect(Ray *ray,Plane *plane,Vector* point,float *distance);
int RayBoxIntersect(Ray *ray,Box *box,Vector *point,float *distance);
bool IsPowerOf2(double number);
void DimensionPowerOf2(int origX,int origY,int& newX,int& newY);

TerrainBlock::TerrainBlock(TerrainBlock* pParent)
{
    m_pChildren = NULL;
    m_pParent = pParent;
    m_pTriangleStrip = NULL;
#ifdef _USE_RAYTRACING_SUPPORT_
    m_pTriangles = NULL;
#endif
}

TerrainBlock::TerrainBlock(int homeVertex,int stride,Terrain* pTerrain,TerrainBlock* pParent)
{
    m_pTriangleStrip = NULL;
    m_pParent = pParent;
    m_HomeIndex = homeVertex;
    m_Stride = stride;
    static int numBlocksBuilt = 0;

    if (Settings::GetInstance()->IsVerbose())
    {
        if (hashDelta <= numBlocksBuilt++)
        {
            cout << "#" << flush;
            numBlocksBuilt = 0;
        }
    }
// Find this block's bounding box.
    float minElevation = pTerrain->GetElevation(homeVertex);
    float maxElevation = pTerrain->GetElevation(homeVertex);
    for (int i = homeVertex; i <= homeVertex + pTerrain->GetWidthVertices() * m_Stride; i += pTerrain->GetWidthVertices())
    {
        for (int j = i; j < i + m_Stride; j++)
        {
            float elevation = pTerrain->GetElevation(j);
            if (elevation < minElevation)
                minElevation = elevation;
            if (maxElevation < elevation)
                maxElevation = elevation;
        }
    }
    Vector boxUpperLeft,boxLowerRight,boxCenter;
    boxUpperLeft.x = pTerrain->m_pVertices[homeVertex].x + pTerrain->m_OffsetX;
    boxUpperLeft.y = pTerrain->m_pVertices[homeVertex].y + pTerrain->m_OffsetY;
    boxUpperLeft.z = maxElevation;
    boxLowerRight.x = pTerrain->m_pVertices[homeVertex + m_Stride * pTerrain->GetWidthVertices() + m_Stride].x + pTerrain->m_OffsetX;
    boxLowerRight.y = pTerrain->m_pVertices[homeVertex + m_Stride * pTerrain->GetWidthVertices() + m_Stride].y + pTerrain->m_OffsetY;
    boxLowerRight.z = minElevation;
    boxCenter.x = (boxUpperLeft.x + boxLowerRight.x) * 0.5f;
    boxCenter.y = (boxUpperLeft.y + boxLowerRight.y) * 0.5f;
    boxCenter.z = (boxUpperLeft.z + boxLowerRight.z) * 0.5f;
    m_BoundingBoxUpperCenter = boxCenter;
    m_BoundingBoxLowerCenter = boxCenter;
    m_BoundingBoxUpperCenter.z = boxUpperLeft.z;
    m_BoundingBoxLowerCenter.z = boxLowerRight.z;

// Recursively build children blocks of this block.
    if (2 < m_Stride)
    {
        m_pChildren = new TerrainBlock*[4];
        int childrenStride = m_Stride / 2;
        m_pChildren[0] = new TerrainBlock(homeVertex,                                                                   childrenStride,pTerrain,this);
        m_pChildren[1] = new TerrainBlock(homeVertex + childrenStride,                                                  childrenStride,pTerrain,this);
        m_pChildren[2] = new TerrainBlock(homeVertex + childrenStride * pTerrain->GetWidthVertices() + childrenStride,  childrenStride,pTerrain,this);
        m_pChildren[3] = new TerrainBlock(homeVertex + childrenStride * pTerrain->GetWidthVertices(),                   childrenStride,pTerrain,this);
    }
    CalculateGeometry(pTerrain);
}

TerrainBlock::~TerrainBlock()
{
    m_pParent = NULL;
    m_pTriangleStrip = NULL;
    if (m_pChildren != NULL && 2 < m_Stride)
    {
        for (int i = 0; i < 4; i++)
        {
            delete m_pChildren[i];
            m_pChildren[i] = NULL;
        }
        delete[] m_pChildren;
    }
}

bool TerrainBlock::IsActive()
{
    return m_bIsActive;
}

void TerrainBlock::Tessellate(double* pMatModelView,double* pMatProjection,int* pViewport,TriangleStrip* pTriangleStrips,TriangleFan* pTriangleFans,int* pCountStrips,int* pCountFans,Terrain* pTerrain)
{
    if ((*pCountStrips < pTerrain->m_MaxNumberOfPrimitives) && pTerrain->CubeInFrustum(m_BoundingBoxUpperCenter.x,m_BoundingBoxUpperCenter.y,m_CubeCenterZ,m_CubeSize))
    {
        if (m_Stride == 2)
        {
            int offset;

            pTerrain->SetVertexStatus(m_HomeIndex,1);
            pTriangleStrips[*pCountStrips].m_pVertices[0] = m_HomeIndex;
            offset = m_HomeIndex + pTerrain->GetWidthVertices();
            pTerrain->SetVertexStatus(offset,1);
            pTriangleStrips[*pCountStrips].m_pVertices[1] = offset;
            offset = m_HomeIndex + 1;
            pTerrain->SetVertexStatus(offset,1);
            pTriangleStrips[*pCountStrips].m_pVertices[2] = offset;
            offset = m_HomeIndex + 1 + pTerrain->GetWidthVertices();
            pTerrain->SetVertexStatus(offset,1);
            pTriangleStrips[*pCountStrips].m_pVertices[3] = offset;
            offset = m_HomeIndex + 2;
            pTerrain->SetVertexStatus(offset,1);
            pTriangleStrips[*pCountStrips].m_pVertices[4] = offset;
            offset = m_HomeIndex + 2 + pTerrain->GetWidthVertices();
            pTerrain->SetVertexStatus(offset,1);
            pTriangleStrips[*pCountStrips].m_pVertices[5] = offset;
            pTriangleStrips[*pCountStrips].m_NumberOfVertices = 6;
            pTriangleStrips[*pCountStrips].m_bEnabled = true;
            *pCountStrips = *pCountStrips + 1;

            if (*pCountStrips < pTerrain->m_MaxNumberOfPrimitives)
            {
                offset = pTerrain->GetWidthVertices() + m_HomeIndex;
                pTerrain->SetVertexStatus(offset,1);
                pTriangleStrips[*pCountStrips].m_pVertices[0] = offset;
                offset = pTerrain->GetWidthVertices() + m_HomeIndex + pTerrain->GetWidthVertices();
                pTerrain->SetVertexStatus(offset,1);
                pTriangleStrips[*pCountStrips].m_pVertices[1] = offset;
                offset = pTerrain->GetWidthVertices() + m_HomeIndex + 1;
                pTerrain->SetVertexStatus(offset,1);
                pTriangleStrips[*pCountStrips].m_pVertices[2] = offset;
                offset = pTerrain->GetWidthVertices() + m_HomeIndex + 1 + pTerrain->GetWidthVertices();
                pTerrain->SetVertexStatus(offset,1);
                pTriangleStrips[*pCountStrips].m_pVertices[3] = offset;
                offset = pTerrain->GetWidthVertices() + m_HomeIndex + 2;
                pTerrain->SetVertexStatus(offset,1);
                pTriangleStrips[*pCountStrips].m_pVertices[4] = offset;
                offset = pTerrain->GetWidthVertices() + m_HomeIndex + 2 + pTerrain->GetWidthVertices();
                pTerrain->SetVertexStatus(offset,1);
                pTriangleStrips[*pCountStrips].m_pVertices[5] = offset;
                pTriangleStrips[*pCountStrips].m_NumberOfVertices = 6;
                pTriangleStrips[*pCountStrips].m_bEnabled = true;
                *pCountStrips = *pCountStrips + 1;
            }

            m_bIsActive = true;
        }
        else
        {
            if (pTerrain->m_MaximumVisibleBlockSize < m_Stride)
            {
                m_pChildren[0]->Tessellate(pMatModelView,pMatProjection,pViewport,pTriangleStrips,pTriangleFans,pCountStrips,pCountFans,pTerrain);
                m_pChildren[1]->Tessellate(pMatModelView,pMatProjection,pViewport,pTriangleStrips,pTriangleFans,pCountStrips,pCountFans,pTerrain);
                m_pChildren[2]->Tessellate(pMatModelView,pMatProjection,pViewport,pTriangleStrips,pTriangleFans,pCountStrips,pCountFans,pTerrain);
                m_pChildren[3]->Tessellate(pMatModelView,pMatProjection,pViewport,pTriangleStrips,pTriangleFans,pCountStrips,pCountFans,pTerrain);
                m_bIsActive = false;
                m_bChildrenActive = true;
            }
            else
            {
                double screenTopX,screenTopY,screenTopZ,screenBottomX,screenBottomY,screenBottomZ;
                float halfWidth = m_CubeSize * 0.5f;
                // Check the vertical screen size of the block
                gluProject(m_BoundingBoxUpperCenter.x,m_BoundingBoxUpperCenter.y,m_BoundingBoxUpperCenter.z,pMatModelView,pMatProjection,pViewport,&screenTopX,&screenTopY,&screenTopZ);
                gluProject(m_BoundingBoxLowerCenter.x,m_BoundingBoxLowerCenter.y,m_BoundingBoxLowerCenter.z,pMatModelView,pMatProjection,pViewport,&screenBottomX,&screenBottomY,&screenBottomZ);
                float screenDistVertical = fabs(screenTopY - screenBottomY);
                // Use both the X and Y axes to find the horizontal screen size of the block by using the larger of the two.
                gluProject(m_BoundingBoxUpperCenter.x - halfWidth,m_BoundingBoxUpperCenter.y,m_BoundingBoxUpperCenter.z,pMatModelView,pMatProjection,pViewport,&screenTopX,&screenTopY,&screenTopZ);
                gluProject(m_BoundingBoxLowerCenter.x + halfWidth,m_BoundingBoxLowerCenter.y,m_BoundingBoxLowerCenter.z,pMatModelView,pMatProjection,pViewport,&screenBottomX,&screenBottomY,&screenBottomZ);
                float screenDistX = fabs(screenTopX - screenBottomX);
                gluProject(m_BoundingBoxUpperCenter.x,m_BoundingBoxUpperCenter.y - halfWidth,m_BoundingBoxUpperCenter.z,pMatModelView,pMatProjection,pViewport,&screenTopX,&screenTopY,&screenTopZ);
                gluProject(m_BoundingBoxLowerCenter.x,m_BoundingBoxLowerCenter.y + halfWidth,m_BoundingBoxLowerCenter.z,pMatModelView,pMatProjection,pViewport,&screenBottomX,&screenBottomY,&screenBottomZ);
                float screenDistY = fabs(screenTopX - screenBottomX);
                float screenDistHorizontal = screenDistX < screenDistY ? screenDistY : screenDistX;

                // Use the smaller of vertical and horizontal screen size to decide whether or not the block should be simplified.
                float screenDist = screenDistHorizontal < screenDistVertical ? screenDistHorizontal : screenDistVertical;

                if (screenDist <= pTerrain->GetDetailThreshold())
                {
                // This block is simplified, so add its triangles to the list and stop recursing.
                    CreateTriangleStrip(pTriangleStrips,pCountStrips,pTerrain);
                    m_bIsActive = true;
                    m_bChildrenActive = false;
                }
                else
                {
                    m_pChildren[0]->Tessellate(pMatModelView,pMatProjection,pViewport,pTriangleStrips,pTriangleFans,pCountStrips,pCountFans,pTerrain);
                    m_pChildren[1]->Tessellate(pMatModelView,pMatProjection,pViewport,pTriangleStrips,pTriangleFans,pCountStrips,pCountFans,pTerrain);
                    m_pChildren[2]->Tessellate(pMatModelView,pMatProjection,pViewport,pTriangleStrips,pTriangleFans,pCountStrips,pCountFans,pTerrain);
                    m_pChildren[3]->Tessellate(pMatModelView,pMatProjection,pViewport,pTriangleStrips,pTriangleFans,pCountStrips,pCountFans,pTerrain);
                    m_bIsActive = false;
                    m_bChildrenActive = true;
                }
            }
        }
    }
    else
    {
        m_bIsActive = false;
        m_bChildrenActive = false;
    }
}

void TerrainBlock::CreateTriangleStrip(TriangleStrip* pTriangleStrips,int* pCount,Terrain* pTerrain)
{
    if (*pCount < pTerrain->m_MaxNumberOfPrimitives)
    {
        pTerrain->SetVertexStatus(m_HomeIndex,1);
        pTriangleStrips[*pCount].m_pVertices[0] = m_HomeIndex;
        int offset = m_HomeIndex + pTerrain->GetWidthVertices() * m_Stride;
        pTerrain->SetVertexStatus(offset,1);
        pTriangleStrips[*pCount].m_pVertices[1] = offset;
        offset = m_HomeIndex + m_Stride;
        pTerrain->SetVertexStatus(offset,1);
        pTriangleStrips[*pCount].m_pVertices[2] = offset;
        offset = m_HomeIndex + m_Stride + pTerrain->GetWidthVertices() * m_Stride;
        pTerrain->SetVertexStatus(offset,1);
        pTriangleStrips[*pCount].m_pVertices[3] = offset;
        pTriangleStrips[*pCount].m_bEnabled = true;
        pTriangleStrips[*pCount].m_NumberOfVertices = 4;

        m_pTriangleStrip = &pTriangleStrips[*pCount];
        *pCount = *pCount + 1;
    }
}

void TerrainBlock::EnableStrip(bool bEnabled)
{
    m_pTriangleStrip->m_bEnabled = false;
}

int TerrainBlock::GetStride()
{
    return m_Stride;
}

TerrainBlock* TerrainBlock::GetParent()
{
    return m_pParent;
}

int TerrainBlock::GetHomeIndex()
{
    return m_HomeIndex;
}

void TerrainBlock::RepairCracks(Terrain* pTerrain,TriangleStrip* pTriangleStrips,TriangleFan* pTriangleFans,int* pCountStrips,int* pCountFans)
{
    if (2 < m_Stride)
    {
        if (m_bIsActive)
        {
            int halfStride = m_Stride / 2;
            int bottomLeft = m_HomeIndex + m_Stride * pTerrain->GetWidthVertices();
            int bottomRight = bottomLeft + m_Stride;
            int i,previousVertex;
            int v0;
            int numVertices = 0;

            bool bNeedToFix = false;
            for (i = m_HomeIndex + m_Stride - 1; m_HomeIndex < i && !bNeedToFix; i--)
                bNeedToFix = (pTerrain->GetVertexStatus(i) == 1);
            if (!bNeedToFix)
            {
                for (i = m_HomeIndex + pTerrain->GetWidthVertices(); i < m_HomeIndex + m_Stride * pTerrain->GetWidthVertices() && !bNeedToFix; i += pTerrain->GetWidthVertices())
                    bNeedToFix = (pTerrain->GetVertexStatus(i) == 1);
                if (!bNeedToFix)
                {
                    for (i = bottomLeft + 1; i < bottomRight && !bNeedToFix; i++)
                        bNeedToFix = (pTerrain->GetVertexStatus(i) == 1);
                    if (!bNeedToFix)
                    {
                        for (i = bottomRight - pTerrain->GetWidthVertices(); m_HomeIndex + m_Stride < i && !bNeedToFix; i -= pTerrain->GetWidthVertices())
                            bNeedToFix = (pTerrain->GetVertexStatus(i) == 1);
                    }
                }
            }

            if (bNeedToFix)
            {
                EnableStrip(false);
                v0 = m_HomeIndex + halfStride + halfStride * pTerrain->GetWidthVertices();
                assert(0 <= v0);
                assert(v0 < pTerrain->GetNumberOfVertices());
                pTriangleFans[*pCountFans].m_pVertices[0] = v0;
                numVertices = 0;
                for (i = m_HomeIndex + m_Stride; m_HomeIndex <= i; i--)
                {
                    assert(0 <= i);
                    assert(i < pTerrain->GetNumberOfVertices());
                    if (pTerrain->GetVertexStatus(i) == 1)
                    {
                        if (++numVertices == MAX_VERTICES_PER_FAN - 1)
                        {
                        // We have reached the maximum size for a fan, so start a new fan.
                            pTriangleFans[*pCountFans].m_NumberOfVertices = numVertices;
                            *pCountFans = *pCountFans + 1;
                            pTriangleFans[*pCountFans].m_pVertices[0] = v0;
                            pTriangleFans[*pCountFans].m_pVertices[1] = previousVertex;
                            numVertices = 2;
                        }
                        pTriangleFans[*pCountFans].m_pVertices[numVertices] = i;
                        previousVertex = i;
                    }
                }
                for (i = m_HomeIndex + pTerrain->GetWidthVertices(); i <= m_HomeIndex + m_Stride * pTerrain->GetWidthVertices(); i += pTerrain->GetWidthVertices())
                {
                    assert(0 <= i);
                    assert(i < pTerrain->GetNumberOfVertices());
                    if(pTerrain->GetVertexStatus(i) == 1)
                    {
                        if (++numVertices == MAX_VERTICES_PER_FAN - 1)
                        {
                        // We have reached the maximum size for a fan, so start a new fan.
                            pTriangleFans[*pCountFans].m_NumberOfVertices = numVertices;
                            *pCountFans = *pCountFans + 1;
                            pTriangleFans[*pCountFans].m_pVertices[0] = v0;
                            pTriangleFans[*pCountFans].m_pVertices[1] = previousVertex;
                            numVertices = 2;
                        }
                        pTriangleFans[*pCountFans].m_pVertices[numVertices] = i;
                        previousVertex = i;
                    }
                }
                for (i = bottomLeft; i <= bottomRight; i++)
                {
                    assert(0 <= i);
                    assert(i < pTerrain->GetNumberOfVertices());
                    if (pTerrain->GetVertexStatus(i) == 1)
                    {
                        if (++numVertices == MAX_VERTICES_PER_FAN - 1)
                        {
                        // We have reached the maximum size for a fan, so start a new fan.
                            pTriangleFans[*pCountFans].m_NumberOfVertices = numVertices;
                            *pCountFans = *pCountFans + 1;
                            pTriangleFans[*pCountFans].m_pVertices[0] = v0;
                            pTriangleFans[*pCountFans].m_pVertices[1] = previousVertex;
                            numVertices = 2;
                        }
                        pTriangleFans[*pCountFans].m_pVertices[numVertices] = i;
                        previousVertex = i;
                    }
                }
                for (i = bottomRight - pTerrain->GetWidthVertices(); m_HomeIndex + m_Stride <= i; i -= pTerrain->GetWidthVertices())
                {
                    assert(0 <= i);
                    assert(i < pTerrain->GetNumberOfVertices());
                    if(pTerrain->GetVertexStatus(i) == 1)
                    {
                        if (++numVertices == MAX_VERTICES_PER_FAN - 1)
                        {
                        // We have reached the maximum size for a fan, so start a new fan.
                            pTriangleFans[*pCountFans].m_NumberOfVertices = numVertices;
                            *pCountFans = *pCountFans + 1;
                            pTriangleFans[*pCountFans].m_pVertices[0] = v0;
                            pTriangleFans[*pCountFans].m_pVertices[1] = previousVertex;
                            numVertices = 2;
                        }
                        pTriangleFans[*pCountFans].m_pVertices[numVertices] = i;
                        previousVertex = i;
                    }
                }
                pTriangleFans[*pCountFans].m_NumberOfVertices = numVertices + 1;
                *pCountFans = *pCountFans + 1;
            }
        }
        else if (m_bChildrenActive)
        {
            m_pChildren[0]->RepairCracks(pTerrain,pTriangleStrips,pTriangleFans,pCountStrips,pCountFans);
            m_pChildren[1]->RepairCracks(pTerrain,pTriangleStrips,pTriangleFans,pCountStrips,pCountFans);
            m_pChildren[2]->RepairCracks(pTerrain,pTriangleStrips,pTriangleFans,pCountStrips,pCountFans);
            m_pChildren[3]->RepairCracks(pTerrain,pTriangleStrips,pTriangleFans,pCountStrips,pCountFans);
        }
    }
}

void TerrainBlock::CalculateGeometry(Terrain* pTerrain)
{
    float height = m_BoundingBoxUpperCenter.z - m_BoundingBoxLowerCenter.z;
    float width = m_Stride * pTerrain->GetVertexSpacing();
    if (width < height)
        m_CubeSize = height;
    else
        m_CubeSize = width;
    m_CubeCenterZ = 0.5f * (m_BoundingBoxUpperCenter.z + m_BoundingBoxLowerCenter.z);

    float halfCube = m_CubeSize / 2.0f;
    m_BoundingBox.m_Min.x = m_BoundingBoxUpperCenter.x - halfCube;
    m_BoundingBox.m_Min.y = m_BoundingBoxUpperCenter.y - halfCube;
    m_BoundingBox.m_Min.z = m_CubeCenterZ - halfCube;
    m_BoundingBox.m_Max.x = m_BoundingBoxUpperCenter.x + halfCube;
    m_BoundingBox.m_Max.y = m_BoundingBoxUpperCenter.y + halfCube;
    m_BoundingBox.m_Max.z = m_CubeCenterZ + halfCube;

#ifdef _USE_RAYTRACING_SUPPORT_
// Build triangles for ray intersection and collision detection.
    if (m_Stride == 2)
    {
        m_pTriangles = new Triangle[8];
        m_pTriangles[0].DefineFromPoints(pTerrain->m_pVertices[m_HomeIndex],pTerrain->m_pVertices[m_HomeIndex + pTerrain->GetWidthVertices()],pTerrain->m_pVertices[m_HomeIndex + 1]);
        m_pTriangles[1].DefineFromPoints(pTerrain->m_pVertices[m_HomeIndex + 1],pTerrain->m_pVertices[m_HomeIndex + pTerrain->GetWidthVertices()],pTerrain->m_pVertices[m_HomeIndex + pTerrain->GetWidthVertices() + 1]);
        m_pTriangles[2].DefineFromPoints(pTerrain->m_pVertices[m_HomeIndex + 1],pTerrain->m_pVertices[m_HomeIndex + 1 + pTerrain->GetWidthVertices()],pTerrain->m_pVertices[m_HomeIndex + 2]);
        m_pTriangles[3].DefineFromPoints(pTerrain->m_pVertices[m_HomeIndex + 2],pTerrain->m_pVertices[m_HomeIndex + 1 + pTerrain->GetWidthVertices()],pTerrain->m_pVertices[m_HomeIndex + 2 + pTerrain->GetWidthVertices()]);
        m_pTriangles[4].DefineFromPoints(pTerrain->m_pVertices[m_HomeIndex + pTerrain->GetWidthVertices()],pTerrain->m_pVertices[m_HomeIndex + pTerrain->GetWidthVertices() + pTerrain->GetWidthVertices()],pTerrain->m_pVertices[m_HomeIndex  + pTerrain->GetWidthVertices() + 1]);
        m_pTriangles[5].DefineFromPoints(pTerrain->m_pVertices[m_HomeIndex + pTerrain->GetWidthVertices() + 1],pTerrain->m_pVertices[m_HomeIndex + pTerrain->GetWidthVertices() + pTerrain->GetWidthVertices()],pTerrain->m_pVertices[m_HomeIndex  + pTerrain->GetWidthVertices() + pTerrain->GetWidthVertices() + 1]);
        m_pTriangles[6].DefineFromPoints(pTerrain->m_pVertices[m_HomeIndex + pTerrain->GetWidthVertices() + 1],pTerrain->m_pVertices[m_HomeIndex + pTerrain->GetWidthVertices() + 1 + pTerrain->GetWidthVertices()],pTerrain->m_pVertices[m_HomeIndex  + pTerrain->GetWidthVertices() + 2]);
        m_pTriangles[7].DefineFromPoints(pTerrain->m_pVertices[m_HomeIndex + pTerrain->GetWidthVertices() + 2],pTerrain->m_pVertices[m_HomeIndex + pTerrain->GetWidthVertices() + 1 + pTerrain->GetWidthVertices()],pTerrain->m_pVertices[m_HomeIndex  + pTerrain->GetWidthVertices() + 2 + pTerrain->GetWidthVertices()]);
    }
#endif
}

void TerrainBlock::Write(FILE* file)
{
    fwrite(&m_Stride,sizeof(int),1,file);
    fwrite(&m_HomeIndex,sizeof(int),1,file);
    fwrite(&m_BoundingBoxUpperCenter,sizeof(Vector),1,file);
    fwrite(&m_BoundingBoxLowerCenter,sizeof(Vector),1,file);
    if (2 < m_Stride)
    {
        for (int i = 0; i < 4; i++)
            m_pChildren[i]->Write(file);
    }
}

void TerrainBlock::Read(FILE* file,Terrain* pTerrain)
{
    fread(&m_Stride,sizeof(int),1,file);
    fread(&m_HomeIndex,sizeof(int),1,file);
    fread(&m_BoundingBoxUpperCenter,sizeof(Vector),1,file);
    fread(&m_BoundingBoxLowerCenter,sizeof(Vector),1,file);
    if (2 < m_Stride)
    {
        m_pChildren = new TerrainBlock*[4];
        for (int i = 0; i < 4; i++)
        {
            m_pChildren[i] = new TerrainBlock(this);
            m_pChildren[i]->Read(file,pTerrain);
        }
    }
    CalculateGeometry(pTerrain);
}

#ifdef _SUPPORT_OPENSCENEGRAPH_
#define CALL_BASE_CONSTRUCTOR : Referenced()
#else
#define CALL_BASE_CONSTRUCTOR
#endif

#ifdef _USE_SDL_
Terrain::Terrain(float* pElevations,int elevWidth,int elevHeight,Uint8* pTextureImage,int textureWidth,int textureHeight,Uint8* pDetailTextureImage,int detailWidth,int detailHeight,float vertexSpacing,float elevationScale,int maxNumTriangles,bool bUseBorders,float offsetX,float offsetY,int numTexturesX,int numTexturesY) CALL_BASE_CONSTRUCTOR
#else
Terrain::Terrain(float* pElevations,int elevWidth,int elevHeight,Uint8* pTextureImage,int textureWidth,int textureHeight,Uint8* pDetailTextureImage,int detailWidth,int detailHeight,float vertexSpacing,float elevationScale,int maxNumTriangles,bool bUseBorders,PFNGLMULTITEXCOORD2FARBPROC pMultiTexCoordFunc,PFNGLACTIVETEXTUREARBPROC pActiveTexFunc,float offsetX,float offsetY,int numTexturesX,int numTexturesY) CALL_BASE_CONSTRUCTOR
#endif
{
#ifndef _USE_SDL_
	glMultiTexCoord2fARB_ptr = pMultiTexCoordFunc;
	glActiveTextureARB_ptr = pActiveTexFunc;
#endif
	Init(pElevations,elevWidth,elevHeight,pTextureImage,textureWidth,textureHeight,pDetailTextureImage,detailWidth,detailHeight,vertexSpacing,elevationScale,maxNumTriangles,bUseBorders,offsetX,offsetY,numTexturesX,numTexturesY);
}

#ifdef _USE_SDL_
Terrain::Terrain(char* szElevationsFilename,char* szTextureFilename,char* szDetailTextureFilename,float vertexSpacing,float elevationScale,int maxNumTriangles,bool bUseBorders,float offsetX,float offsetY,int numTexturesX,int numTexturesY) CALL_BASE_CONSTRUCTOR
{
#ifdef _USE_GDAL_
	bool useGDAL = (strstr(szElevationsFilename,".ddf") || strstr(szElevationsFilename,".DDF") || strstr(szElevationsFilename,".dem") || strstr(szElevationsFilename,".DEM"));
#endif

	char szFullFilename[MAX_FILENAME_LENGTH];
	Settings::GetInstance()->PrependMediaPath(szElevationsFilename,szFullFilename);
	
	float* pImageData = NULL;
	int elevWidth;
	int elevHeight;

#ifdef _USE_GDAL_
	if(useGDAL) 
	{
		GDALDataset *poDataset;
		GDALAllRegister();
		poDataset = (GDALDataset*)GDALOpen(szFullFilename,GA_ReadOnly);
		if (poDataset == NULL)
			cout << "TERRAIN: Failed to load DEM data" << endl;
		cout << "TERRAIN: Input size is " << poDataset->GetRasterXSize() << "x" << poDataset->GetRasterYSize() << "x" << poDataset->GetRasterCount() << endl;
		int dataWidth = poDataset->GetRasterXSize();
		int dataHeight = poDataset->GetRasterYSize();
		DimensionPowerOf2(poDataset->GetRasterXSize(),poDataset->GetRasterYSize(),elevWidth,elevHeight);
		cout << "TERRAIN: Changing size to " << elevWidth << "," << elevHeight << endl;
		
		GDALRasterBand *poBand;
		int nBlockXSize, nBlockYSize;
		poBand = poDataset->GetRasterBand(1);
		poBand->GetBlockSize(&nBlockXSize,&nBlockYSize);

		pImageData = new float[elevWidth * elevHeight];
		
		int i;
		for (i = 0; i < elevHeight * elevWidth; i++)
			pImageData[i] = 0.0f;

		for(i = 0; i < dataHeight; i++)
		{
			float *pafScanline;
			int nXSize = poBand->GetXSize();
			pafScanline = (float *) CPLMalloc(sizeof(float)*nXSize);
			poBand->RasterIO(GF_Read,0,i,nXSize,1,pafScanline,nXSize,1,GDT_Float32,0,0);
			int index = i * elevWidth;
			for (int j = 0; j < dataWidth; j++)
				pImageData[index + j] = pafScanline[j];
		}

		delete poDataset;
	}
	else
	{
#endif
		SDL_Surface* pImage = IMG_Load(szFullFilename);
		if (pImage == NULL)
		{
			string msg("Failed to load elevation image file '");
			msg += szFullFilename;
			msg += "'; This means that the file was not found or it is not an image type that Demeter can read.";
			throw new DemeterException(msg);
		}
		else
		{
			// Test to see if the image is a power of 2 in both width and height.
			if (!IsPowerOf2((double)pImage->w) || !IsPowerOf2((double)pImage->h))
			{
				string msg("The elevation image file '");
				msg += szFullFilename;
				msg += "' is NOT a power of 2 in both width and height. Elevation files must be a power of 2 in both width and height.";
				throw new DemeterException(msg);
			}

			// Test to see if the image is 24-bit color.
			if (pImage->format->BytesPerPixel < 3)
			{
				string msg("The elevation image file '");
				msg += szFullFilename;
				msg += "' is NOT a 24-bit image. Elevation files must be 24-bit images.";
				throw new DemeterException(msg);
			}

			// Load the elevation data.
			int i,j,k;
			int size = pImage->pitch * pImage->h;
			pImageData = new float[pImage->w * pImage->h];
			for (i = 0,k = 0; i < size; i += pImage->pitch)
			{
				for (j = 0; j < pImage->pitch; j += pImage->format->BytesPerPixel,k++)
				{
					int index = i + j;
					Uint32* pCurrentPixel = (Uint32*)(((Uint8*)pImage->pixels) + index);
					Uint8 red,green,blue;
					SDL_GetRGB(*pCurrentPixel,pImage->format,&red,&green,&blue);
					pImageData[k] = (float)red;
				}
			}
			elevWidth = pImage->w;
			elevHeight = pImage->h;
			SDL_FreeSurface(pImage);
		}
#ifdef _USE_GDAL_
	}
#endif

	// Load the texture data.
	int texWidth,texHeight;
	Uint8* pTextureImage = NULL;
	if (szTextureFilename != NULL)
	{
		LoadImage(szTextureFilename,&texWidth,&texHeight,&pTextureImage,false);
		if (texWidth == 0)
		{
			string msg("Failed to load texture image file '");
			msg += szTextureFilename;
			msg += "'; This means that the file was not found or it is not an image type that Demeter can read.";
			throw new DemeterException(msg);
		}
	}
	
	// Load the detail texture data.
	int detailWidth,detailHeight;
	Uint8* pDetailImage = NULL;
	if (szDetailTextureFilename != NULL)
	{
		LoadImage(szDetailTextureFilename,&detailWidth,&detailHeight,&pDetailImage,false);
		if (detailWidth == 0)
		{
			string msg("Failed to load detail texture image file '");
			msg += szDetailTextureFilename;
			msg += "'; This means that the file was not found or it is not an image type that Demeter can read.";
			throw new DemeterException(msg);
		}
	}

	Init(pImageData,elevWidth,elevHeight,pTextureImage,texWidth,texHeight,pDetailImage,detailWidth,detailHeight,vertexSpacing,elevationScale,maxNumTriangles,bUseBorders,offsetX,offsetY,numTexturesX,numTexturesY);

	delete[] pDetailImage;
	delete[] pTextureImage;
	delete[] pImageData;
}
#endif

Terrain::~Terrain()
{
    while (!m_Textures.empty())
    {
        vector<Texture*>::iterator iter = m_Textures.begin();
        Texture* pTexture = *iter;
        m_Textures.erase(iter);
        delete pTexture;
    }

    if (m_pCommonTexture != NULL)
        delete m_pCommonTexture;

    delete[] m_pTriangleStrips;
    delete[] m_pTriangleFans;
    delete[] m_pVertices;
    delete m_pVertexStatus;
    delete m_pRootBlock;
}

void Terrain::Init(float* pElevations,int elevWidth,int elevHeight,Uint8* pTextureImage,int textureWidth,int textureHeight,Uint8* pDetailTextureImage,int detailWidth,int detailHeight,float vertexSpacing,float elevationScale,int maxNumTriangles,bool bUseBorders,float offsetX,float offsetY,int numTexturesX,int numTexturesY)
{
    m_pTextureFactory = NULL;
    m_MaxNumberOfPrimitives = maxNumTriangles / 4;
    m_pTriangleStrips = NULL;
    m_pTriangleFans = NULL;
    m_pVertices = NULL;
    m_CommonTextureRepeats = 20.0f;
    m_pVertexStatus = NULL;
    m_pRootBlock = NULL;
    m_pCommonTexture = NULL;
    m_VertexSpacing = vertexSpacing;
    m_MaximumVisibleBlockSize = 8;
    m_OffsetX = offsetX;
    m_OffsetY = offsetY;

    Uint8 red;
    Uint8 green;
    Uint8 blue;

	if (!IsPowerOf2(elevWidth) || !IsPowerOf2(elevHeight))
	{
		string msg("The elevation data is NOT a power of 2 in both width and height. Elevation data must be a power of 2 in both width and height.");
		throw new DemeterException(msg);
	}

	m_WidthVertices = elevWidth + 1; // Add 1 dummy pixel line to edge for block strides
	m_HeightVertices = elevHeight + 1;
	m_NumberOfVertices = m_WidthVertices * m_HeightVertices;
	m_pVertices = new Vector[m_WidthVertices * m_HeightVertices];
	int i,j;
	float x,y;
	y = 0.0f;
	m_MaxElevation = 0.0f;
	int end = elevWidth * elevHeight;
	for (i = 0,j = 0; i < end; i += elevWidth,y += m_VertexSpacing)
	{
		float* pImageRow = pElevations + i;
		x = 0.0f;
		for (float* pImagePixel = pImageRow; pImagePixel < pImageRow + elevWidth; pImagePixel++,j++,x += m_VertexSpacing)
		{
			m_pVertices[j].x = x;
			m_pVertices[j].y = y;
			m_pVertices[j].z = *pImagePixel * elevationScale;
			if (m_MaxElevation < m_pVertices[j].z)
				m_MaxElevation = m_pVertices[j].z;
		}
		// Account for dummy column on right edge
		m_pVertices[j].x = (m_WidthVertices - 1) * m_VertexSpacing;
		m_pVertices[j].y = y;
		m_pVertices[j].z = m_pVertices[j - 1].z;
		if (m_MaxElevation < m_pVertices[j].z)
			m_MaxElevation = m_pVertices[j].z;
		j++; 
		x += m_VertexSpacing;
	}

	x = 0.0f;
	for (i = m_NumberOfVertices - m_WidthVertices; i < m_NumberOfVertices; i++,x += m_VertexSpacing)
	{
		m_pVertices[i].x = x;
		m_pVertices[i].y = (m_HeightVertices - 1) * m_VertexSpacing;
		m_pVertices[i].z = m_pVertices[i - m_WidthVertices].z;
	}

    m_DetailThreshold = 9.0f;
    BuildBlocks();
    if (pTextureImage != NULL)
        SetTexture(pTextureImage,textureWidth,textureHeight,bUseBorders);
    if (pDetailTextureImage != NULL)
        SetCommonTexture(pDetailTextureImage,detailWidth,detailHeight);

#ifdef _USE_SDL_
	if (glActiveTextureARB_ptr == NULL && !Settings::GetInstance()->IsHeadless())
	{
		glActiveTextureARB_ptr = (PFNGLACTIVETEXTUREARBPROC)SDL_GL_GetProcAddress("glActiveTextureARB");
		glMultiTexCoord2fARB_ptr = (PFNGLMULTITEXCOORD2FARBPROC)SDL_GL_GetProcAddress("glMultiTexCoord2fARB");
		if(!glActiveTextureARB_ptr || !glMultiTexCoord2fARB_ptr)
		{
			cout << "TERRAIN: ERROR: Multitexture extensions not supported by this OpenGL vendor!" << endl;
			bMultiTextureSupported = false;
		}
	}
#endif

	if (0 < numTexturesX)
	{
		m_TextureTileWidth = (int)(((float)(m_WidthVertices / (float)numTexturesX)) * m_VertexSpacing);
		m_TextureTileHeight = (int)(((float)(m_HeightVertices / (float)numTexturesY)) * m_VertexSpacing);
		m_NumberOfTextureTilesWidth = numTexturesX;
		m_NumberOfTextureTilesHeight = numTexturesY;
		m_NumberOfTextureTiles = numTexturesX * numTexturesY;
		m_TileSize = numTexturesX * m_TextureTileWidth;
	}
}

bool Terrain::IsMultiTextureSupported()
{
    return bMultiTextureSupported;
}

int Terrain::GetNumberOfVertices()
{
    return m_NumberOfVertices;
}

void Terrain::UpdateNeighbor(Terrain* pTerrain,Terrain::DIRECTION direction)
{
    int thisVertex,otherVertex;
    if (direction == Terrain::DIR_SOUTH)
    {
        for (thisVertex = 0,otherVertex = m_NumberOfVertices - m_WidthVertices; thisVertex < m_WidthVertices; thisVertex++,otherVertex++)
        {
            if (GetVertexStatus(thisVertex))
                pTerrain->SetVertexStatus(otherVertex,true);
        }
    }
    else if (direction == Terrain::DIR_NORTH)
    {
        for (thisVertex = m_NumberOfVertices - m_WidthVertices,otherVertex = 0; thisVertex < m_NumberOfVertices; thisVertex++,otherVertex++)
        {
            if (GetVertexStatus(thisVertex))
                pTerrain->SetVertexStatus(otherVertex,true);
        }
    }
    else if (direction == Terrain::DIR_WEST)
    {
        for (thisVertex = 0,otherVertex = m_WidthVertices - 1; thisVertex < m_NumberOfVertices; thisVertex += m_WidthVertices,otherVertex += m_WidthVertices)
            if (GetVertexStatus(thisVertex))
                pTerrain->SetVertexStatus(otherVertex,true);
    }
    else if (direction == Terrain::DIR_EAST)
    {
        for (thisVertex = m_WidthVertices - 1,otherVertex = 0; thisVertex < m_NumberOfVertices; thisVertex += m_WidthVertices,otherVertex += m_WidthVertices)
            if (GetVertexStatus(thisVertex))
                pTerrain->SetVertexStatus(otherVertex,true);
    }
    else if (direction == Terrain::DIR_NORTHWEST)
    {
        if (GetVertexStatus(m_NumberOfVertices - m_WidthVertices))
            pTerrain->SetVertexStatus(m_WidthVertices - 1,true);
    }
    else if (direction == Terrain::DIR_NORTHEAST)
    {
        if (GetVertexStatus(m_NumberOfVertices - 1))
            pTerrain->SetVertexStatus(0,true);
    }
    else if (direction == Terrain::DIR_SOUTHEAST)
    {
        if (GetVertexStatus(m_WidthVertices - 1))
            pTerrain->SetVertexStatus(m_NumberOfVertices - m_WidthVertices,true);
    }
    else if (direction == Terrain::DIR_SOUTHWEST)
    {
        if (GetVertexStatus(0))
            pTerrain->SetVertexStatus(m_NumberOfVertices - 1,true);
    }
}

float Terrain::GetElevation(int index)
{
    return m_pVertices[index].z;
}

float Terrain::GetElevation(float x,float y)
{
    Plane plane;
    int vertexID;
    float elevation;

    x -= m_OffsetX;
    y -= m_OffsetY;

#ifdef _PROTECT_ACCESS_
    if (x < 0.0f || y < 0.0f || GetWidth() < x || GetHeight() < y)
        elevation = 0.0f;
    else
    {
#endif
        vertexID = ((int)(y / m_VertexSpacing)) * m_WidthVertices + ((int)(x / m_VertexSpacing));

        if ((fmod(y,m_VertexSpacing) + fmod(x,m_VertexSpacing)) <= m_VertexSpacing)
            plane.defineFromPoints(m_pVertices[vertexID],m_pVertices[vertexID + m_WidthVertices],m_pVertices[vertexID + 1]);
        else
            plane.defineFromPoints(m_pVertices[vertexID + 1],m_pVertices[vertexID + 1 + m_WidthVertices],m_pVertices[vertexID + m_WidthVertices]);

        elevation = -1.0f * ((plane.a * x + plane.b * y + plane.d) / plane.c);
#ifdef _PROTECT_ACCESS_
    }
#endif

    return elevation;
}

void Terrain::GetNormal(float x,float y,float& normalX,float& normalY,float& normalZ)
{
    Plane plane;
    int vertexID;
#ifdef _PROTECT_ACCESS_
    if (x < 0.0f || y < 0.0f || GetWidth() < x || GetHeight() < y)
    {
        normalX = normalY = 0.0f;
        normalZ = 1.0f;
    }
    else
    {
#endif
        vertexID = (int)(y / m_VertexSpacing) * m_WidthVertices + (int)(x / m_VertexSpacing);

        if ((fmod(y,m_VertexSpacing) - fmod(x,m_VertexSpacing)) >= 0.0f)
            plane.defineFromPoints(m_pVertices[vertexID],m_pVertices[vertexID + 1],m_pVertices[vertexID + m_WidthVertices]);
        else
            plane.defineFromPoints(m_pVertices[vertexID + 1],m_pVertices[vertexID + 1 + m_WidthVertices],m_pVertices[vertexID + m_WidthVertices]);

        normalX = plane.a;
        normalY = plane.b;
        normalZ = plane.c;
#ifdef _PROTECT_ACCESS_
    }
#endif
}

void Terrain::SetDetailThreshold(float threshold)
{
    m_DetailThreshold = threshold;
}

float Terrain::GetDetailThreshold()
{
    return m_DetailThreshold;
}

int Terrain::GetWidthVertices()
{
    return m_WidthVertices;
}

int Terrain::GetHeightVertices()
{
    return m_HeightVertices;
}

float Terrain::GetWidth() const
{
    return (float)(m_WidthVertices - 1) * m_VertexSpacing;
}

float Terrain::GetHeight() const
{
    return (float)(m_HeightVertices - 1) * m_VertexSpacing;
}

float Terrain::GetMaxElevation() const
{
    return m_MaxElevation;
}

void Terrain::SetTextureFactory(TextureFactory* pFactory)
{
	m_pTextureFactory = pFactory;
}

float Terrain::GetVertexElevation(int index) const
{
#ifdef _PROTECT_ACCESS_
    if (index < 0 || m_NumberOfVertices <= index)
        return 0.0f;
    else
#endif
        return m_pVertices[index].z;
}

void Terrain::SetVertexElevation(int index,float newElevation)
{
#ifdef _PROTECT_ACCESS_
    if (0 <= index && index < m_NumberOfVertices)
#endif
        m_pVertices[index].z = newElevation;
}

float Terrain::GetVertexSpacing()
{
    return m_VertexSpacing;
}

void Terrain::BuildBlocks()
{
    if (Settings::GetInstance()->IsHeadless())
        return;
    numLevels = 0.0f;
    numBlocks = 0.0f;
    for (int i = m_WidthVertices - 1; 2 <= i; i /= 2)
        numLevels += 1.0f;
    for (double j = 0.0f; j < numLevels; j += 1.0f)
        numBlocks += pow(4,j);
    if (Settings::GetInstance()->IsVerbose())
    {
        cout << "TERRAIN: Building " << numBlocks << " blocks; please wait..." << endl;
#ifdef _USE_RAYTRACING_SUPPORT_
        cout << "TERRAIN: Memory required at runtime for blocks = " << numBlocks * (sizeof(TerrainBlock) + 8 * sizeof(Triangle)) << " bytes" << endl;
#else
        cout << "TERRAIN: Memory required at runtime for blocks = " << numBlocks * sizeof(TerrainBlock) << " bytes" << endl;
#endif
        cout << ".............................." << endl;
        hashDelta = (double)numBlocks / 30.0f;
        cout << "#" << flush;
    }
    m_pVertexStatus = new BitArray(m_WidthVertices * m_HeightVertices);
    // We assume that the terrain's width is always a power of 2 + 1!
    m_pRootBlock = new TerrainBlock(0,m_WidthVertices - 1,this,NULL);
    if (Settings::GetInstance()->IsVerbose())
        cout << endl;
}

void Terrain::SetVertexStatus(int vertexIndex,bool status)
{
    if (status)
        m_pVertexStatus->SetBit(vertexIndex);
    else
        m_pVertexStatus->ClearBit(vertexIndex);
}

bool Terrain::GetVertexStatus(int vertexIndex)
{
    return m_pVertexStatus->IsBitSet(vertexIndex);
}

int Terrain::Tessellate()
{
    if (m_pTriangleStrips == NULL)
    {
        long int maxNumStrips = (GetWidthVertices() - 1) * (GetHeightVertices() - 1);
        try
        {
            if (m_MaxNumberOfPrimitives < maxNumStrips)
                maxNumStrips = m_MaxNumberOfPrimitives;
            if (Settings::GetInstance()->IsVerbose())
                cout << "TERRAIN: Allocating " << maxNumStrips << " triangle strips and fans (" << maxNumStrips * sizeof(TriangleStrip) + maxNumStrips * sizeof(TriangleFan) << " bytes)\n" << endl;
            m_pTriangleStrips = new TriangleStrip[maxNumStrips];
            m_pTriangleFans = new TriangleFan[maxNumStrips];
            if (m_pTriangleStrips == NULL || m_pTriangleFans == NULL)
            {
                cout << "TERRAIN: " << "Not enough memory to build terrain triangles" << endl;
                exit(1);
            }
        }
        catch(...)
        {
            cout << "TERRAIN: " << "Not enough memory to build terrain triangles" << endl;
            exit(1);
        }
    }

    double matModelview[16];
    double matProjection[16];
    GLint viewport[4];
    glGetDoublev(GL_MODELVIEW_MATRIX, matModelview);
    glGetDoublev(GL_PROJECTION_MATRIX, matProjection);
    glGetIntegerv(GL_VIEWPORT,viewport);

    ExtractFrustum();

    m_pVertexStatus->Clear();
    m_CountStrips = m_CountFans = 0;
    m_pRootBlock->Tessellate(matModelview,matProjection,viewport,m_pTriangleStrips,m_pTriangleFans,&m_CountStrips,&m_CountFans,this);
    return m_CountStrips * 2 + m_CountFans * 6;
}

bool Terrain::SetCommonTexture(Uint8* pBuffer,int width,int height)
{
    bool bSuccess = false;
	// Test to see if the image is a power of 2 in both width and height.
	if (!IsPowerOf2(width) || !IsPowerOf2(height))
	{
		string msg("The detail texture image file is NOT a power of 2 in both width and height.\nTexture files must be a power of 2 in both width and height.");
		throw new DemeterException(msg);
	}
	m_pCommonTexture = new Texture(pBuffer,width,height,width,0,false,Settings::GetInstance()->IsTextureCompression());
	bSuccess = true;
	if (Settings::GetInstance()->IsVerbose())
		cout << "TERRAIN: Common texture set successfully" << endl;
    return bSuccess;
}

#ifdef _USE_SDL_
bool Terrain::SetCommonTexture(char* szFilename)
{
    bool bSuccess = false;
    char szFullFilename[MAX_FILENAME_LENGTH];
    Settings::GetInstance()->PrependMediaPath(szFilename,szFullFilename);
    if (!Settings::GetInstance()->IsCompilerOnly())
    {
        if (Settings::GetInstance()->IsVerbose())
                cout << "TERRAIN: Setting common texture to " << szFilename << endl;
        int width,height;
        Uint8* pBuffer;
        LoadImage(szFullFilename,&width,&height,&pBuffer);
        if (width == 0)
        {
            string msg("Failed to load detail texture image file '");
            msg += szFullFilename;
            msg += "'; This means that the file was not found or it is not an image type that Demeter can read.";
            throw new DemeterException(msg);
        }
        else
        {
            // Test to see if the image is a power of 2 in both width and height.
            if (!IsPowerOf2(width) || !IsPowerOf2(height))
            {
                string msg("The detail texture image file '");
                msg += szFullFilename;
                msg += "' is NOT a power of 2 in both width and height.\nTexture files must be a power of 2 in both width and height.";
                throw new DemeterException(msg);
            }
            m_pCommonTexture = new Texture(pBuffer,width,height,width,0,false,Settings::GetInstance()->IsTextureCompression());
            delete[] pBuffer;
            bSuccess = true;
            if (Settings::GetInstance()->IsVerbose())
                cout << "TERRAIN: Common texture set successfully" << endl;
        }
    }
    else
    {
        FILE* testFile = fopen(szFullFilename,"rb");
        if (testFile)
        {
            bSuccess = true;
            fclose(testFile);
        }
    }
    return bSuccess;
}
#endif

bool Terrain::SetTexture(Uint8* pBuffer,int width,int height,bool bUseBorders)
{
    if (Settings::GetInstance()->IsHeadless())
        return true;
    bool bSuccess = false;
	// Test to see if the image is a power of 2 in both width and height.
	if (!IsPowerOf2(width) || !IsPowerOf2(height))
	{
		string msg("The texture is NOT a power of 2 in both width and height.\nTextures must be a power of 2 in both width and height.");
		throw new DemeterException(msg);
	}
	ChopTexture(pBuffer,width,height,256,bUseBorders);
	bSuccess = true;
	if (Settings::GetInstance()->IsVerbose())
		cout << "TERRAIN: Texture set successfully" << endl;
    return bSuccess;
}

#ifdef _USE_SDL_
bool Terrain::SetTexture(char* szFilename,bool bUseBorders)
{
    if (Settings::GetInstance()->IsHeadless())
        return true;
    bool bSuccess = false;
    char szFullFilename[MAX_FILENAME_LENGTH];
    Settings::GetInstance()->PrependMediaPath(szFilename,szFullFilename);
    if (!Settings::GetInstance()->IsCompilerOnly())
    {
        if (Settings::GetInstance()->IsVerbose())
            cout << "TERRAIN: Setting texture to '" << szFullFilename << "'" << endl;
        int width,height;
        Uint8* pBuffer;
        LoadImage(szFullFilename,&width,&height,&pBuffer);
        if (width == 0)
        {
            string msg("Failed to load texture image file '");
            msg += szFullFilename;
            msg += "'; This means that the file was not found or it is not an image type that Demeter can read.";
            throw new DemeterException(msg);
        }
        else
        {
            // Test to see if the image is a power of 2 in both width and height.
            if (!IsPowerOf2(width) || !IsPowerOf2(height))
            {
                string msg("The texture image file '");
                msg += szFullFilename;
                msg += "' is NOT a power of 2 in both width and height.\nTexture files must be a power of 2 in both width and height.";
                throw new DemeterException(msg);
            }
            ChopTexture(pBuffer,width,height,256,bUseBorders);
            delete[] pBuffer;
            bSuccess = true;
            if (Settings::GetInstance()->IsVerbose())
                cout << "TERRAIN: Texture set successfully" << endl;
        }
    }
    else
    {
        FILE* fileTest = fopen(szFullFilename,"rb");
        if (fileTest)
        {
            bSuccess = true;
            fclose(fileTest);
        }
    }
    return bSuccess;
}
#endif

void Terrain::ChopTexture(Uint8* pImage,int width,int height,int tileSize,bool bUseBorders)
{
    // It is assumed that the image is in a 3-byte per pixel, RGB format, with no padding on the pixel rows
    m_TextureTileWidth = (int)(((float)(m_WidthVertices / (width / tileSize))) * m_VertexSpacing);
    m_TextureTileHeight = (int)(((float)(m_HeightVertices / (height / tileSize))) * m_VertexSpacing);
    m_NumberOfTextureTilesWidth = width / tileSize;
    m_NumberOfTextureTilesHeight = height / tileSize;
    m_NumberOfTextureTiles = (width / tileSize) * (height / tileSize);
    m_TileSize = tileSize;

    if (bUseBorders)
    {
        unsigned char* pBorderedImage = new Uint8[(width + 2) * (height + 2) * 3];

        int i;
        // Create a temporary version of the original texture that is surrounded by a 1-pixel border.
        for (i = 0; i < height; i++)
            for (int j = 0; j < width * 3; j++)
                pBorderedImage[(i + 1) * (width + 2) * 3 + j + 3] = pImage[i * width * 3 + j];

        // Create texture tiles by roaming across the bordered image.
        int k = 0;
        for (i = 0; i < height; i += tileSize)
        {
            for (int j = 0; j < width; j += tileSize)
            {
                Uint8* pTile = pBorderedImage + i * (width + 2) * 3 + j * 3;
#ifdef _USE_VERTEX_ARRAYS_
                Texture* pTexture = new Texture(pTile,tileSize,tileSize,width + 2,1,false,Settings::GetInstance()->IsTextureCompression());
                m_Textures.push_back(pTexture);
#else
                Texture* pTexture = new Texture(pTile,tileSize,tileSize,width + 2,1,true,Settings::GetInstance()->IsTextureCompression());
                m_Textures.push_back(pTexture);
#endif
            }
        }
        delete[] pBorderedImage;
    }
    else
    {
        // Create texture tiles by roaming across the bordered image.
        int k = 0;
        for (int i = 0; i < height; i += tileSize)
        {
            for (int j = 0; j < width; j += tileSize)
            {
                Uint8* pTile = pImage + i * width * 3 + j * 3;
#ifdef _USE_VERTEX_ARRAYS_
                Texture* pTexture = new Texture(pTile,tileSize,tileSize,width,0,false,Settings::GetInstance()->IsTextureCompression());
                m_Textures.push_back(pTexture);
#else
                Texture* pTexture = new Texture(pTile,tileSize,tileSize,width,0,true,Settings::GetInstance()->IsTextureCompression());
                m_Textures.push_back(pTexture);
#endif
            }
        }
    }
}

void Terrain::SetMaximumVisibleBlockSize(int stride)
{
    m_MaximumVisibleBlockSize = stride;
}

float Terrain::GetCommonTextureRepeats()
{
    return m_CommonTextureRepeats;
}

void Terrain::SetCommonTextureRepeats(float commonTextureRepeats)
{
    m_CommonTextureRepeats = commonTextureRepeats;
}

int Terrain::ModelViewMatrixChanged()
{
    int count = Tessellate();
    m_pRootBlock->RepairCracks(this,m_pTriangleStrips,m_pTriangleFans,&m_CountStrips,&m_CountFans);
    return count;
}

void Terrain::Render()
{
    int i,j;
    bool bUseMultiTexture = (bMultiTextureSupported && m_pCommonTexture != NULL);
    
	glFrontFace(GL_CW);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);

    if (bUseMultiTexture && m_pCommonTexture != NULL)
    {
        glActiveTextureARB_ptr(GL_TEXTURE0_ARB);
        glEnable(GL_TEXTURE_2D);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
        glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, GL_REPLACE);

        glActiveTextureARB_ptr(GL_TEXTURE1_ARB);
        glEnable(GL_TEXTURE_2D);
        glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE, GL_MODULATE);
        glTexEnvf(GL_TEXTURE_ENV,GL_COMBINE_RGB_EXT,GL_ADD);
		GLuint cTexId = m_pCommonTexture->UploadTexture();
        glBindTexture(GL_TEXTURE_2D,cTexId);
    }
    else
        glEnable(GL_TEXTURE_2D);

    for (i = 0; i < m_CountStrips; i++)
        m_pTriangleStrips[i].Setup(this);

    for (i = 0; i < m_CountFans; i++)
        m_pTriangleFans[i].Setup(this);

    if (bUseMultiTexture)
        glActiveTextureARB_ptr(GL_TEXTURE0_ARB);

    glNormal3f(0.0f,0.0f,1.0f);

    // Prevent state thrashing by rendering one texture at a time...
    for (i = 0; i < m_NumberOfTextureTiles; i++)
    {
        bool firstTime = true;
        for (j = 0; j < m_CountStrips; j++)
        {
            if (m_pTriangleStrips[j].textureId == i)
            {
                if (firstTime)
                {
                    glBindTexture(GL_TEXTURE_2D,GetTerrainTile(i));
                    firstTime = false;
                }
                m_pTriangleStrips[j].Render(this);
            }
        }
        for (j = 0; j < m_CountFans; j++)
        {
            if (m_pTriangleFans[j].textureId == i)
            {
                if (firstTime)
                {
                    glBindTexture(GL_TEXTURE_2D,GetTerrainTile(i));
                    firstTime = false;
                }
                m_pTriangleFans[j].Render(this);
            }
        }
        if (Settings::GetInstance()->UseDynamicTextures() && firstTime)
        {
            UnloadTerrainTile(i);
        }
    }
    glFrontFace(GL_CCW);
    // Turn multi-texture back off again so the client application doesn't end up in an unexpected state.
	if (bUseMultiTexture)
	{
		glActiveTextureARB_ptr(GL_TEXTURE1_ARB);
		glDisable(GL_TEXTURE_2D);
		glActiveTextureARB_ptr(GL_TEXTURE0_ARB);
		glDisable(GL_TEXTURE_2D);
	}
}

void Terrain::DisableTextures()
{
    if (bMultiTextureSupported)
    {
        glActiveTextureARB_ptr(GL_TEXTURE0_ARB);
        glDisable(GL_TEXTURE_2D);
        glActiveTextureARB_ptr(GL_TEXTURE1_ARB);
        glDisable(GL_TEXTURE_2D);
    }
    else
        glDisable(GL_TEXTURE_2D);
}

void Terrain::EnableTextures()
{
    if (bMultiTextureSupported)
    {
        glActiveTextureARB_ptr(GL_TEXTURE0_ARB);
        glEnable(GL_TEXTURE_2D);
        glActiveTextureARB_ptr(GL_TEXTURE1_ARB);
        glEnable(GL_TEXTURE_2D);
    }
    else
        glEnable(GL_TEXTURE_2D);
}

TriangleStrip::TriangleStrip()
{
}

TriangleStrip::~TriangleStrip()
{
}

void TriangleStrip::Setup(Terrain* pTerrain)
{
    if (m_bEnabled)
    {
        minX = pTerrain->m_pVertices[m_pVertices[0]].x;
        minY = pTerrain->m_pVertices[m_pVertices[0]].y;
        for (int i = 1; i < m_NumberOfVertices; i++)
        {
            if (pTerrain->m_pVertices[m_pVertices[i]].x < minX)
                minX = pTerrain->m_pVertices[m_pVertices[i]].x;
            if (pTerrain->m_pVertices[m_pVertices[i]].y < minY)
                minY = pTerrain->m_pVertices[m_pVertices[i]].y;
        }
        int tileY = (int)minY / pTerrain->GetTextureTileHeight();
        int tileX = (int)minX / pTerrain->GetTextureTileWidth();
        textureId = tileY * pTerrain->GetNumberOfTextureTilesWidth() + tileX;
    }
}

void TriangleStrip::Render(Terrain* pTerrain)
{
    if (m_bEnabled)
    {
// IMPORTANT NOTE: Using glDrawElements() is a modest speed gain, but we lose multi-texturing and the use of different
// texture coordinates for a single vertex! If there is some way to still have these features while using glDrawElements,
// then re-enable this code in place of the rest of this method.
#ifdef _USE_VERTEX_ARRAYS_
        glDrawElements(GL_TRIANGLE_STRIP,m_NumberOfVertices,GL_UNSIGNED_INT,m_pVertices);
#else
        float texU,texV;
        glBegin(GL_TRIANGLE_STRIP);
        for (int i = 0; i < m_NumberOfVertices; i++)
        {
            texU = pTerrain->m_pVertices[m_pVertices[i]].x / (float)pTerrain->GetTextureTileWidth();
            texV = pTerrain->m_pVertices[m_pVertices[i]].y / (float)pTerrain->GetTextureTileHeight();
            if ((float)texU - (int)texU < EPSILON)
                texU = 1.0f;
            else
                texU = (float)texU - (int)texU;
            if ((float)texV - (int)texV < EPSILON)
                texV = 1.0f;
            else
                texV = (float)texV - (int)texV;
            if (fabs(1.0f - texU) <= EPSILON && fabs(minX - pTerrain->m_pVertices[m_pVertices[i]].x) <= EPSILON)
                texU = 0.0f;
            if (fabs(1.0f - texV) <= EPSILON && fabs(minY - pTerrain->m_pVertices[m_pVertices[i]].y) <= EPSILON)
                texV = 0.0f;
            if (!bMultiTextureSupported)
                glTexCoord2f(texU,texV);
            else
            {
                glMultiTexCoord2fARB_ptr(GL_TEXTURE0_ARB,texU,texV);
                glMultiTexCoord2fARB_ptr(GL_TEXTURE1_ARB,texU * pTerrain->GetCommonTextureRepeats(),texV * pTerrain->GetCommonTextureRepeats());
            }
            glVertex3f(pTerrain->m_pVertices[m_pVertices[i]].x + pTerrain->m_OffsetX,pTerrain->m_pVertices[m_pVertices[i]].y + pTerrain->m_OffsetY,pTerrain->m_pVertices[m_pVertices[i]].z);
        }
        glEnd();
#endif
    }
}

TriangleFan::TriangleFan()
{
}

TriangleFan::~TriangleFan()
{
}

void TriangleFan::Setup(Terrain* pTerrain)
{
    minX = pTerrain->m_pVertices[m_pVertices[0]].x;
    minY = pTerrain->m_pVertices[m_pVertices[0]].y;
    for (int i = 1; i < m_NumberOfVertices; i++)
    {
        if (pTerrain->m_pVertices[m_pVertices[i]].x < minX)
            minX = pTerrain->m_pVertices[m_pVertices[i]].x;
        if (pTerrain->m_pVertices[m_pVertices[i]].y < minY)
            minY = pTerrain->m_pVertices[m_pVertices[i]].y;
    }
    int tileY = (int)minY / pTerrain->GetTextureTileHeight();
    int tileX = (int)minX / pTerrain->GetTextureTileWidth();
    textureId = tileY * pTerrain->GetNumberOfTextureTilesWidth() + tileX;
}

void TriangleFan::Render(Terrain* pTerrain)
{
// IMPORTANT NOTE: Using glDrawElements() is a modest speed gain, but we lose multi-texturing and the use of different
// texture coordinates for a single vertex! If there is some way to still have these features while using glDrawElements,
// then re-enable this code in place of the rest of this method.
#ifdef _USE_VERTEX_ARRAYS_
    glDrawElements(GL_TRIANGLE_FAN,m_NumberOfVertices,GL_UNSIGNED_INT,m_pVertices);
#else
    float texU,texV;
    glBegin(GL_TRIANGLE_FAN);
    for (int i = 0; i < m_NumberOfVertices; i++)
    {
        texU = (pTerrain->m_pVertices[m_pVertices[i]].x) / (float)pTerrain->GetTextureTileWidth();
        texV = (pTerrain->m_pVertices[m_pVertices[i]].y) / (float)pTerrain->GetTextureTileHeight();
        if ((float)texU - (int)texU < EPSILON)
            texU = 1.0f;
        else
            texU = (float)texU - (int)texU;
        if ((float)texV - (int)texV < EPSILON)
            texV = 1.0f;
        else
            texV = (float)texV - (int)texV;
        if (fabs(1.0f - texU) <= EPSILON && fabs(minX - pTerrain->m_pVertices[m_pVertices[i]].x) <= EPSILON)
            texU = 0.0f;
        if (fabs(1.0f - texV) <= EPSILON && fabs(minY - pTerrain->m_pVertices[m_pVertices[i]].y) <= EPSILON)
            texV = 0.0f;
        if (!bMultiTextureSupported)
            glTexCoord2f(texU,texV);
        else
        {
            glMultiTexCoord2fARB_ptr(GL_TEXTURE0_ARB,texU,texV);
            glMultiTexCoord2fARB_ptr(GL_TEXTURE1_ARB,texU * pTerrain->GetCommonTextureRepeats(),texV * pTerrain->GetCommonTextureRepeats());
        }
        glVertex3f(pTerrain->m_pVertices[m_pVertices[i]].x + pTerrain->m_OffsetX,pTerrain->m_pVertices[m_pVertices[i]].y + pTerrain->m_OffsetY,pTerrain->m_pVertices[m_pVertices[i]].z);
    }
    glEnd();
#endif
}

int Terrain::GetTextureTileWidth()
{
    return m_TextureTileWidth;
}

int Terrain::GetTextureTileHeight()
{
    return m_TextureTileHeight;
}

int Terrain::GetNumberOfTextureTilesWidth()
{
    return m_NumberOfTextureTilesWidth;
}

int Terrain::GetNumberOfTextureTilesHeight()
{
    return m_NumberOfTextureTilesHeight;
}

GLuint Terrain::GetTerrainTile(int index)
{
    GLuint texID;
    if (m_pTextureFactory == NULL)
    {
        Texture* pTexture = m_Textures[index];
        texID = pTexture->UploadTexture();
    }
    else
    {
		int tileX = index % (int)m_NumberOfTextureTilesWidth;
		int tileY = index / (int)m_NumberOfTextureTilesHeight;
        float originX = tileX * m_TextureTileWidth;
        float originY = tileY * m_TextureTileHeight;
        texID = m_pTextureFactory->GetTexture(index,originX,originY,m_TextureTileWidth,m_TextureTileHeight);
    }
    return texID;
}

void Terrain::UnloadTerrainTile(int index)
{
    if (m_pTextureFactory == NULL)
    {
        Texture* pTexture = m_Textures[index];
        pTexture->UnloadTexture();
    }
    else
    {
        float originX = (float)(index % (int)m_TextureTileWidth);
        float originY = (index / (float)m_TextureTileHeight);
        m_pTextureFactory->UnloadTexture(index,originX,originY);
    }
}

void Terrain::ExtractFrustum()
{
// Thanks to Mark Morley for this OpenGL frustum testing code.
// See all of it at http://www.markmorley.com/opengl/frustumculling.html.
    float proj[16];
    float modl[16];
    float clip[16];
    float t;
    /* Get the current PROJECTION matrix from OpenGL */
    glGetFloatv(GL_PROJECTION_MATRIX, proj);
    /* Get the current MODELVIEW matrix from OpenGL */
    glGetFloatv(GL_MODELVIEW_MATRIX, modl);
    /* Combine the two matrices (multiply projection by modelview) */
    clip[ 0] = modl[ 0] * proj[ 0] + modl[ 1] * proj[ 4] + modl[ 2] * proj[ 8] + modl[ 3] * proj[12];
    clip[ 1] = modl[ 0] * proj[ 1] + modl[ 1] * proj[ 5] + modl[ 2] * proj[ 9] + modl[ 3] * proj[13];
    clip[ 2] = modl[ 0] * proj[ 2] + modl[ 1] * proj[ 6] + modl[ 2] * proj[10] + modl[ 3] * proj[14];
    clip[ 3] = modl[ 0] * proj[ 3] + modl[ 1] * proj[ 7] + modl[ 2] * proj[11] + modl[ 3] * proj[15];
    clip[ 4] = modl[ 4] * proj[ 0] + modl[ 5] * proj[ 4] + modl[ 6] * proj[ 8] + modl[ 7] * proj[12];
    clip[ 5] = modl[ 4] * proj[ 1] + modl[ 5] * proj[ 5] + modl[ 6] * proj[ 9] + modl[ 7] * proj[13];
    clip[ 6] = modl[ 4] * proj[ 2] + modl[ 5] * proj[ 6] + modl[ 6] * proj[10] + modl[ 7] * proj[14];
    clip[ 7] = modl[ 4] * proj[ 3] + modl[ 5] * proj[ 7] + modl[ 6] * proj[11] + modl[ 7] * proj[15];
    clip[ 8] = modl[ 8] * proj[ 0] + modl[ 9] * proj[ 4] + modl[10] * proj[ 8] + modl[11] * proj[12];
    clip[ 9] = modl[ 8] * proj[ 1] + modl[ 9] * proj[ 5] + modl[10] * proj[ 9] + modl[11] * proj[13];
    clip[10] = modl[ 8] * proj[ 2] + modl[ 9] * proj[ 6] + modl[10] * proj[10] + modl[11] * proj[14];
    clip[11] = modl[ 8] * proj[ 3] + modl[ 9] * proj[ 7] + modl[10] * proj[11] + modl[11] * proj[15];
    clip[12] = modl[12] * proj[ 0] + modl[13] * proj[ 4] + modl[14] * proj[ 8] + modl[15] * proj[12];
    clip[13] = modl[12] * proj[ 1] + modl[13] * proj[ 5] + modl[14] * proj[ 9] + modl[15] * proj[13];
    clip[14] = modl[12] * proj[ 2] + modl[13] * proj[ 6] + modl[14] * proj[10] + modl[15] * proj[14];
    clip[15] = modl[12] * proj[ 3] + modl[13] * proj[ 7] + modl[14] * proj[11] + modl[15] * proj[15];
    /* Extract the numbers for the RIGHT plane */
    m_Frustum[0][0] = clip[ 3] - clip[ 0];
    m_Frustum[0][1] = clip[ 7] - clip[ 4];
    m_Frustum[0][2] = clip[11] - clip[ 8];
    m_Frustum[0][3] = clip[15] - clip[12];
    /* Normalize the result */
    t = sqrt(m_Frustum[0][0] * m_Frustum[0][0] + m_Frustum[0][1] * m_Frustum[0][1] + m_Frustum[0][2] * m_Frustum[0][2]);
    m_Frustum[0][0] /= t;
    m_Frustum[0][1] /= t;
    m_Frustum[0][2] /= t;
    m_Frustum[0][3] /= t;
    /* Extract the numbers for the LEFT plane */
    m_Frustum[1][0] = clip[ 3] + clip[ 0];
    m_Frustum[1][1] = clip[ 7] + clip[ 4];
    m_Frustum[1][2] = clip[11] + clip[ 8];
    m_Frustum[1][3] = clip[15] + clip[12];
    /* Normalize the result */
    t = sqrt(m_Frustum[1][0] * m_Frustum[1][0] + m_Frustum[1][1] * m_Frustum[1][1] + m_Frustum[1][2] * m_Frustum[1][2]);
    m_Frustum[1][0] /= t;
    m_Frustum[1][1] /= t;
    m_Frustum[1][2] /= t;
    m_Frustum[1][3] /= t;
    /* Extract the BOTTOM plane */
    m_Frustum[2][0] = clip[ 3] + clip[ 1];
    m_Frustum[2][1] = clip[ 7] + clip[ 5];
    m_Frustum[2][2] = clip[11] + clip[ 9];
    m_Frustum[2][3] = clip[15] + clip[13];
    /* Normalize the result */
    t = sqrt(m_Frustum[2][0] * m_Frustum[2][0] + m_Frustum[2][1] * m_Frustum[2][1] + m_Frustum[2][2] * m_Frustum[2][2]);
    m_Frustum[2][0] /= t;
    m_Frustum[2][1] /= t;
    m_Frustum[2][2] /= t;
    m_Frustum[2][3] /= t;
    /* Extract the TOP plane */
    m_Frustum[3][0] = clip[ 3] - clip[ 1];
    m_Frustum[3][1] = clip[ 7] - clip[ 5];
    m_Frustum[3][2] = clip[11] - clip[ 9];
    m_Frustum[3][3] = clip[15] - clip[13];
    /* Normalize the result */
    t = sqrt(m_Frustum[3][0] * m_Frustum[3][0] + m_Frustum[3][1] * m_Frustum[3][1] + m_Frustum[3][2] * m_Frustum[3][2]);
    m_Frustum[3][0] /= t;
    m_Frustum[3][1] /= t;
    m_Frustum[3][2] /= t;
    m_Frustum[3][3] /= t;
    /* Extract the FAR plane */
    m_Frustum[4][0] = clip[ 3] - clip[ 2];
    m_Frustum[4][1] = clip[ 7] - clip[ 6];
    m_Frustum[4][2] = clip[11] - clip[10];
    m_Frustum[4][3] = clip[15] - clip[14];
    /* Normalize the result */
    t = sqrt(m_Frustum[4][0] * m_Frustum[4][0] + m_Frustum[4][1] * m_Frustum[4][1] + m_Frustum[4][2] * m_Frustum[4][2]);
    m_Frustum[4][0] /= t;
    m_Frustum[4][1] /= t;
    m_Frustum[4][2] /= t;
    m_Frustum[4][3] /= t;
    /* Extract the NEAR plane */
    m_Frustum[5][0] = clip[ 3] + clip[ 2];
    m_Frustum[5][1] = clip[ 7] + clip[ 6];
    m_Frustum[5][2] = clip[11] + clip[10];
    m_Frustum[5][3] = clip[15] + clip[14];
    /* Normalize the result */
    t = sqrt(m_Frustum[5][0] * m_Frustum[5][0] + m_Frustum[5][1] * m_Frustum[5][1] + m_Frustum[5][2] * m_Frustum[5][2]);
    m_Frustum[5][0] /= t;
    m_Frustum[5][1] /= t;
    m_Frustum[5][2] /= t;
    m_Frustum[5][3] /= t;
}

bool Terrain::CubeInFrustum(float x, float y, float z, float size)
{
// Thanks to Mark Morley for this OpenGL frustum testing code.
// See all of it at http://www.markmorley.com/opengl/frustumculling.html.
    for(int p = 0; p < 6; p++)
    {
        if(m_Frustum[p][0] * (x - size) + m_Frustum[p][1] * (y - size) + m_Frustum[p][2] * (z - size) + m_Frustum[p][3] > 0)
            continue;
        if(m_Frustum[p][0] * (x + size) + m_Frustum[p][1] * (y - size) + m_Frustum[p][2] * (z - size) + m_Frustum[p][3] > 0)
            continue;
        if(m_Frustum[p][0] * (x - size) + m_Frustum[p][1] * (y + size) + m_Frustum[p][2] * (z - size) + m_Frustum[p][3] > 0)
            continue;
        if(m_Frustum[p][0] * (x + size) + m_Frustum[p][1] * (y + size) + m_Frustum[p][2] * (z - size) + m_Frustum[p][3] > 0)
            continue;
        if(m_Frustum[p][0] * (x - size) + m_Frustum[p][1] * (y - size) + m_Frustum[p][2] * (z + size) + m_Frustum[p][3] > 0)
            continue;
        if(m_Frustum[p][0] * (x + size) + m_Frustum[p][1] * (y - size) + m_Frustum[p][2] * (z + size) + m_Frustum[p][3] > 0)
            continue;
        if(m_Frustum[p][0] * (x - size) + m_Frustum[p][1] * (y + size) + m_Frustum[p][2] * (z + size) + m_Frustum[p][3] > 0)
            continue;
        if(m_Frustum[p][0] * (x + size) + m_Frustum[p][1] * (y + size) + m_Frustum[p][2] * (z + size) + m_Frustum[p][3] > 0)
            continue;
        return false;
    }
    return true;
}

#ifdef _USE_RAYTRACING_SUPPORT_
float Terrain::IntersectRay(float startX,float startY,float startZ,float dirX,float dirY,float dirZ,float& intersectX,float& intersectY,float& intersectZ)
{
    Ray ray;
    float distance = INFINITY;
    Vector point;
    point.x = point.y = point.z = -1.0f;

    ray.m_Origin.x = startX;
    ray.m_Origin.y = startY;
    ray.m_Origin.z = startZ;
    ray.m_Direction.x = dirX;
    ray.m_Direction.y = dirY;
    ray.m_Direction.z = dirZ;
    ray.m_Direction.Normalize();
    m_pRootBlock->IntersectRay(ray,point,distance,this);
    intersectX = point.x;
    intersectY = point.y;
    intersectZ = point.z;
    return distance;
}
#endif

void Terrain::SetLatticePosition(int x,int y)
{
    m_LatticePositionX = x;
    m_LatticePositionY = y;
}

void Terrain::GetLatticePosition(int& x,int& y)
{
    x = m_LatticePositionX;
    y = m_LatticePositionY;
}

#ifdef _USE_RAYTRACING_SUPPORT_
void TerrainBlock::IntersectRay(Ray& ray,Vector& intersectionPoint,float& lowestDistance,Terrain* pTerrain)
{
// First test ray against this block's bounding box.
    if (RayBoxIntersect(&ray,&m_BoundingBox,NULL,NULL))
    {
        if (2 < m_Stride)
        {
            m_pChildren[0]->IntersectRay(ray,intersectionPoint,lowestDistance,pTerrain);
            m_pChildren[1]->IntersectRay(ray,intersectionPoint,lowestDistance,pTerrain);
            m_pChildren[2]->IntersectRay(ray,intersectionPoint,lowestDistance,pTerrain);
            m_pChildren[3]->IntersectRay(ray,intersectionPoint,lowestDistance,pTerrain);
        }
        else
        {
            float distance;
            Vector point;
            for (int i = 0; i < 8; i++)
            {
                if (RayPlaneIntersect(&ray,m_pTriangles[i].GetPlane(),&point,&distance) == 1)
                {
                    if (i == 0 || i == 2 || i == 4 || i == 6)
                    {
                        if (m_pTriangles[i].GetVertex(0)->x <= point.x &&
                            m_pTriangles[i].GetVertex(0)->y <= point.y &&
                            point.x <= m_pTriangles[i].GetVertex(2)->x &&
                            point.y <= m_pTriangles[i].GetVertex(1)->y &&
                            ((fmod(point.y,pTerrain->m_VertexSpacing) + fmod(point.x,pTerrain->m_VertexSpacing)) <= pTerrain->m_VertexSpacing))
                        {
                            if (distance < lowestDistance)
                            {
                                lowestDistance = distance;
                                intersectionPoint.x = point.x;
                                intersectionPoint.y = point.y;
                                intersectionPoint.z = point.z;
                            }
                        }
                    }
                    else
                    {
                        if (m_pTriangles[i].GetVertex(1)->x <= point.x &&
                            m_pTriangles[i].GetVertex(0)->y <= point.y &&
                            point.x <= m_pTriangles[i].GetVertex(2)->x &&
                            point.y <= m_pTriangles[i].GetVertex(1)->y &&
                            ((fmod(point.y,pTerrain->m_VertexSpacing) + fmod(point.x,pTerrain->m_VertexSpacing)) >= pTerrain->m_VertexSpacing))
                        {
                            if (distance < lowestDistance)
                            {
                                lowestDistance = distance;
                                intersectionPoint.x = point.x;
                                intersectionPoint.y = point.y;
                                intersectionPoint.z = point.z;
                            }
                        }
                    }
                }
            }
        }
    }
}

int RayPlaneIntersect(Ray *ray,Plane *plane,Vector* point,float *distance)
{
    float           vd,vo,PnDOTRo,t;

    vd = plane->a * ray->m_Direction.x + plane->b * ray->m_Direction.y + plane->c * ray->m_Direction.z;
    if (vd == 0.0)
        // The plane is parallel to the ray. I've never seen this happen but someday it will . . .
        return -1;
    if (vd > 0.0)
    {
        // The plane is facing away from the ray so no intersection occurs.
        return -2;
    }
    PnDOTRo = plane->a * ray->m_Origin.x + plane->b * ray->m_Origin.y + plane->c * ray->m_Origin.z;
    vo = -1.0f * (PnDOTRo + plane->d);
    t = vo / vd;
    if (t < 0.0f)
        // The intersection occurs behind the ray's origin.
        return -3;
    point->x = ray->m_Origin.x + ray->m_Direction.x * t;
    point->y = ray->m_Origin.y + ray->m_Direction.y * t;
    point->z = ray->m_Origin.z + ray->m_Direction.z * t;
    if (distance != NULL)
        *distance = t;
    return 1;
}

int RayBoxIntersect(Ray *ray,Box *box,Vector *point,float *distance)
{
    float       tnear,tfar,t1,t2;

    tnear = -INFINITY;
    tfar = INFINITY;

// Find intersection with x-aligned planes of box.
    // If the ray is parallel to the box and not within the planes of the box it misses.
    if (ray->m_Direction.x == 0.0)
        if ((ray->m_Origin.x < box->m_Min.x) && (ray->m_Origin.x > box->m_Max.x))
            return 0;
    // Calculate intersection distance with the box's planes.
    t1 = (box->m_Min.x - ray->m_Origin.x) / ray->m_Direction.x;
    t2 = (box->m_Max.x - ray->m_Origin.x) / ray->m_Direction.x;
    if (t1 > t2)
    {
        float tmp = t1;
        t1 = t2;
        t2 = tmp;
    }
    if (t1 > tnear)
        tnear = t1;
    if (t2 < tfar)
        tfar = t2;
    if (tnear > tfar)
        return 0;
    if (tfar < 0.0)
        return 0;
// Find intersection with y-aligned planes of box.
    // If the ray is parallel to the box and not within the planes of the box it misses.
    if (ray->m_Direction.y == 0.0)
        if ((ray->m_Origin.y < box->m_Min.y) && (ray->m_Origin.y > box->m_Max.y))
            return 0;
    // Calculate intersection distance with the box's planes.
    t1 = (box->m_Min.y - ray->m_Origin.y) / ray->m_Direction.y;
    t2 = (box->m_Max.y - ray->m_Origin.y) / ray->m_Direction.y;
    if (t1 > t2)
    {
        float tmp = t1;
        t1 = t2;
        t2 = tmp;
    }
    if (t1 > tnear)
        tnear = t1;
    if (t2 < tfar)
        tfar = t2;
    if (tnear > tfar)
        return 0;
    if (tfar < 0.0)
        return 0;
// Find intersection with z-aligned planes of box.
    // If the ray is parallel to the box and not within the planes of the box it misses.
    if (ray->m_Direction.z == 0.0)
        if ((ray->m_Origin.z < box->m_Min.z) && (ray->m_Origin.z > box->m_Max.z))
            return 0;
    // Calculate intersection distance with the box's planes.
    t1 = (box->m_Min.z - ray->m_Origin.z) / ray->m_Direction.z;
    t2 = (box->m_Max.z - ray->m_Origin.z) / ray->m_Direction.z;
    if (t1 > t2)
    {
        float tmp = t1;
        t1 = t2;
        t2 = tmp;
    }
    if (t1 > tnear)
        tnear = t1;
    if (t2 < tfar)
        tfar = t2;
    if (tnear > tfar)
        return 0;
    if (tfar < 0.0)
        return 0;
// If we survived all of the tests, the box is hit.
    if (point != NULL)
    {
        point->x = ray->m_Origin.x + tnear * ray->m_Direction.x;
        point->y = ray->m_Origin.y + tnear * ray->m_Direction.y;
        point->z = ray->m_Origin.z + tnear * ray->m_Direction.z;
        *distance = tnear;
    }
    return 1;
}
#endif

#ifdef _USE_SDL_
void LoadImage(char* szFilename,int* pWidth,int* pHeight,Uint8** pBuffer,bool bColorKey)
{
	char szFullFilename[MAX_FILENAME_LENGTH];
	Settings::GetInstance()->PrependMediaPath(szFilename,szFullFilename);
    SDL_Surface* pImage = IMG_Load(szFullFilename);
    if (pImage != NULL)
    {
        *pWidth = pImage->w;
        *pHeight = pImage->h;
        Uint8* pBufferTemp;
        if (bColorKey)
            pBufferTemp = new Uint8[*pWidth * *pHeight * 4];
        else
            pBufferTemp = new Uint8[*pWidth * *pHeight * 3];
        int i,j;
        Uint8* pImagePixels = (Uint8*)pImage->pixels;
        for (i = 0,j = 0; i < pImage->h * pImage->pitch; i += pImage->pitch)
        {
            Uint8* pImageRow = pImagePixels + i;
            for (Uint8* pImagePixel = pImageRow; pImagePixel < pImageRow + pImage->w * pImage->format->BytesPerPixel; pImagePixel += pImage->format->BytesPerPixel)
            {
                Uint8 red,green,blue,alpha;
                Uint32* pCurrentPixel = (Uint32*)pImagePixel;
                if (bColorKey)
                    SDL_GetRGBA(*pCurrentPixel,pImage->format,&red,&green,&blue,&alpha);
                else
                    SDL_GetRGBA(*pCurrentPixel,pImage->format,&red,&green,&blue,&alpha);
                pBufferTemp[j++] = red;
                pBufferTemp[j++] = green;
                pBufferTemp[j++] = blue;
                if (bColorKey)
                {
                    pBufferTemp[j++] = alpha;
                }
            }
        }
        *pBuffer = pBufferTemp;
        SDL_FreeSurface(pImage);
    }
    else
    {
        *pWidth = 0;
        *pHeight = 0;
        *pBuffer = NULL;
    }
}
#endif

GLuint CreateTexture(Uint8* pTexels,int width,int height,int rowLength,int border,int internalFormat,bool bClamp,bool bColorKey)
{
    GLuint texId;
    glGenTextures(1,&texId);
    glBindTexture(GL_TEXTURE_2D,texId);
    if (bClamp)
    {
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
    }
    else
    {
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
    }
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);
    glPixelStorei(GL_UNPACK_ROW_LENGTH,rowLength);
    glPixelStorei(GL_UNPACK_ALIGNMENT,1);
    if (bColorKey)
    {
        glTexImage2D(GL_TEXTURE_2D,0,internalFormat,width + 2 * border,height + 2 * border,border,GL_RGBA,GL_UNSIGNED_BYTE,pTexels);
        gluBuild2DMipmaps(GL_TEXTURE_2D,GL_RGBA,width,height,GL_RGBA,GL_UNSIGNED_BYTE,pTexels);
    }
    else
    {
        glTexImage2D(GL_TEXTURE_2D,0,internalFormat,width + 2 * border,height + 2 * border,border,GL_RGB,GL_UNSIGNED_BYTE,pTexels);
        gluBuild2DMipmaps(GL_TEXTURE_2D,GL_RGB,width,height,GL_RGB,GL_UNSIGNED_BYTE,pTexels);
    }
    glPixelStorei(GL_UNPACK_ROW_LENGTH,0);
    return texId;
}

bool IsPowerOf2(double number)
{
    const int MAX_POWER = 1024;
    bool isPowerOf2 = false;
    for(int i = 0; i < MAX_POWER && !isPowerOf2; i++)
    {
        if (pow(2,i) == number)
            isPowerOf2 = true;
    }
    return isPowerOf2;
}

Settings::Settings()
{
    m_szMediaPath = NULL;
    m_bVerbose = false;
    m_bIsCompilerOnly = false;
    m_IsHeadless = false;
    m_UseDynamicTextures = false;
}

Settings::~Settings()
{
    if (m_szMediaPath != NULL)
        delete[] m_szMediaPath;
}

Settings* Settings::GetInstance()
{
    if (pSettingsInstance == NULL)
        pSettingsInstance = new Settings();
    return pSettingsInstance;
}

bool Settings::SetProperty(char* szProperty,char* szValue)
{
    bool bSuccess = false;

    if (strcmp(szProperty,"verbose") == 0)
    {
        SetVerbose(strcmp(szValue,"true") == 0);
        bSuccess = true;
    }

    return bSuccess;
}

bool Settings::GetProperty(char* szProperty,char* szValue)
{
    bool bSuccess = false;

    if (strcmp(szProperty,"verbose") == 0)
    {
        sprintf(szValue,m_bVerbose ? "true" : "false");
        bSuccess = true;
    }
    else if (strcmp(szProperty,"glinfo") == 0)
    {
        sprintf(szValue,"OpenGL Vendor: %s; OpenGL Extensions Supported: %s",glGetString(GL_VENDOR),glGetString(GL_EXTENSIONS));
        bSuccess = true;
    }

    return bSuccess;
}

void Settings::SetUseDynamicTextures(bool use)
{
    m_UseDynamicTextures = use;
}

bool Settings::UseDynamicTextures()
{
    return m_UseDynamicTextures;
}

void Settings::SetMediaPath(char* szPath)
{
#ifdef _WIN32
    char separator = '\\';
#else
    char separator = '/';
#endif
    m_bCompressTextures = false;
    if (m_szMediaPath != NULL)
        delete[] m_szMediaPath;
    if (szPath[strlen(szPath) - 1] == separator)
    {
        m_szMediaPath = new char[strlen(szPath) + 1];
        sprintf(m_szMediaPath,szPath);
    }
    else
    {
        m_szMediaPath = new char[strlen(szPath) + 2];
        sprintf(m_szMediaPath,"%s%c",szPath,separator);
    }
}

void Settings::GetMediaPath(char** szPath)
{
    char* szOutput = new char[strlen(m_szMediaPath) + 1];
    sprintf(szOutput,m_szMediaPath);
    *szPath = szOutput;
}

void Settings::PrependMediaPath(char* szFilename,char* szFullFilename)
{
    sprintf(szFullFilename,"%s%s",m_szMediaPath,szFilename);
}

void Settings::SetHeadless(bool isHeadless)
{
    m_IsHeadless = isHeadless;
}

bool Settings::IsHeadless()
{
    return m_IsHeadless;
}

void Settings::SetVerbose(bool bVerbose)
{
    m_bVerbose = bVerbose;
}

bool Settings::IsVerbose()
{
    return m_bVerbose;
}

int Settings::GetScreenWidth()
{
    return m_ScreenWidth;
}

void Settings::SetScreenWidth(int width)
{
    m_ScreenWidth = width;
}

int Settings::GetScreenHeight()
{
    return m_ScreenHeight;
}

void Settings::SetScreenHeight(int height)
{
    m_ScreenHeight = height;
}

bool Settings::IsCompilerOnly()
{
        return m_bIsCompilerOnly;
}

void Settings::SetCompilerOnly(bool bIsCompilerOnly)
{
        m_bIsCompilerOnly = bIsCompilerOnly;
}

void Settings::SetTextureCompression(bool bCompress)
{
    m_bCompressTextures = bCompress;
}

bool Settings::IsTextureCompression()
{
    return m_bCompressTextures;
}

Vector& Vector::operator = (const Vector& v)
{
    x = v.x;
    y = v.y;
    z = v.z;
    return *this;
}

float Vector::Normalize(float tolerance)
{
    float length = GetLength();

    if (length > tolerance)
    {
        float invLength = 1.0f / length;
        x *= invLength;
        y *= invLength;
        z *= invLength;
    }
    else
    {
        length = 0.0;
    }

    return length;
}

float Vector::GetLength()
{
    return sqrt(x*x + y*y + z*z);
}

Plane::Plane(Vector& p1,Vector& p2,Vector& p3)
{
    defineFromPoints(p1,p2,p3);
}

void Plane::defineFromPoints(Vector& p1,Vector& p2,Vector& p3)
{
    Vector v1,v2,normal;
// Find the normal of the polygon defined by the three points(cross product of 2 vertex vectors)
    v1.x = p2.x - p1.x;
    v1.y = p2.y - p1.y;
    v1.z = p2.z - p1.z;

    v2.x = p3.x - p1.x;
    v2.y = p3.y - p1.y;
    v2.z = p3.z - p1.z;

    v1.Normalize();
    v2.Normalize();
// Find surface normal based on cross product.
    normal.x = v1.y * v2.z - v2.y * v1.z;
    normal.y = v2.x * v1.z - v1.x * v2.z;
    normal.z = v1.x * v2.y - v2.x * v1.y;
    normal.Normalize();
// This surface normal represents the a,b,c components of the plane equation.
    a = normal.x;
    b = normal.y;
    c = normal.z;
// The d component is calculated from Ax + By + Cz + D = 0
    d = -(a * p1.x + b * p1.y + c * p1.z);
}

Triangle::Triangle()
{
}

Triangle::~Triangle()
{
}

void Triangle::DefineFromPoints(Vector& p1,Vector& p2,Vector& p3)
{
    m_pVertices[0].x = p1.x;
    m_pVertices[0].y = p1.y;
    m_pVertices[0].z = p1.z;
    m_pVertices[1].x = p2.x;
    m_pVertices[1].y = p2.y;
    m_pVertices[1].z = p2.z;
    m_pVertices[2].x = p3.x;
    m_pVertices[2].y = p3.y;
    m_pVertices[2].z = p3.z;
#ifdef _USE_RAYTRACING_SUPPORT_
    m_Plane.defineFromPoints(p3,p2,p1);
#endif
}

Vector* Triangle::GetVertex(int index)
{
    return &m_pVertices[index];
}

#ifdef _USE_RAYTRACING_SUPPORT_
Plane* Triangle::GetPlane()
{
    return &m_Plane;
}
#endif

char szElevName[1024];
char szTexName[1024];

#ifdef _USE_SDL_
TerrainLattice::TerrainLattice(char* szBaseName,char* szExtensionElev,char* szExtensionTex,char* szDetailTextureName,float vertexSpacing,float elevationScale,int maxNumTriangles,bool bUseBorders,int widthTerrains,int heightTerrains)
{
	m_szBaseName = new char[strlen(szBaseName)+1];
	sprintf(m_szBaseName,szBaseName);
	m_szExtensionElev = new char[strlen(szExtensionElev)+1];
	sprintf(m_szExtensionElev,szExtensionElev);
	m_szExtensionTex = new char[strlen(szExtensionTex)+1];
	sprintf(m_szExtensionTex,szExtensionTex);
	m_szDetailTextureName = new char[strlen(szDetailTextureName)+1];
	sprintf(m_szDetailTextureName,szDetailTextureName);
	m_VertexSpacing = vertexSpacing;
	m_ElevationScale = elevationScale;
	m_MaxNumTriangles = maxNumTriangles;
	m_bUseBorders = bUseBorders;
    m_WidthTerrains = widthTerrains;
    m_HeightTerrains = heightTerrains;
    m_pTerrains = new Terrain*[widthTerrains * heightTerrains];
    for (int i = 0; i < widthTerrains * heightTerrains; i++)
        m_pTerrains[i] = NULL;
    m_WidthActiveTerrains = m_HeightActiveTerrains = 1;

    sprintf(szElevName,"%sElev.%s.%ld-%ld",m_szBaseName,m_szExtensionElev,0,0);
    sprintf(szTexName,"%sTex.%s.%ld-%ld",m_szBaseName,m_szExtensionTex,0,0);
	Terrain* pTerrain = new Terrain(szElevName,szTexName,m_szDetailTextureName,m_VertexSpacing,m_ElevationScale,m_MaxNumTriangles,m_bUseBorders);
	m_TerrainWidth = pTerrain->GetWidth();
	m_TerrainHeight = pTerrain->GetHeight();
	delete pTerrain;
}

TerrainLattice::~TerrainLattice()
{
	delete[] m_szBaseName;
	delete[] m_szExtensionElev;
	delete[] m_szExtensionTex;
	delete[] m_szDetailTextureName;
    delete[] m_pTerrains;
}

void TerrainLattice::AddTerrain(Terrain* pTerrain,int positionX,int positionY)
{
    m_pTerrains[positionY * m_WidthTerrains + positionX] = pTerrain;
    pTerrain->SetLatticePosition(positionX,positionY);
}

Terrain* TerrainLattice::GetTerrain(int positionX,int positionY)
{
    return m_pTerrains[positionY * m_WidthTerrains + positionX];
}

Terrain* TerrainLattice::GetTerrainRelative(Terrain* pTerrain,Terrain::DIRECTION direction)
{
    int offsetX,offsetY;
    switch (direction)
    {
        case Terrain::DIR_NORTH:
            offsetX = 0;
            offsetY = 1;
            break;
        case Terrain::DIR_NORTHEAST:
            offsetX = 1;
            offsetY = 1;
            break;
        case Terrain::DIR_EAST:
            offsetX = 1;
            offsetY = 0;
            break;
        case Terrain::DIR_SOUTHEAST:
            offsetX = 1;
            offsetY = -1;
            break;
        case Terrain::DIR_SOUTH:
            offsetX = 0;
            offsetY = -1;
            break;
        case Terrain::DIR_SOUTHWEST:
            offsetX = -1;
            offsetY = -1;
            break;
        case Terrain::DIR_WEST:
            offsetX = -1;
            offsetY = 0;
            break;
        case Terrain::DIR_NORTHWEST:
            offsetX = -1;
            offsetY = 1;
            break;
    }
    return GetTerrainRelative(pTerrain,offsetX,offsetY);
}


Terrain* TerrainLattice::GetTerrainRelative(Terrain* pTerrain,int positionX,int positionY)
{
    Terrain* pRequestedTerrain = NULL;
    if (-1 <= positionX && positionX <= 1 && -1 <= positionY && positionY <= 1)
    {
        int x,y;
        pTerrain->GetLatticePosition(x,y);
        x += positionX;
        y += positionY;

        Terrain* pTerrainCenter = m_pTerrains[m_CurrentTerrainIndex[Terrain::DIR_CENTER]];
        int centerX,centerY;
        pTerrainCenter->GetLatticePosition(centerX,centerY);
        if (abs(x - centerX) <= m_WidthActiveTerrains && abs(y - centerY) <= m_HeightActiveTerrains)
        {
            if (0 <= x && x < m_WidthTerrains && 0 <= y && y < m_HeightTerrains)
                pRequestedTerrain = GetTerrain(x,y);
        }
    }
    return pRequestedTerrain;
}

Terrain* TerrainLattice::GetTerrainAtPoint(float x,float y)
{
    int indexX = (int)x / (int)m_TerrainWidth;
    int indexY = (int)y / (int)m_TerrainHeight;
    if (0 <= indexX && indexX < m_WidthTerrains && 0 <= indexY && indexY < m_HeightTerrains)
        return GetTerrain(indexX,indexY);
    else
        return NULL;
}

void TerrainLattice::SetCameraPosition(float x,float y,float z)
{
    int indexX = (int)x / (int)m_TerrainWidth;
    int indexY = (int)y / (int)m_TerrainHeight;
    m_CurrentTerrainIndex[Terrain::DIR_CENTER] = indexY * m_WidthTerrains + indexX;
    m_CurrentTerrainIndex[Terrain::DIR_SOUTH] = 0 < indexY ? m_CurrentTerrainIndex[Terrain::DIR_CENTER] - m_WidthTerrains : -1;
    m_CurrentTerrainIndex[Terrain::DIR_SOUTHEAST] = (0 < indexY && indexX < m_WidthTerrains - 1) ? m_CurrentTerrainIndex[Terrain::DIR_CENTER] - m_WidthTerrains + 1 : -1;
    m_CurrentTerrainIndex[Terrain::DIR_SOUTHWEST] = 0 < indexX && 0 < indexY ? m_CurrentTerrainIndex[Terrain::DIR_CENTER] - m_WidthTerrains - 1 : -1;
    m_CurrentTerrainIndex[Terrain::DIR_EAST] = indexX < m_WidthTerrains - 1 ? m_CurrentTerrainIndex[Terrain::DIR_CENTER] + 1 : -1;
    m_CurrentTerrainIndex[Terrain::DIR_WEST] = 0 < indexX ? m_CurrentTerrainIndex[Terrain::DIR_CENTER] - 1 : -1;
    m_CurrentTerrainIndex[Terrain::DIR_NORTH] = indexY < m_HeightTerrains - 1 ? m_CurrentTerrainIndex[Terrain::DIR_CENTER] + m_WidthTerrains : -1;
    m_CurrentTerrainIndex[Terrain::DIR_NORTHEAST] = (indexX < m_WidthTerrains - 1 && indexY < m_HeightTerrains - 1) ? m_CurrentTerrainIndex[Terrain::DIR_CENTER] + m_WidthTerrains + 1 : -1;
    m_CurrentTerrainIndex[Terrain::DIR_NORTHWEST] = (indexY < m_HeightTerrains - 1 && 0 < indexX) ? m_CurrentTerrainIndex[Terrain::DIR_CENTER] + m_WidthTerrains - 1 : -1;

    for (int i = 0; i < m_WidthTerrains * m_HeightTerrains; i++)
    {
        bool active = false;
        for (int dir = 0; dir < 9 && !active; dir++)
            active = (i == m_CurrentTerrainIndex[dir]);
        if (!active && m_pTerrains[i] != NULL)
        {
            vector<TerrainLoadListener*>::iterator iter = m_TerrainLoadListeners.begin();
            while (iter != m_TerrainLoadListeners.end())
            {
                TerrainLoadListener* pListener = *iter;
                pListener->TerrainUnloading(m_pTerrains[i]);
                iter++;
            }
            delete m_pTerrains[i];
            m_pTerrains[i] = NULL;
        }
        else if (active && m_pTerrains[i] == NULL)
        {
            LoadTerrain(i);
            vector<TerrainLoadListener*>::iterator iter = m_TerrainLoadListeners.begin();
            while (iter != m_TerrainLoadListeners.end())
            {
                TerrainLoadListener* pListener = *iter;
                pListener->TerrainLoaded(m_pTerrains[i]);
                iter++;
            }
        }
    }
}

void TerrainLattice::LoadTerrain(int index)
{
    int indexX = index % m_WidthTerrains;
    int indexY = index / m_WidthTerrains;
    sprintf(szElevName,"%sElev.%s.%ld-%ld",m_szBaseName,m_szExtensionElev,indexX,indexY);
    sprintf(szTexName,"%sTex.%s.%ld-%ld",m_szBaseName,m_szExtensionTex,indexX,indexY);
    
	Terrain* pTerrain = new Terrain(szElevName,szTexName,m_szDetailTextureName,m_VertexSpacing,m_ElevationScale,m_MaxNumTriangles,m_bUseBorders,(float)indexX * m_TerrainWidth,(float)indexY * m_TerrainHeight);
    if (m_TerrainWidth == 0.0f)
    {
        m_TerrainWidth = pTerrain->GetWidth();
        m_TerrainHeight = pTerrain->GetHeight();
    }
    AddTerrain(pTerrain,indexX,indexY);
}

void TerrainLattice::SetDetailThreshold(float threshold)
{
    for (int i = 0; i < 9; i++)
    {
        if (0 <= m_CurrentTerrainIndex[i])
            m_pTerrains[m_CurrentTerrainIndex[i]]->SetDetailThreshold(threshold);
    }
}

float TerrainLattice::GetElevation(float x,float y)
{
    Terrain* pTerrain = GetTerrainAtPoint(x,y);
    if (pTerrain != NULL)
        return pTerrain->GetElevation(x,y);
    else
        return 0.0f;
}

Terrain::DIRECTION TerrainLattice::GetOppositeDirection(Terrain::DIRECTION direction)
{
    Terrain::DIRECTION oppositeDirection;
    switch (direction)
    {
        case Terrain::DIR_NORTH:
            oppositeDirection = Terrain::DIR_SOUTH;
            break;
        case Terrain::DIR_NORTHEAST:
            oppositeDirection = Terrain::DIR_SOUTHWEST;
            break;
        case Terrain::DIR_EAST:
            oppositeDirection = Terrain::DIR_WEST;
            break;
        case Terrain::DIR_SOUTHEAST:
            oppositeDirection = Terrain::DIR_NORTHWEST;
            break;
        case Terrain::DIR_SOUTH:
            oppositeDirection = Terrain::DIR_NORTH;
            break;
        case Terrain::DIR_SOUTHWEST:
            oppositeDirection = Terrain::DIR_NORTHEAST;
            break;
        case Terrain::DIR_WEST:
            oppositeDirection = Terrain::DIR_EAST;
            break;
        case Terrain::DIR_NORTHWEST:
            oppositeDirection = Terrain::DIR_SOUTHEAST;
            break;
        case Terrain::DIR_CENTER:
            oppositeDirection = Terrain::DIR_CENTER;
            break;
    }
    return oppositeDirection;
}

void TerrainLattice::Tessellate()
{
    int i;
    for (i = 0; i < 9; i++)
    {
        if (m_CurrentTerrainIndex[i] != -1)
            m_pTerrains[m_CurrentTerrainIndex[i]]->Tessellate();
    }

    for (i = 0; i < 9; i++)
    {
        if (m_CurrentTerrainIndex[i] != -1)
        {
            for (int direction = 0; direction < 8; direction++)
            {
                if (direction != Terrain::DIR_CENTER)
                {
                    Terrain* pTerrain = GetTerrainRelative(m_pTerrains[m_CurrentTerrainIndex[i]],(Terrain::DIRECTION)direction);
                    if (pTerrain != NULL)
                    {
                        m_pTerrains[m_CurrentTerrainIndex[i]]->UpdateNeighbor(pTerrain,(Terrain::DIRECTION)direction);
                        pTerrain->UpdateNeighbor(m_pTerrains[m_CurrentTerrainIndex[i]],GetOppositeDirection((Terrain::DIRECTION)direction));
                    }
                }
            }
        }
    }

    for (i = 0; i < 9; i++)
    {
        if (m_CurrentTerrainIndex[i] != -1)
            m_pTerrains[m_CurrentTerrainIndex[i]]->m_pRootBlock->RepairCracks(m_pTerrains[m_CurrentTerrainIndex[i]],m_pTerrains[m_CurrentTerrainIndex[i]]->m_pTriangleStrips,m_pTerrains[m_CurrentTerrainIndex[i]]->m_pTriangleFans,&m_pTerrains[m_CurrentTerrainIndex[i]]->m_CountStrips,&m_pTerrains[m_CurrentTerrainIndex[i]]->m_CountFans);
    }

}

void TerrainLattice::Render()
{
    for (int i = 0; i < 9; i++)
    {
        if (m_CurrentTerrainIndex[i] != -1)
            m_pTerrains[m_CurrentTerrainIndex[i]]->Render();
    }
}

void TerrainLattice::AddTerrainLoadListener(TerrainLoadListener& listener)
{
    m_TerrainLoadListeners.push_back(&listener);
}

void TerrainLattice::RemoveTerrainLoadListener(TerrainLoadListener& listener)
{
    bool found = false;
    vector<TerrainLoadListener*>::iterator iter = m_TerrainLoadListeners.begin();
    while (iter != m_TerrainLoadListeners.end() && !found)
    {
        TerrainLoadListener* pListener = *iter;
        if (pListener == &listener)
        {
            m_TerrainLoadListeners.erase(iter);
            found = true;
        }
        else
            iter++;
    }
}

float TerrainLattice::GetWidth()
{
    return m_WidthTerrains * m_TerrainWidth;
}

float TerrainLattice::GetHeight()
{
    return m_HeightTerrains * m_TerrainHeight;
}

Texture::Texture(Uint8* pBuffer,int width,int height,int rowLength,int borderSize,bool bClamp,bool useCompression)
{
    m_pBuffer = new Uint8[height * width * 3];
    int k = 0;
    for (int i = 0; i < height; i++)
    {
        for (int j = 0; j < width * 3; j++,k++)
        {
            m_pBuffer[k] = pBuffer[i * rowLength * 3 + j];
        }
    }
    m_Width = width;
    m_Height = height;
    m_UseCompression = useCompression;
    m_BorderSize = borderSize;
    m_TextureID = 0;
    m_bClamp = bClamp;
}

Texture::~Texture()
{
	UnloadTexture();
    if (m_pBuffer != NULL)
    {
        delete[] m_pBuffer;
        m_pBuffer = NULL;
    }
}

GLuint Texture::UploadTexture()
{
    if (m_TextureID == 0)
    {
        int textureFormat;
        textureFormat = m_UseCompression ? COMPRESSED_RGB_S3TC_DXT1_EXT : GL_RGB4;
        m_TextureID = CreateTexture(m_pBuffer,m_Width,m_Height,m_Width,0,textureFormat,m_bClamp);
    }
    return m_TextureID;
}

void Texture::UnloadTexture()
{
    if (m_TextureID != 0)
    {
        glDeleteTextures(1,&m_TextureID);
        m_TextureID = 0;
    }
}

void DimensionPowerOf2(int origX,int origY,int& newX,int& newY)
{
	newX = newY = 0;
	for (int numPows = 20; newX == 0 || newY == 0; numPows *= 2)
	{
		int* powers = new int[numPows];
		int i;
		for (i = 0; i < numPows; i++)
			powers[i] = (int)pow(2,i);
		newX = 0;
		for (i = 0; i < numPows && newX == 0; i++)
		{
			if (origX <= powers[i])
				newX = powers[i];
		}
		newY = 0;
		for (i = 0; i < numPows && newY == 0; i++)
		{
			if (origY <= powers[i])
				newY = powers[i];
		}
		delete[] powers;
	}
}
#endif
