/*
 * This is a very simple demonstration of how to provide procedural textures to Demeter.
 * This sample texture factory simply creates texels based on the terrain elevation at each texel.
 * Colors become more yellow at higher elevations and more blue at lower elevations.
 * For more interesting procedural texture algorithms and ideas, see the Open Mountains website
 * (openmountains.sourceforge.net) for some good ideas and some good links.
 */

#include "SampleTextureFactory.h"

using namespace Demeter;

const float texelsPerMeter = 0.25f;
const float texelSpacing = 1.0f / texelsPerMeter;

GLuint CreateTex(Uint8* pTexels,int width,int height,int rowLength,int border,int internalFormat,bool bClamp,bool bColorKey);

SampleTextureFactory::SampleTextureFactory(Terrain* pTerrain)
{
	m_pTerrain = pTerrain;
}

GLuint SampleTextureFactory::GetTexture(int index,float originX,float originY,float width,float height)
{
	// The Demeter terrain calls this method when a texture has become visible and needs to be rendered.
	// Your implementation of TextureFactory must return a valid OpenGL texture, generated by whatever
	// means you like. Notice that a map of textures is kept as a cache so that subsequent requests for this
	// texture will be nearly instantaneous. You don't have to do this (e.g. for extremely large terrains
	// it wouldn't be practical), but you need some kind of strategy to keep your texture requests fast since
	// they happen at runtime.
	GLuint* pTexID = m_Textures[index];
	if (pTexID == NULL)
	{
		cout << "SAMPLETEXTUREFACTORY: width = " << width << " height = " << height << endl;
		cout << m_pTerrain->GetWidth() << "," << m_pTerrain->GetHeight() << endl;
		cout << "SAMPLETEXTUREFACTORY: Generating texture of size " << (int)(width * texelsPerMeter * height * texelsPerMeter) * 3 << " bytes" << endl;
		Uint8* pImage = new Uint8[(int)(width * texelsPerMeter * height * texelsPerMeter) * 3];
		int imageIndex = 0;
		for (float y = originY; y < originY + height; y += texelSpacing)
		{
			for (float x = originX; x < originX + width; x += texelSpacing,imageIndex+=3)
			{
				float elev = m_pTerrain->GetElevation(x,y);
				Uint8 red = (Uint8)(255.0f * elev / m_pTerrain->GetMaxElevation());
				Uint8 green = (Uint8)(255.0f * elev / m_pTerrain->GetMaxElevation());
				Uint8 blue = (Uint8)255 - (Uint8)(255.0f * elev / m_pTerrain->GetMaxElevation());
				pImage[imageIndex] = red;
				pImage[imageIndex + 1] = green;
				pImage[imageIndex + 2] = blue;
			}
		}
		pTexID = new GLuint;
		*pTexID = CreateTex(pImage,(int)(width * texelsPerMeter),(int)(height * texelsPerMeter),(int)(width * texelsPerMeter),0,GL_RGB4,true,false);
		delete[] pImage;
		m_Textures[index] = pTexID;	
	}
	return *pTexID;
}

void SampleTextureFactory::UnloadTexture(int index,float originX,float originY)
{
	// Called by the Demeter terrain when a texture is no longer visible.
	// We could unload the texture here, set it to a lower priority for OpenGL, etc.
	// Keep the texture for now ...
}

GLuint CreateTex(Uint8* pTexels,int width,int height,int rowLength,int border,int internalFormat,bool bClamp,bool bColorKey)
{
	// Simple helper function to create an OpenGL texture from a byte buffer. In implementing your own
	// TextureFactory classes, you can create textures any way you like.
    GLuint texId;
    glGenTextures(1,&texId);
    glBindTexture(GL_TEXTURE_2D,texId);
    if (bClamp)
    {
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
    }
    else
    {
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
    }
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);
    glPixelStorei(GL_UNPACK_ROW_LENGTH,rowLength);
    glPixelStorei(GL_UNPACK_ALIGNMENT,1);
    if (bColorKey)
    {
        glTexImage2D(GL_TEXTURE_2D,0,internalFormat,width + 2 * border,height + 2 * border,border,GL_RGBA,GL_UNSIGNED_BYTE,pTexels);
        gluBuild2DMipmaps(GL_TEXTURE_2D,GL_RGBA,width,height,GL_RGBA,GL_UNSIGNED_BYTE,pTexels);
    }
    else
    {
        glTexImage2D(GL_TEXTURE_2D,0,internalFormat,width + 2 * border,height + 2 * border,border,GL_RGB,GL_UNSIGNED_BYTE,pTexels);
        gluBuild2DMipmaps(GL_TEXTURE_2D,GL_RGB,width,height,GL_RGB,GL_UNSIGNED_BYTE,pTexels);
    }
    glPixelStorei(GL_UNPACK_ROW_LENGTH,0);
    return texId;
}


