// Demeter Terrain Visualization Library by Clay Fowler
// Copyright (C) 2001 Clay Fowler

/* 
IMPORTANT NOTE:
This file is not really a part of Demeter itself. This is simply a sample application that shows how to use Demeter.
As such, this file is written to be easy to follow, not to be a robust application. 

See the file mainOSG.cpp for a slightly more robust (and more interesting) application example.
*/

/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.
*/

#include "Terrain.h"
#include <new.h>
#include <iostream>
#include "Camera.h"
#include "matrix.h"
#include "SampleTextureFactory.h"

// It is silly to limit the frame rate in a real application. See the OSG Sample Application for a much
// better way of handling this issue.
#define _LIMIT_FRAME_RATE_ // Causes rendering rate to be clamped down to a maximum rate - due to coarse timers, this actually causes the framerate to be very low on some systems. See the OSG Sample Application for a much better way of handling this issue.

const float PI = 3.14159f;
const float PI_OVER_TWO = 3.1459f * 0.5f;
const float CAMERA_SPEED = 6.75f; 
const float CAMERA_ROTATE_SPEED = 0.09f; 
const float CAMERA_BORDER_SIZE = 500.0f; 
const float CAMERA_MIN_ALTITUDE = 5.0f;
const float CAMERA_MAX_ALTITUDE = 100.0f;
const int TERRAIN_ANIMATION_PERIOD = 300; 
const float MAX_VIEW_DISTANCE = 4500.0f;
const int MAX_NUM_PARAMETERS = 16;
const int MAX_PARAM_LENGTH = 2048;
const float FOG_RED = 0.5f;
const float FOG_GREEN = 0.75f;
const float FOG_BLUE = 1.0f;
const float FOG_ALPHA = 0.0f;

#ifdef _LIMIT_FRAME_RATE_
	const Uint32 CYCLE_PERIOD = 30; // milliseconds (33 = approx. 30 frames per second)
#endif

using namespace Demeter;

void ShowUsage();
void ToggleWireframe(char* szMsg);
void GetParam(char* szCommand,int paramNum,char* szParamOut);
void WriteLog(char* szMessage);

Settings* pSettings;
bool bWireframe = false;

#ifdef _USE_RAYTRACING_SUPPORT_
class Marker
{
public:
                        Marker();
                        ~Marker();
    void                Render();
    void	 			SetPosition(float x,float y,float z);
private:
	Vector				m_Position;
    GLUquadric*			m_pQuadric;
    float				m_Radius;
};
const float MARKER_RADIUS = 10.0f;
#endif

Terrain* pTerrain = NULL;

#ifdef _WIN32
int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPSTR lpCmdLine,int nCmdShow)
#else
int main(int argc,char** argv)
#endif
{
	char* szElevFilename = NULL;
	char* szTextureFilename = NULL;
	char* szDetailTextureFilename = NULL;
	char* szElevScale = NULL;
	char* szVertexSpacing = NULL;
	
	bool bFullscreen = false;
	bool bUseBorders = false;
	bool bUseTriangleStrips = true;
	bool bUseLowRes = false;
	bool bUseProcedural = false;
    int ScreenWidth;
    int ScreenHeight;
	Vector lookAt;
	Vector lookUp;
	Vector cameraAngle;
#ifdef _USE_RAYTRACING_SUPPORT_
    Marker marker;
#endif
	float threshold = 12.0f; // A high threshold to compensate for the naive approach toward limiting frame rate.
	bool continueRunning = true;
	bool firstTime = true;
	bool bVerboseCycle = false;
	Uint32 totalRenderTime = 0;
	int cycles = 0;
    char szMsg[128];

// Parse the command line
#ifdef _WIN32
	int argc = 1;
	char** argv;
	argv = new char*[16];
	for (int win32cmd = 0; win32cmd < 16; win32cmd++)
		argv[win32cmd] = new char[512];
	sprintf(argv[0],"Viewer.exe");
	char szParam[512];
	do
	{
		GetParam(lpCmdLine,argc,szParam);
		if (0 < strlen(szParam))
		{
			sprintf(argv[argc],szParam);
			argc++;
		}
	}
	while (0 < strlen(szParam));
#endif

	try
	{
		if (argc < 7)
			ShowUsage();
		else
		{
			pSettings = Settings::GetInstance();

			if (8 <= argc)
			{
				int arg;
				for (arg = 1; arg < argc - 6; arg++)
				{
					if (strcmp(argv[arg],"-v") == 0)
						pSettings->SetVerbose(true);
					else if (strcmp(argv[arg],"-f") == 0)
						bFullscreen = true;
					else if (strcmp(argv[arg],"-b") == 0)
						bUseBorders = true;
					else if (strcmp(argv[arg],"-s") == 0)
						bUseTriangleStrips = false;
					else if (strcmp(argv[arg],"-l") == 0)
						bUseLowRes = true;
					else if (strcmp(argv[arg],"-p") == 0)
						bUseProcedural = true;
				}
				szElevFilename = argv[arg];
				szTextureFilename = argv[arg + 1];
				szDetailTextureFilename = argv[arg + 2];
				szVertexSpacing = argv[arg + 3];
				szElevScale = argv[arg + 4];
				pSettings->SetMediaPath(argv[arg + 5]);
			}
			else
			{
				szElevFilename = argv[1];
				szTextureFilename = argv[2];
				szDetailTextureFilename = argv[3];
				szVertexSpacing = argv[4];
				szElevScale = argv[5];
				pSettings->SetMediaPath(argv[6]);
			}
			if (bUseLowRes)
			{
				ScreenWidth = 640;
				ScreenHeight = 480;
			}
			else
			{
				ScreenWidth = 800;
				ScreenHeight = 600;
			}
			pSettings->SetScreenWidth(ScreenWidth);
			pSettings->SetScreenHeight(ScreenHeight);
		// Initialize the rendering surface
			if (pSettings->IsVerbose())
				cout << "Initializing video" << endl;
			if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0)
			{
				printf("Unable to initialize SDL: %s\n",SDL_GetError());
				exit(1);
			}
			atexit(SDL_Quit);
			SDL_GL_SetAttribute(SDL_GL_BUFFER_SIZE,32);
			SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,1);
			SDL_EnableUNICODE(1);
			if (pSettings->IsVerbose())
				printf("Setting video mode (OpenGL)\n");
			Uint32 screenFlags = SDL_OPENGL;
			if (bFullscreen)
				screenFlags |= SDL_FULLSCREEN;
			SDL_Surface* pScreen = SDL_SetVideoMode(pSettings->GetScreenWidth(),pSettings->GetScreenHeight(),32,screenFlags);
			if (pScreen == NULL)
			{
				cout << "Unable to set screen resolution" << endl;
				exit(2);
			}
			SDL_WM_SetCaption("Demeter","testgl");
			if (pSettings->IsVerbose())
				cout << "VIEWER: OpenGL vendor: " << glGetString(GL_VENDOR) << endl;
			if (pSettings->IsVerbose())
				cout << "VIEWER: OpenGL extensions supported: " << glGetString(GL_EXTENSIONS) << endl;	
		// Create the terrain
			try
			{
				const int MAX_NUM_VISIBLE_TRIANGLES = 40000; // Chosen based on the expected number of triangles that will be visible on-screen at any one time (the terrain mesh will typically have far more triangles than are seen at one time, especially with dynamic tessellation)
				if (bUseProcedural)
				{
					pTerrain = new Terrain(szElevFilename,NULL,szDetailTextureFilename,atof(szVertexSpacing),atof(szElevScale),MAX_NUM_VISIBLE_TRIANGLES,bUseBorders,0.0f,0.0f,100,100);
					pTerrain->SetTextureFactory(new SampleTextureFactory(pTerrain));
				}
				else
					pTerrain = new Terrain(szElevFilename,szTextureFilename,szDetailTextureFilename,atof(szVertexSpacing),atof(szElevScale),MAX_NUM_VISIBLE_TRIANGLES,bUseBorders);
			}
			catch(DemeterException* pEx)
			{
				cout << pEx->GetErrorMessage() << endl;
				exit(0);
			}
			catch(...)
			{
				cout << "VIEWER: Unexpected exception while creating terrain" << endl;
				exit(0);
			}
			pTerrain->SetMaximumVisibleBlockSize(64);
			pTerrain->SetCommonTextureRepeats(50.0f);
		// Setup 3D scene
		// Vanilla OpenGL setup
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glViewport(0,0,pSettings->GetScreenWidth(),pSettings->GetScreenHeight());
			glFrustum(-0.5f,0.5f,-0.5f,0.5f,0.65f,MAX_VIEW_DISTANCE);
			glClearColor(0.5f,0.75f,1.0f,0.0f);
			glDisable(GL_NORMALIZE);
			glEnable(GL_DEPTH_TEST);
			glShadeModel(GL_SMOOTH);
		// Setup camera
			Vector cameraPosition;
			cameraPosition.x = pTerrain->GetWidth() / 2.0f + 400.0f;
			cameraPosition.y = pTerrain->GetHeight() / 2.0f - 351.0f;
			cameraPosition.z = 20.0f;
			Camera camera;
			camera.SetPosition(cameraPosition.x,cameraPosition.y,cameraPosition.z);			
			cameraAngle.x = 0.0f;
			cameraAngle.y = cameraAngle.z = 0.0f;
		// Setup fog - it is highly recommended that apps ALWAYS use fog for realistic terrains...
			float fogColor[4];
			fogColor[0] = FOG_RED;
			fogColor[1] = FOG_GREEN;
			fogColor[2] = FOG_BLUE;
			fogColor[3] = FOG_ALPHA;
			glEnable(GL_FOG);
			glFogf(GL_FOG_MODE,GL_LINEAR);
			glFogfv(GL_FOG_COLOR,fogColor);
			glFogf(GL_FOG_START,100.0f);
			glFogf(GL_FOG_END,MAX_VIEW_DISTANCE - 100.0f);
			glHint(GL_FOG_HINT,GL_FASTEST);
		
		// Main control and rendering loop
//			SDL_WM_GrabInput(SDL_GRAB_ON);
			Uint32 startTime = SDL_GetTicks();
			Uint32 wireframeToggledTime = startTime;
			Uint32 numTrianglesRendered = 0;
			while (continueRunning)
			{
	#ifdef _LIMIT_FRAME_RATE_
				Uint32 cycleStartTime = SDL_GetTicks();
	#endif
			// Setup viewing transformation
				Matrix rotateX,rotateZ,cameraTransform; // TBD: Quick and dirty viewing system - for a real app something more robust would probably be required
				rotateX.SetRotateX(cameraAngle.x);
				rotateZ.SetRotateZ(cameraAngle.z);
				cameraTransform = rotateX * rotateZ;
				lookAt.x = 0.0f;
				lookAt.y = 100000.0f; // Look at a suitably (and arbitrarily) distant point - if your terrain's bigger than this then increase this value
				lookAt.z = 0.0f;
				lookUp.x = 0.0f;
				lookUp.y = 0.5f;
				lookUp.z = 1.0f;
				lookUp.Normalize();
				Vector currentLookAt = cameraTransform * lookAt;
				Vector currentLookUp = cameraTransform * lookUp;
				camera.SetLookAt(cameraPosition.x + currentLookAt.x,cameraPosition.y + currentLookAt.y,cameraPosition.z + currentLookAt.z);
				camera.SetLookUp(currentLookUp.x,currentLookUp.y,currentLookUp.z);
				float currentCameraElevation = pTerrain->GetElevation(cameraPosition.x,cameraPosition.y) + cameraPosition.z;
				camera.SetPosition(cameraPosition.x,cameraPosition.y,currentCameraElevation);
				camera.UpdateViewTransform();
			// Render the terrain
				pTerrain->SetDetailThreshold(threshold);
				numTrianglesRendered += pTerrain->ModelViewMatrixChanged();
				glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
				pTerrain->Render();
	#ifdef _USE_RAYTRACING_SUPPORT_
			// Render everything else in the scene
				marker.Render();
	#endif
			// Swap the double buffer so we can see what we just rendered
				SDL_GL_SwapBuffers();
			// Handle mouse camera commands
				int mouseX,mouseY;
				SDL_GetMouseState(&mouseX,&mouseY);
				if (mouseX == 0)
					cameraAngle.z += CAMERA_ROTATE_SPEED;
				if (mouseX == pSettings->GetScreenWidth() - 1)
					cameraAngle.z -= CAMERA_ROTATE_SPEED;
				if (mouseY == 0)
				{
					Vector lookAt2D = currentLookAt;
					lookAt2D.z = 0.0f;
					lookAt2D.Normalize();
					cameraPosition.x += lookAt2D.x * CAMERA_SPEED;
					cameraPosition.y += lookAt2D.y * CAMERA_SPEED;
					cameraPosition.z += lookAt2D.z * CAMERA_SPEED;
				}
				if (mouseY == pSettings->GetScreenHeight() - 1)
				{
					Vector lookAt2D = currentLookAt;
					lookAt2D.z = 0.0f;
					lookAt2D.Normalize();
					cameraPosition.x += lookAt2D.x * -CAMERA_SPEED;
					cameraPosition.y += lookAt2D.y * -CAMERA_SPEED;
					cameraPosition.z += lookAt2D.z * -CAMERA_SPEED;
				}
			// Handle all other input
				SDL_PumpEvents();
				SDL_Event event;
				while(SDL_PollEvent(&event))
				{
					if (event.type == SDL_QUIT)
						continueRunning = false;			
	#ifdef _USE_RAYTRACING_SUPPORT_
					else if (event.type == SDL_MOUSEBUTTONDOWN)
					{
					// When the mouse button is clicked, do an intersection test with the terrain to demonstrate
					// mouse picking. A simple "marker" object will be placed at the clicked point. This is simply
					// to demonstrate that applications can find out where on the terrain a user is clicking.
						if (event.button.button == 1)
						{
							Vector direction;
							direction.x = ((float)event.button.x / (float)ScreenWidth) - 0.5f;
							direction.z = ((float)event.button.y / (float)ScreenHeight) - 0.5f;
							direction.y = 1.0f;
							direction.z *= -1.0f;
							direction.Normalize();
							direction = cameraTransform * direction;
							Vector intersect;
							pTerrain->IntersectRay(cameraPosition.x,cameraPosition.y,currentCameraElevation,direction.x,direction.y,direction.z,intersect.x,intersect.y,intersect.z);
							marker.SetPosition(intersect.x,intersect.y,intersect.z + MARKER_RADIUS);
						}
					}
	#endif
				}

				int numKeys;
				Uint8* pKeys = SDL_GetKeyState(&numKeys);
				// Lots of redundant code that would need to be modularized for a real app in here...
				if (pKeys[SDLK_t])
				{
					if (1000 < SDL_GetTicks() - wireframeToggledTime)
					{
						ToggleWireframe(NULL);
						wireframeToggledTime = SDL_GetTicks();
					}
				}
				if (pKeys[SDLK_w])
				{
				// Move the camera forward
					Vector lookAt2D = currentLookAt;
					lookAt2D.z = 0.0f;
					lookAt2D.Normalize();
					cameraPosition.x += lookAt2D.x * CAMERA_SPEED;
					cameraPosition.y += lookAt2D.y * CAMERA_SPEED;
				}
				if (pKeys[SDLK_x])
				{
				// Move the camera backward
					Vector lookAt2D = currentLookAt;
					lookAt2D.z = 0.0f;
					lookAt2D.Normalize();
					cameraPosition.x += -lookAt2D.x * CAMERA_SPEED;
					cameraPosition.y += -lookAt2D.y * CAMERA_SPEED;
				}
				if (pKeys[SDLK_a])
				{
				// Move the camera left
					Vector lookAt2D = currentLookAt;
					lookAt2D.z = 0.0f;
					lookAt2D.Normalize();
					Matrix mat;
					mat.SetRotateZ(PI_OVER_TWO);
					lookAt2D = mat * lookAt2D;
					cameraPosition.x += lookAt2D.x * CAMERA_SPEED;
					cameraPosition.y += lookAt2D.y * CAMERA_SPEED;
				}
				if (pKeys[SDLK_d])
				{
				// Move the camera right
					Vector lookAt2D = currentLookAt;
					lookAt2D.z = 0.0f;
					lookAt2D.Normalize();
					Matrix mat;
					mat.SetRotateZ(-PI_OVER_TWO);
					lookAt2D = mat * lookAt2D;
					cameraPosition.x += lookAt2D.x * CAMERA_SPEED;
					cameraPosition.y += lookAt2D.y * CAMERA_SPEED;
				}
				if (pKeys[SDLK_q])
				// Move the camera up
					cameraPosition.z += CAMERA_SPEED;
				if (pKeys[SDLK_z])
				// Move the camera down
					cameraPosition.z -= CAMERA_SPEED;
				if (pKeys[SDLK_ESCAPE])
				// Quit the application
					continueRunning = false;

				if (pKeys[SDLK_KP_PLUS])
				{
				// Increase the detail threshold (make the terrain less detailed - worse looking but faster)
					threshold += 0.5f;
					sprintf(szMsg,"VIEWER: Threshold = %.1f",threshold);
					WriteLog(szMsg);
				}
				if (pKeys[SDLK_KP_MINUS])
				{
				// Decrease the detail threshold (make the terrain more detailed - better looking but slower)
					threshold -= 0.5f;
					sprintf(szMsg,"VIEWER: Threshold = %.1f",threshold);
					WriteLog(szMsg);
				}

			// Clamp all control values to legal limits
				if (threshold < 0.0f)
					threshold = 0.0f;
				if (cameraPosition.x < CAMERA_BORDER_SIZE)
					cameraPosition.x = CAMERA_BORDER_SIZE;
				if (pTerrain->GetWidth() - CAMERA_BORDER_SIZE < cameraPosition.x)
					cameraPosition.x = pTerrain->GetWidth() - CAMERA_BORDER_SIZE;
				if (cameraPosition.y < CAMERA_BORDER_SIZE)
					cameraPosition.y = CAMERA_BORDER_SIZE;
				if (pTerrain->GetHeight() - CAMERA_BORDER_SIZE < cameraPosition.y)
					cameraPosition.y = pTerrain->GetHeight() - CAMERA_BORDER_SIZE;
				if (cameraPosition.z < CAMERA_MIN_ALTITUDE)
					cameraPosition.z = CAMERA_MIN_ALTITUDE;
				if (CAMERA_MAX_ALTITUDE < cameraPosition.z)
					cameraPosition.z = CAMERA_MAX_ALTITUDE;

				Uint32 elapsedTime = SDL_GetTicks() - startTime;
				if (10000 < elapsedTime)
				{
					if (pSettings->IsVerbose())
					{
						float speed = (float)cycles / ((float)elapsedTime / 1000.0f);
						sprintf(szMsg,"VIEWER: %.3f frames per second",speed);
						WriteLog(szMsg);
	#ifdef _LIMIT_FRAME_RATE_
						sprintf(szMsg,"**Frame rate clamped - compile without _LIMIT_FRAME_RATE_ for max speed");
						WriteLog(szMsg);
	#endif
					}
					cycles = 0;
					startTime = SDL_GetTicks();
				}
	#ifdef _LIMIT_FRAME_RATE_
			// Clamp to the maximum cycle rate. See the OSG sample application for a much better approach to this problem.
				Uint32 cycleTime = SDL_GetTicks() - cycleStartTime;
				if (cycleTime < CYCLE_PERIOD)
					SDL_Delay(CYCLE_PERIOD - cycleTime);
	#endif
				cycles++;
			}	
		}	
		SDL_WM_GrabInput(SDL_GRAB_OFF);
		SDL_Quit();
	}
	catch(DemeterException* pEx)
	{
		cout << pEx->GetErrorMessage() << endl;
		SDL_WM_GrabInput(SDL_GRAB_OFF);
		SDL_Quit();
	}
	catch(...)
	{
		SDL_WM_GrabInput(SDL_GRAB_OFF);
		SDL_Quit();
	}
	return 1;
}

void ShowUsage()
{
	cout << "Viewer [-w | -v | -b | -f | -s | -l | -p] <elevFilename> <textureFilename> <detailTextureFilename> <vertexSpacing> <elevScale> <media path>" << endl;
	cout << "\tOptions: " << endl;
	cout << "\tp = use sample procedural texture algorithm" << endl; 		
	cout << "\tb = use texture borders to eliminate seams on terrain" << endl; 		
	cout << "\tw = render terrain as wireframe rather than solid polygons" << endl; 		
	cout << "\tv = verbose output (diagnostic)" << endl;
	cout << "\tf = run fullscreen" << endl; 		
	cout << "\ts = force use of triangle soup instead of triangle strips (and fans)" << endl;
	cout << "\tl = use low resolution video (640x480)" << endl;
}

void ToggleWireframe(char* szMsg)
{
	bWireframe = !bWireframe;
	if (bWireframe)
		glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
	else
		glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
}

void GetParam( char* szCommand,int paramNum,char* szParamOut )
{
	char* szParam = szCommand;

	int numParamsSkipped = 0;

	for (int i = 0; i < paramNum; i++)
	{
		char* szWrite = szParamOut;
		while(*szParam == ' ') szParam++;
		int length = 0;
		while (*szParam != '\0' && *szParam != ' ' && ++length < MAX_PARAM_LENGTH)
			*szWrite++ = *szParam++;
		*szWrite = '\0';
	}
}

void WriteLog(char* szMessage)
{
	cout << szMessage << endl;
}

#ifdef _USE_RAYTRACING_SUPPORT_
Marker::Marker()
{
    m_pQuadric = gluNewQuadric();
    m_Radius = MARKER_RADIUS;
}

Marker::~Marker()
{
    gluDeleteQuadric(m_pQuadric);
}

void Marker::Render()
{
	glFrontFace(GL_CCW);

	float mat[3] = {1.0f,1.0f,1.0f};
	glMaterialfv(GL_FRONT,GL_AMBIENT,mat);
	glMaterialfv(GL_FRONT,GL_DIFFUSE,mat);
	glMaterialfv(GL_FRONT,GL_SPECULAR,mat);
	glMaterialf(GL_FRONT,GL_SHININESS,0.0f);

    pTerrain->DisableTextures();

    glPushMatrix();
    glTranslatef(m_Position.x,m_Position.y,m_Position.z);
    gluSphere(m_pQuadric,m_Radius,16,16);
    glPopMatrix();

    pTerrain->EnableTextures();
}

void Marker::SetPosition(float x,float y,float z)
{
    m_Position.x = x;
    m_Position.y = y;
    m_Position.z = z;
}
#endif
