// Demeter Terrain Visualization Library by Clay Fowler
// Copyright (C) 2001 Clay Fowler

/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.
*/

#include "Terrain.h"
#include <new.h>
#include <iostream>
#include <osg/Geode>
#include <osg/GeoSet>
#include <osg/Material>
#include <osg/Vec3>
#include <osg/Transform>
#include <osg/Texture>
#include <osg/TexEnv>
#include <osg/Transparency>
#include <osg/AlphaFunc>
#include <osg/Billboard>
#include <osg/Fog>
#include <osgDB/Registry>
#include <osgDB/ReadFile>
#include <osgDB/FileUtils>
#include <osgUtil/SceneView>
#include <osgUtil/DisplayListVisitor>
#include <math.h>
#include "DemeterDrawable.h"
#include "HoverCamera.h"

using namespace Demeter;
using namespace osg;
using namespace osgUtil;

Geode* createCube();
void CreateTrees(Terrain* pTerrain,Group* pParentGroup);
GeoSet* CreateTree(float width,float height);
Node* CreateTerrainNode(Terrain* pTerrain);

const float CAMERA_SPEED = 100.0f; // Meters per second
const float CAMERA_ROT_SPEED = 72.0f; // Degrees per second
const float MAX_VIEW_DISTANCE = 5000.0f;

const float SKY_RED = 0.5f;
const float SKY_GREEN = 0.75f;
const float SKY_BLUE = 1.0f;
const float SKY_ALPHA = 1.0f;

#ifdef _WIN32
int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPSTR lpCmdLine,int nCmdShow)
#else
int main(int argc, char* argv[])
#endif
{
	int ScreenWidth = 800;
	int ScreenHeight = 600;
	float detailThreshold = 9.0f;
	
	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0)
	{
		printf("MAIN: Unable to initialize SDL: %s\n",SDL_GetError());
		exit(1);
	}
	atexit(SDL_Quit);

	try
	{
		SDL_GL_SetAttribute(SDL_GL_BUFFER_SIZE,32);
		SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,1);
		SDL_GL_SetAttribute(SDL_GL_RED_SIZE,5);
		SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE,5);
		SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,5);
		SDL_EnableUNICODE(1);
		Uint32 screenFlags = SDL_OPENGL | SDL_OPENGLBLIT | SDL_FULLSCREEN;
		SDL_Surface* pScreen = SDL_SetVideoMode(ScreenWidth,ScreenHeight,32,screenFlags);
		if (pScreen == NULL)
		{
			cout << "MAIN: Unable to set screen resolution" << endl;
			exit(2);
		}
		cout << "MAIN: Screen is " << (int)(pScreen->format->BytesPerPixel) << " bytes per pixel" << endl;
			
		SDL_WM_SetCaption("Demeter","testgl");

		SceneView* pView = new SceneView();
		pView->setDefaults();
		pView->setViewport(0,0,ScreenWidth,ScreenHeight);
		Vec4 bkgColor;
		bkgColor[0] = SKY_RED;
		bkgColor[1] = SKY_GREEN;
		bkgColor[2] = SKY_BLUE;
		bkgColor[3] = SKY_ALPHA;
		pView->setBackgroundColor(bkgColor);
		pView->setCalcNearFar(false);
		
	#ifdef _WIN32
		char fileSeparator = '\\';
	#else
		char fileSeparator = '/';
	#endif
		char szMediaPath[9];
		sprintf(szMediaPath,"maps%c",fileSeparator,fileSeparator);
		Settings::GetInstance()->SetMediaPath(szMediaPath);
		osgDB::Registry::instance();
		osgDB::setFilePath(szMediaPath);

		Group* pRootNode = new Group;

		Settings::GetInstance()->SetVerbose(true);
		Settings::GetInstance()->SetScreenWidth(800);
		Settings::GetInstance()->SetScreenHeight(600);
		const int MAX_NUM_VISIBLE_TRIANGLES = 80000; // Chosen based on the expected number of triangles that will be visible on-screen at any one time (the terrain mesh will typically have far more triangles than are seen at one time, especially with dynamic tessellation)
		Terrain* pTerrain = new Terrain("LlanoElev.jpg","LlanoTex.jpg","dirt.jpg",20.0f,2.0f,MAX_NUM_VISIBLE_TRIANGLES,true);
		pTerrain->SetDetailThreshold(detailThreshold);
		Node* pTerrainNode = CreateTerrainNode(pTerrain);
		StateSet* pState = new StateSet;
		pState->setMode(GL_LIGHTING,StateAttribute::OFF);
		pTerrainNode->setStateSet(pState);

		pState = new StateSet;
		Fog* fog = new Fog;
		fog->setMode(Fog::LINEAR);
		fog->setDensity(0.5f);
		fog->setStart(100.0f);
		fog->setEnd(MAX_VIEW_DISTANCE - 100.0f);
		fog->setColor(bkgColor);
		pState->setAttributeAndModes(fog,StateAttribute::ON);
		pRootNode->setStateSet(pState);

		CreateTrees(pTerrain,pRootNode);
			
		pRootNode->addChild(pTerrainNode);
					
		pView->setSceneData(pRootNode);

		Matrix identityMat;
		
		float currentAngle = 0.0f;
		HoverCamera camera(pTerrain,pView->getCamera(),MAX_VIEW_DISTANCE,(float)Settings::GetInstance()->GetScreenWidth() / (float)Settings::GetInstance()->GetScreenHeight());
		camera.SetHoverElevation(25.0f);
		camera.SetPosition(pTerrain->GetWidth() / 2.0f + 400.0f,pTerrain->GetHeight() / 2.0f - 100.0f);

		char szMsg[128];
		Uint32 startPeriodTime = SDL_GetTicks();
		float fps,timePerFrame = 0.01f;
		int cycles = 0;
		int showFPSCyclePeriod = 0;
	//	SDL_WM_GrabInput(SDL_GRAB_ON);
		bool continueRunning = true;
		while (continueRunning)
		{
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			pView->cull();
			pView->draw();
			glFinish();
			SDL_GL_SwapBuffers();

			// Adjust simulation velocities based on the user's computer speed, so we have real velocities in units/second (independent of rendering frame rate.)
			float distance = CAMERA_SPEED * timePerFrame;
			float angle = CAMERA_ROT_SPEED * timePerFrame;

			int mouseX,mouseY;
			SDL_GetMouseState(&mouseX,&mouseY);
			if (mouseX == 0)
				camera.AddYaw(-angle);
			if (mouseX == ScreenWidth - 1)
				camera.AddYaw(angle);
			if (mouseY == 0)
				camera.TranslateForward(distance);
			if (mouseY == ScreenHeight - 1)
				camera.TranslateBackward(distance);
			SDL_PumpEvents();
			SDL_Event event;
			while(SDL_PollEvent(&event))
			{
				if (event.type == SDL_QUIT)
					continueRunning = false;
			}
			int numKeys;
			Uint8* pKeys = SDL_GetKeyState(&numKeys);
			if (pKeys[SDLK_t]);
			if (pKeys[SDLK_w])
				camera.TranslateForward(distance);
			if (pKeys[SDLK_x])
				camera.TranslateBackward(distance);
			if (pKeys[SDLK_a])
				camera.TranslateLeft(distance);
			if (pKeys[SDLK_d])
				camera.TranslateRight(distance);
			if (pKeys[SDLK_q])
				camera.SetHoverElevation(camera.GetHoverElevation() + distance / 10.0f);
			if (pKeys[SDLK_z])
				camera.SetHoverElevation(camera.GetHoverElevation() - distance / 10.0f);
			if (pKeys[SDLK_ESCAPE])
				continueRunning = false;
			if (pKeys[SDLK_n])
			{
				detailThreshold += 0.1f;
				pTerrain->SetDetailThreshold(detailThreshold);
				cout << detailThreshold << endl;
			}
			if (pKeys[SDLK_m])
			{
				if (0.1f < detailThreshold)
				{
					detailThreshold -= 0.1f;
					pTerrain->SetDetailThreshold(detailThreshold);
					cout << detailThreshold << endl;
				}
			}
			if (pKeys[SDLK_KP_PLUS]);
			if (pKeys[SDLK_KP_MINUS]);

			if (cycles++ == 50)
			{
				Uint32 now = SDL_GetTicks();
				Uint32 elapsedTime = now - startPeriodTime;
				fps = (float)cycles / ((float)elapsedTime / 1000.0f);
				if (Settings::GetInstance()->IsVerbose() && ++showFPSCyclePeriod == 10)
				{
					cout << "MAIN: " << fps << " frames per second" << endl;
					showFPSCyclePeriod = 0;	
				}
				timePerFrame = 1.0f / fps;
				startPeriodTime = now;
				cycles = 0;
			}
		}	
	//	SDL_WM_GrabInput(SDL_GRAB_OFF);
		SDL_Quit();
		return 1;
	}
	catch(DemeterException* pEx)
	{
		cout << pEx->GetErrorMessage() << endl;
		SDL_Quit();
		exit(0);
	}
	catch(...)
	{
		cout << "MAINOSG: Unexpected exception" << endl;
		SDL_Quit();
		exit(0);
	}
}

Node* CreateTerrainNode(Terrain* pTerrain)
{
	Geode* pGeode = NULL;

	try
	{
		DemeterDrawable* pDrawable = new DemeterDrawable;
		pDrawable->SetTerrain(pTerrain);
		pGeode = new Geode;
		pGeode->addDrawable(pDrawable);
	}
	catch(...)
	{
	}
	
	return pGeode;
}

GeoSet* CreateTree(float width,float height)
{
	osg::Texture* pTex = new osg::Texture;
	Image* pImage = osgDB::readImageFile("tree0.rgba");
    pTex->setImage(pImage);

    StateSet *dstate = new StateSet;
    dstate->setAttributeAndModes(pTex,StateAttribute::ON);
    dstate->setAttribute(new TexEnv);
    dstate->setAttributeAndModes(new Transparency, StateAttribute::ON);
    AlphaFunc* alphaFunc = new AlphaFunc;
    alphaFunc->setFunction(AlphaFunc::GEQUAL,0.05f);
    dstate->setAttributeAndModes(alphaFunc, StateAttribute::ON);
    dstate->setMode(GL_LIGHTING, StateAttribute::OFF);
    dstate->setRenderingHint(StateSet::TRANSPARENT_BIN);

	Vec3* pVertices = new Vec3[4];
	pVertices[0].x() = -0.5f * width;
	pVertices[0].y() =  0.0f;
	pVertices[0].z() =  0.0f;
	pVertices[1].x() =  0.5f * width;
	pVertices[1].y() =  0.0f;
	pVertices[1].z() =  0.0f;
	pVertices[2].x() =  0.5f * width;
	pVertices[2].y() =  0.0f;
	pVertices[2].z() =  height;
	pVertices[3].x() = -0.5f * width;
	pVertices[3].y() =  0.0f;
	pVertices[3].z() =  height;
	
	Vec2* pTexCoords = new Vec2[4];
	pTexCoords[0].x() = 0.0f;
	pTexCoords[0].y() = 0.0f;
	pTexCoords[1].x() = 1.0f;
	pTexCoords[1].y() = 0.0f;
	pTexCoords[2].x() = 1.0f;
	pTexCoords[2].y() = 1.0f;
	pTexCoords[3].x() = 0.0f;
	pTexCoords[3].y() = 1.0f;
	
	Vec4* color = new Vec4(1.0f,1.0f,1.0f,1.0f);
	
    GeoSet *pSet = new GeoSet;
    pSet->setCoords(pVertices);
    pSet->setTextureCoords(pTexCoords);
    pSet->setTextureBinding(GeoSet::BIND_PERVERTEX);
    pSet->setColors(color);
    pSet->setColorBinding(GeoSet::BIND_OVERALL);
    pSet->setPrimType(GeoSet::QUADS);
    pSet->setNumPrims(1);
    pSet->setStateSet(dstate);
	return pSet;
}

void CreateTrees(Terrain* pTerrain,Group* pParentGroup)
{
	const float GROUP_DIAMETER = 500.0f;
	const float GROUP_FREQUENCY = 0.1;
	const int MAX_NUM_TREES_PER_GROUP = 30;
	const int MIN_NUM_TREES_PER_GROUP = 10;	
	GeoSet* pTree = NULL;
	
	for (float groupX = GROUP_DIAMETER / 2.0f; groupX < pTerrain->GetWidth() - (GROUP_DIAMETER / 2.0f); groupX += GROUP_DIAMETER)
	{
		for (float groupY = GROUP_DIAMETER / 2.0f; groupY < pTerrain->GetWidth() - (GROUP_DIAMETER / 2.0f); groupY += GROUP_DIAMETER)
		{
			if (((float)rand() / (float)RAND_MAX) <= GROUP_FREQUENCY)
			{
			// There is a group of trees centered at groupX,groupY.
				Group* pGroup = new Group;				
				int numTrees = (int)(((float)rand() / (float)RAND_MAX) * MAX_NUM_TREES_PER_GROUP);
				if (numTrees < MIN_NUM_TREES_PER_GROUP)
					numTrees = MIN_NUM_TREES_PER_GROUP;
				for (int i = 0; i < numTrees; i++)
				{
					float angle = ((float)rand() / (float)RAND_MAX) * 360.0f;
					Matrix matrix;
					matrix.makeRot(angle,0.0f,0.0f,1.0f);					
					
					Vec3 dir;
					dir.x() = 0.0f;
					dir.y() = 1.0f;
					dir.z() = 0.0f;
					dir = dir * matrix;
					
					float distance = ((float)rand() / (float)RAND_MAX) * GROUP_DIAMETER;
					Vec3 position;
					position.x() = groupX + dir.x() * distance;
					position.y() = groupY + dir.y() * distance;
					position.z() = pTerrain->GetElevation(position.x(),position.y());
					if (!pTree)
						pTree = CreateTree(30.0f,40.0f);
					Billboard* pBillboard = new Billboard;
					pBillboard->addDrawable(pTree,position);
					pGroup->addChild(pBillboard);
				}
				pParentGroup->addChild(pGroup);
			}
		}	
	}
}

