// Demeter Terrain Visualization Library by Clay Fowler
// Copyright (C) 2001 Clay Fowler

/* 
IMPORTANT NOTE:
This file is not really a part of Demeter itself. This is simply a sample application that shows how to use Demeter.
As such, this file is written to be easy to follow, not to be a robust application. 
See the file mainOSG.cpp for a slightly more robust (and more interesting) application example.
*/

/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.
*/

#include "Terrain.h"
#include <new.h>
#include <iostream>
#include "Camera.h"
#include "matrix.h"

const float CAMERA_SPEED = 100.0f; // Meters per second
const float CAMERA_ROTATE_SPEED = 0.7f; // Radians per second
const float CAMERA_BORDER_SIZE = 1000.0f; // Meters
const float CAMERA_MIN_ALTITUDE = 5.0f; // Meters
const float CAMERA_MAX_ALTITUDE = 100.0f; // Meters
const float MAX_VIEW_DISTANCE = 7000.0f;
const int MAX_NUM_PARAMETERS = 16;
const int MAX_PARAM_LENGTH = 2048;
const float FOG_RED = 0.5f;
const float FOG_GREEN = 0.75f;
const float FOG_BLUE = 1.0f;
const float FOG_ALPHA = 0.0f;
const float PI = 3.1459f;
const float PI_OVER_TWO = 3.1459f * 0.5f;

using namespace Demeter;

void ToggleWireframe(char* szMsg);
void GetParam(char* szCommand,int paramNum,char* szParamOut);
void ShowUsage();

Settings* pSettings;
bool bWireframe = false;

class LoadListener : public TerrainLoadListener
{
public:
	virtual void TerrainLoaded(Terrain* pTerrain)
	{
		int x,y;
		pTerrain->GetLatticePosition(x,y);
		cout << "Terrain at " << x << "," << y << " loaded!" << endl;
	}

	virtual void TerrainUnloading(Terrain* pTerrain)
	{
		int x,y;
		pTerrain->GetLatticePosition(x,y);
		cout << "Terrain at " << x << "," << y << " unloading!" << endl;
	}
};

#ifdef _WIN32
int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPSTR lpCmdLine,int nCmdShow)
#else
int main(int argc,char** argv)
#endif
{
	bool bFullscreen = true;
	bool bUseLowRes = false;
    int ScreenWidth;
    int ScreenHeight;
	Vector lookAt;
	Vector lookUp;
	Vector cameraAngle;
	float threshold = 12.0f;
	bool continueRunning = true;
	bool firstTime = true;
	int cycles = 0;

// Parse the command line
#ifdef _WIN32
	int argc = 1;
	char** argv;
	argv = new char*[16];
	for (int win32cmd = 0; win32cmd < 16; win32cmd++)
		argv[win32cmd] = new char[512];
	sprintf(argv[0],"pagingdemo.exe");
	char szParam[512];
	do
	{
		GetParam(lpCmdLine,argc,szParam);
		if (0 < strlen(szParam))
		{
			sprintf(argv[argc],szParam);
			argc++;
		}
	}
	while (0 < strlen(szParam));
#endif

	try
	{
		pSettings = Settings::GetInstance();

		if (argc != 2)
		{
			ShowUsage();
			exit(1);
		}

		pSettings->SetMediaPath(argv[1]);

		ScreenWidth = 800;
		ScreenHeight = 600;

		pSettings->SetScreenWidth(ScreenWidth);
		pSettings->SetScreenHeight(ScreenHeight);
		
	// Initialize the rendering surface
	//////////////////////////////
		if (pSettings->IsVerbose())
			cout << "VIEWER: Initializing video" << endl;
		if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0)
		{
			printf("ERROR: Unable to initialize SDL: %s\n",SDL_GetError());
			exit(2);
		}
		atexit(SDL_Quit);
		SDL_GL_SetAttribute(SDL_GL_BUFFER_SIZE,32);
		SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,1);
		SDL_EnableUNICODE(1);
		if (pSettings->IsVerbose())
			printf("VIEWER: Setting video mode (OpenGL)\n");
		Uint32 screenFlags = SDL_OPENGL;
		if (bFullscreen)
			screenFlags |= SDL_FULLSCREEN;
		SDL_Surface* pScreen = SDL_SetVideoMode(pSettings->GetScreenWidth(),pSettings->GetScreenHeight(),32,screenFlags);
		if (pScreen == NULL)
		{
			cout << "ERROR: Unable to set screen resolution" << endl;
			exit(3);
		}
		SDL_WM_SetCaption("Demeter","testgl");
		if (pSettings->IsVerbose())
			cout << "VIEWER: OpenGL vendor: " << glGetString(GL_VENDOR) << endl;
		if (pSettings->IsVerbose())
			cout << "VIEWER: OpenGL extensions supported: " << glGetString(GL_EXTENSIONS) << endl;	

	// Create the terrain
		const int MAX_NUM_VISIBLE_TRIANGLES = 40000; // Chosen based on the expected number of triangles that will be visible on-screen at any one time (the terrain mesh will typically have far more triangles than are seen at one time, especially with dynamic tessellation)
		TerrainLattice* pLattice = new TerrainLattice("Utah","bmp","bmp","dirt.jpg",10.0f,8.0f,MAX_NUM_VISIBLE_TRIANGLES,false,9,9);
		
		try
		{
			pLattice->AddTerrainLoadListener(*new LoadListener);
		}
		catch(DemeterException* pEx)
		{
			cout << "VIEWER: WARNING - in order to run the PagingDemo, you MUST first download the paging demo data!" << endl;
			cout << pEx->GetErrorMessage() << endl;
			exit(0);
		}
		catch(...)
		{
			cout << "VIEWER: Unexpected exception while creating terrain" << endl;
			exit(0);
		}
		
	// Vanilla OpenGL setup
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glViewport(0,0,pSettings->GetScreenWidth(),pSettings->GetScreenHeight());
		glFrustum(-0.5f,0.5f,-0.5f,0.5f,0.45f,MAX_VIEW_DISTANCE);
		glClearColor(0.5f,0.75f,1.0f,0.0f);
		glDisable(GL_NORMALIZE);
		glEnable(GL_DEPTH_TEST);
		glShadeModel(GL_SMOOTH);
	// Setup camera
		Vector cameraPosition;
		cameraPosition.x = 50.0f;
		cameraPosition.y = 50.0f;
		cameraPosition.z = 110.0f;
		Camera camera;
		camera.SetPosition(cameraPosition.x,cameraPosition.y,cameraPosition.z);			
		cameraAngle.x = 0.0f;
		cameraAngle.y = 0.0f;
		cameraAngle.z = 0.0f;
		pLattice->SetCameraPosition(cameraPosition.x,cameraPosition.y,0.0f);
	// Setup fog - it is highly recommended that apps ALWAYS use fog for realistic terrains...
		float fogColor[4];
		fogColor[0] = FOG_RED;
		fogColor[1] = FOG_GREEN;
		fogColor[2] = FOG_BLUE;
		fogColor[3] = FOG_ALPHA;
		glEnable(GL_FOG);
		glFogf(GL_FOG_MODE,GL_LINEAR);
		glFogfv(GL_FOG_COLOR,fogColor);
		glFogf(GL_FOG_START,100.0f);
		glFogf(GL_FOG_END,MAX_VIEW_DISTANCE - 100.0f);
		glHint(GL_FOG_HINT,GL_FASTEST);
	
	// Main control and rendering loop
		Uint32 startTime = SDL_GetTicks();
		Uint32 wireframeToggledTime = startTime;
		float cyclesPerSecond = 30.0f;
		
		while (continueRunning)
		{
		// Setup viewing transformation
			Matrix rotateX,rotateZ,cameraTransform; // TBD: Quick and dirty viewing system - for a real app something more robust would probably be required
			rotateX.SetRotateX(cameraAngle.x);
			rotateZ.SetRotateZ(cameraAngle.z);
			cameraTransform = rotateX * rotateZ;
			lookAt.x = 0.0f;
			lookAt.y = 100000.0f; // Look at a suitably (and arbitrarily) distant point - if your terrain's bigger than this then increase this value
			lookAt.z = 0.0f;
			lookUp.x = 0.0f;
			lookUp.y = 0.5f;
			lookUp.z = 1.0f;
			lookUp.Normalize();
			Vector currentLookAt = cameraTransform * lookAt;
			Vector currentLookUp = cameraTransform * lookUp;
			camera.SetLookAt(cameraPosition.x + currentLookAt.x,cameraPosition.y + currentLookAt.y,cameraPosition.z + currentLookAt.z);
			camera.SetLookUp(currentLookUp.x,currentLookUp.y,currentLookUp.z);
			float currentCameraElevation = pLattice->GetElevation(cameraPosition.x,cameraPosition.y) + cameraPosition.z;
			camera.SetPosition(cameraPosition.x,cameraPosition.y,currentCameraElevation);
			camera.UpdateViewTransform();
		// Render the terrain
			pLattice->SetCameraPosition(cameraPosition.x,cameraPosition.y,currentCameraElevation);
			pLattice->SetDetailThreshold(threshold);
			pLattice->Tessellate();
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			pLattice->Render();
		// Swap the double buffer so we can see what we just rendered
			SDL_GL_SwapBuffers();
		// Handle mouse camera commands
			int mouseX,mouseY;
			SDL_GetMouseState(&mouseX,&mouseY);
			float speed = (float)CAMERA_SPEED / cyclesPerSecond;
			float rotateSpeed = (float)CAMERA_ROTATE_SPEED / cyclesPerSecond;
			if (mouseX == 0)
				cameraAngle.z += rotateSpeed;
			if (mouseX == pSettings->GetScreenWidth() - 1)
				cameraAngle.z -= rotateSpeed;
			if (mouseY == 0)
			{
				Vector lookAt2D = currentLookAt;
				lookAt2D.z = 0.0f;
				lookAt2D.Normalize();
				cameraPosition.x += lookAt2D.x * speed;
				cameraPosition.y += lookAt2D.y * speed;
				cameraPosition.z += lookAt2D.z * speed;
			}
			if (mouseY == pSettings->GetScreenHeight() - 1)
			{
				Vector lookAt2D = currentLookAt;
				lookAt2D.z = 0.0f;
				lookAt2D.Normalize();
				cameraPosition.x += lookAt2D.x * -speed;
				cameraPosition.y += lookAt2D.y * -speed;
				cameraPosition.z += lookAt2D.z * -speed;
			}
		// Handle all other input
			SDL_PumpEvents();
			SDL_Event event;
			while(SDL_PollEvent(&event))
			{
				if (event.type == SDL_QUIT)
					continueRunning = false;	
			}

			int numKeys;
			Uint8* pKeys = SDL_GetKeyState(&numKeys);
			// Lots of redundant code that would need to be modularized for a real app in here...
			if (pKeys[SDLK_t])
			{
				if (1000 < SDL_GetTicks() - wireframeToggledTime)
				{
					ToggleWireframe(NULL);
					wireframeToggledTime = SDL_GetTicks();
				}
			}
			
			if (pKeys[SDLK_e])
				speed *= 10.0f;
			if (pKeys[SDLK_w])
			{
			// Move the camera forward
				Vector lookAt2D = currentLookAt;
				lookAt2D.z = 0.0f;
				lookAt2D.Normalize();
				cameraPosition.x += lookAt2D.x * speed;
				cameraPosition.y += lookAt2D.y * speed;
			}
			if (pKeys[SDLK_x])
			{
			// Move the camera backward
				Vector lookAt2D = currentLookAt;
				lookAt2D.z = 0.0f;
				lookAt2D.Normalize();
				cameraPosition.x += -lookAt2D.x * speed;
				cameraPosition.y += -lookAt2D.y * speed;
			}
			if (pKeys[SDLK_a])
			{
			// Move the camera left
				Vector lookAt2D = currentLookAt;
				lookAt2D.z = 0.0f;
				lookAt2D.Normalize();
				Matrix mat;
				mat.SetRotateZ(PI_OVER_TWO);
				lookAt2D = mat * lookAt2D;
				cameraPosition.x += lookAt2D.x * speed;
				cameraPosition.y += lookAt2D.y * speed;
			}
			if (pKeys[SDLK_d])
			{
			// Move the camera right
				Vector lookAt2D = currentLookAt;
				lookAt2D.z = 0.0f;
				lookAt2D.Normalize();
				Matrix mat;
				mat.SetRotateZ(-PI_OVER_TWO);
				lookAt2D = mat * lookAt2D;
				cameraPosition.x += lookAt2D.x * speed;
				cameraPosition.y += lookAt2D.y * speed;
			}
			if (pKeys[SDLK_q])
			// Move the camera up
				cameraPosition.z += speed;
			if (pKeys[SDLK_z])
			// Move the camera down
				cameraPosition.z -= speed;
			if (pKeys[SDLK_ESCAPE])
			// Quit the application
				continueRunning = false;

			if (pKeys[SDLK_KP_PLUS])
			{
			// Increase the detail threshold (make the terrain less detailed - worse looking but faster)
				threshold += 0.5f;
				cout << "VIEWER: Threshold = " << threshold << endl;
			}
			if (pKeys[SDLK_KP_MINUS])
			{
			// Decrease the detail threshold (make the terrain more detailed - better looking but slower)
				threshold -= 0.5f;
				cout << "VIEWER: Threshold = " << threshold << endl;
			}

		// Clamp all control values to legal limits
			if (threshold < 0.0f)
				threshold = 0.0f;
			if (cameraPosition.x < CAMERA_BORDER_SIZE)
				cameraPosition.x = CAMERA_BORDER_SIZE;
			if (pLattice->GetWidth() - CAMERA_BORDER_SIZE < cameraPosition.x)
				cameraPosition.x = pLattice->GetWidth() - CAMERA_BORDER_SIZE;
			if (cameraPosition.y < CAMERA_BORDER_SIZE)
				cameraPosition.y = CAMERA_BORDER_SIZE;
			if (pLattice->GetHeight() - CAMERA_BORDER_SIZE < cameraPosition.y)
				cameraPosition.y = pLattice->GetHeight() - CAMERA_BORDER_SIZE;
			if (cameraPosition.z < CAMERA_MIN_ALTITUDE)
				cameraPosition.z = CAMERA_MIN_ALTITUDE;
			if (CAMERA_MAX_ALTITUDE < cameraPosition.z)
				cameraPosition.z = CAMERA_MAX_ALTITUDE;

			Uint32 elapsedTime = SDL_GetTicks() - startTime;
			if (1000 < elapsedTime)
			{
				if (pSettings->IsVerbose())
				{
					cyclesPerSecond = (float)cycles / ((float)elapsedTime / 1000.0f);
					cout << "VIEWER: " << cyclesPerSecond << " frames per second" << endl;
				}
				cycles = 0;
				startTime = SDL_GetTicks();
			}
			cycles++;
		}	
		SDL_Quit();
	}
	catch(DemeterException* pEx)
	{
		cout << pEx->GetErrorMessage() << endl;
		SDL_Quit();
	}
	catch(...)
	{
		SDL_Quit();
	}
	return 1;
}

void ShowUsage()
{
	cout << "PagingDemo <media path>" << endl;
}

void ToggleWireframe(char* szMsg)
{
	bWireframe = !bWireframe;
	if (bWireframe)
		glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
	else
		glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
}

void GetParam( char* szCommand,int paramNum,char* szParamOut )
{
	char* szParam = szCommand;

	int numParamsSkipped = 0;

	for (int i = 0; i < paramNum; i++)
	{
		char* szWrite = szParamOut;
		while(*szParam == ' ') szParam++;
		int length = 0;
		while (*szParam != '\0' && *szParam != ' ' && ++length < MAX_PARAM_LENGTH)
			*szWrite++ = *szParam++;
		*szWrite = '\0';
	}
}
