/***************************************************************************
                          morph.cpp  -  morphing methods
                             -------------------
    begin                : Sun Oct 29 2000
    copyright            : (C) 2000 by Tero Kontkanen
    email                : tkontkanen@mail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "tgl.h"

namespace TGL{

void tgl_object::set_loop( int begin, int end )
{
	// bad value
	if( begin<0 )
	{
		tgl_msg( "set_loop: 'begin' value out of range (%d)", TGL_WARNING, (void*)begin );
		begin = 0;
	}
	
	if( begin>=numframes )
	{
		tgl_msg( "set_loop: 'begin' value out of range (%d)", TGL_WARNING, (void*)begin );
		begin = numframes-1;
	}

	if( end<0 )
	{
		tgl_msg( "set_loop: 'end' value out of range (%d)", TGL_WARNING, (void*)end );
		end = 0;
	}
	
	if( end>=numframes )
	{
		tgl_msg( "set_loop: 'end' value out of range (%d)", TGL_WARNING, (void*)end );
		end = numframes-1;
	}

	loop_begin = begin;
	loop_end = end;
	
	set_morph( morphfrom, morphto ); // check if these are out of loop range
}

void tgl_object::set_morph( int from, int to )
{
	// bad value
	if( from<loop_begin )
	{
		tgl_msg( "set_morph: 'from' value out of loop range (%d)", TGL_WARNING, (void*)from );
		from = 0;
	}
	if( from>loop_end )
	{
		tgl_msg( "set_morph: 'from' value out of loop range (%d)", TGL_WARNING, (void*)from );
		from = numframes-1;
	}

	if( to<loop_begin )
	{
		tgl_msg( "set_morph: 'to' value out of loop range (%d)", TGL_WARNING, (void*)to );
		to = 0;
	}
	if( to>loop_end )
	{
		tgl_msg( "set_morph: 'to' value out of loop range (%d)", TGL_WARNING, (void*)to );
		to = numframes-1;
	}

	// if changed...
	if( morphfrom != from ) do_morph = 1;
	if( morphto != to ) do_morph = 1;

	morphfrom = from;
	morphto = to;

	// ...update vertexmorph lookup table
	if( do_morph ) calc_morphtable();
}

void tgl_object::calc_morphtable( void )
{
	int i, morphfrom_i=morphfrom*numvertices, morphto_i=morphto*numvertices;
	int ifrom, ito;

	for( i=0; i<numvertices; i++ )
	{
		ifrom = morphfrom_i+i;
		ito = morphto_i+i;

		vertexmorph[i].pos[0] = vertices[ito].pos[0] - vertices[ifrom].pos[0];
		vertexmorph[i].pos[1] = vertices[ito].pos[1] - vertices[ifrom].pos[1];
		vertexmorph[i].pos[2] = vertices[ito].pos[2] - vertices[ifrom].pos[2];

		if( coloring )
		{
			vertexmorph[i].col[0] = vertices[ito].col[0] - vertices[ifrom].col[0];
			vertexmorph[i].col[1] = vertices[ito].col[1] - vertices[ifrom].col[1];
			vertexmorph[i].col[2] = vertices[ito].col[2] - vertices[ifrom].col[2];
			vertexmorph[i].col[3] = vertices[ito].col[3] - vertices[ifrom].col[3];
		}
		
		if( lighting || hfre )
		{
			vertexmorph[i].vec[0] = vertices[ito].vec[0] - vertices[ifrom].vec[0];
			vertexmorph[i].vec[1] = vertices[ito].vec[1] - vertices[ifrom].vec[1];
			vertexmorph[i].vec[2] = vertices[ito].vec[2] - vertices[ifrom].vec[2];
		}
		
		if( texturing )
		{
			vertexmorph[i].tex[0] = vertices[ito].tex[0] - vertices[ifrom].tex[0];
			vertexmorph[i].tex[1] = vertices[ito].tex[1] - vertices[ifrom].tex[1];
		}
	}

	if( tgl_debug >= 3 ) tgl_msg( "calc_morphtable: morphtable updated", TGL_INFO );
}

void tgl_object::calc_currframe( void )
{
	int i, morphfrom_i=morphfrom*numvertices, ifrom;

  if( pos_morph ) calc_currframe_pos();
  if( vec_morph && lighting || hfre ) calc_currframe_vec();
  if( col_morph && coloring ) calc_currframe_col();
  if( tex_morph && texturing ) calc_currframe_tex();
		
	prevframenum = currframenum;
}

void tgl_object::calc_currframe_pos( void )
{
	int i, morphfrom_i=morphfrom*numvertices, ifrom;
	
	switch( pos_morph ) // mode #
	{
		case 0: // none
			break;
			
		case 1: // nearest
			if( prevframenum != currframenum )
				for( i=0; i<numvertices; i++ ) {
					ifrom = i + numvertices*currframenum;
					currframe[i].pos[0] = vertices[ifrom].pos[0];
					currframe[i].pos[1] = vertices[ifrom].pos[1];
					currframe[i].pos[2] = vertices[ifrom].pos[2];
				}				
			break;
			
		case 2: // linear
			for( i=0; i<numvertices; i++ ) {
				ifrom = i + morphfrom_i;
				currframe[i].pos[0] = vertices[ifrom].pos[0] + currmorph*vertexmorph[i].pos[0];
				currframe[i].pos[1] = vertices[ifrom].pos[1] + currmorph*vertexmorph[i].pos[1];
				currframe[i].pos[2] = vertices[ifrom].pos[2] + currmorph*vertexmorph[i].pos[2];
			}				
			break;
			
		case 3: // cubic
			for( i=0; i<numvertices; i++ ) {
				currframe[i].pos[0] = cubic_morph( 9, morphfrom, morphto, i, currmorph );
				currframe[i].pos[1] = cubic_morph( 10, morphfrom, morphto, i, currmorph );
				currframe[i].pos[2] = cubic_morph( 11, morphfrom, morphto, i, currmorph );		
			}
			break;
			
		default: // not supported?
			tgl_msg( "calc_currframe_pos: Invalid morphing mode (%d)", TGL_WARNING, (void*)pos_morph );
			break;
	}
}

void tgl_object::calc_currframe_vec( void )
{
	int i, morphfrom_i=morphfrom*numvertices, ifrom;
	
	switch( vec_morph ) // mode #
	{
		case 0: // none
			break;
			
		case 1: // nearest
			if( prevframenum != currframenum )
				for( i=0; i<numvertices; i++ ) {
					ifrom = i + numvertices*currframenum;
					currframe[i].vec[0] = vertices[ifrom].vec[0];
					currframe[i].vec[1] = vertices[ifrom].vec[1];
					currframe[i].vec[2] = vertices[ifrom].vec[2];
				}				
			break;
			
		case 2: // linear
			for( i=0; i<numvertices; i++ ) {
				ifrom = i + morphfrom_i;
				currframe[i].vec[0] = vertices[ifrom].vec[0] + currmorph*vertexmorph[i].vec[0];
				currframe[i].vec[1] = vertices[ifrom].vec[1] + currmorph*vertexmorph[i].vec[1];
				currframe[i].vec[2] = vertices[ifrom].vec[2] + currmorph*vertexmorph[i].vec[2];
			}				
			break;
			
		case 3: // cubic
			for( i=0; i<numvertices; i++ ) {
				currframe[i].vec[0] = cubic_morph( 6, morphfrom, morphto, i, currmorph );
				currframe[i].vec[1] = cubic_morph( 7, morphfrom, morphto, i, currmorph );
				currframe[i].vec[2] = cubic_morph( 8, morphfrom, morphto, i, currmorph );		
			}
			break;
			
		default: // not supported?
			tgl_msg( "calc_currframe_vec: Invalid morphing mode (%d)", TGL_WARNING, (void *)vec_morph );
			break;
	}
}

void tgl_object::calc_currframe_col( void )
{
	int i, morphfrom_i=morphfrom*numvertices, ifrom;
	
	switch( col_morph ) // mode #
	{
		case 0: // none
			break;
			
		case 1: // nearest
			if( prevframenum != currframenum )
				for( i=0; i<numvertices; i++ ) {
					ifrom = i + numvertices*currframenum;
					currframe[i].col[0] = vertices[ifrom].col[0];
					currframe[i].col[1] = vertices[ifrom].col[1];
					currframe[i].col[2] = vertices[ifrom].col[2];
					currframe[i].col[3] = vertices[ifrom].col[3];
				}				
			break;
			
		case 2: // linear
			for( i=0; i<numvertices; i++ ) {
				ifrom = i + morphfrom_i;
				currframe[i].col[0] = vertices[ifrom].col[0] + currmorph*vertexmorph[i].col[0];
				currframe[i].col[1] = vertices[ifrom].col[1] + currmorph*vertexmorph[i].col[1];
				currframe[i].col[2] = vertices[ifrom].col[2] + currmorph*vertexmorph[i].col[2];
				currframe[i].col[3] = vertices[ifrom].col[3] + currmorph*vertexmorph[i].col[3];
			}				
			break;
			
		case 3: // cubic
			for( i=0; i<numvertices; i++ ) {
				currframe[i].col[0] = cubic_morph( 2, morphfrom, morphto, i, currmorph );
				currframe[i].col[1] = cubic_morph( 3, morphfrom, morphto, i, currmorph );
				currframe[i].col[2] = cubic_morph( 4, morphfrom, morphto, i, currmorph );
				currframe[i].col[3] = cubic_morph( 5, morphfrom, morphto, i, currmorph );
			}
			break;
		
		default: // not supported?
			tgl_msg( "calc_currframe_col: Invalid morphing mode (%d)", TGL_WARNING, (void*)col_morph );
			break;
	}
}

void tgl_object::calc_currframe_tex( void )
{
	int i, morphfrom_i=morphfrom*numvertices, ifrom;
	
	switch( tex_morph ) // mode #
	{
		case 0: // none
			break;
			
		case 1: // nearest
			if( prevframenum != currframenum )
				for( i=0; i<numvertices; i++ ) {
					ifrom = i + numvertices*currframenum;
					currframe[i].tex[0] = vertices[ifrom].tex[0];
					currframe[i].tex[1] = vertices[ifrom].tex[1];
				}				
			break;
			
		case 2: // linear
			for( i=0; i<numvertices; i++ ) {
				ifrom = i + morphfrom_i;
				currframe[i].tex[0] = vertices[ifrom].tex[0] + currmorph*vertexmorph[i].tex[0];
				currframe[i].tex[1] = vertices[ifrom].tex[1] + currmorph*vertexmorph[i].tex[1];
			}				
			break;
			
		case 3: // cubic
			for( i=0; i<numvertices; i++ ) {
				currframe[i].tex[0] = cubic_morph( 0, morphfrom, morphto, i, currmorph );
				currframe[i].tex[1] = cubic_morph( 1, morphfrom, morphto, i, currmorph );
			}
			break;
			
		default: // not supported?
			tgl_msg( "calc_currframe_tex: Invalid morphing mode (%d)", TGL_WARNING, (void*)tex_morph );
			break;			
	}
}


void tgl_object::morph( float pos )
{
	if( pos < 0.0 )
	{
		tgl_msg( "morph: position value out of range (model=%s)", TGL_WARNING, modelname );
		pos = 0.0;
	}
	if( pos > 1.0 )
	{
		tgl_msg( "morph: position value out of range (model=%s)", TGL_WARNING, modelname );
		pos = 1.0;
	}

	if( pos != currmorph ) do_morph = 1;
	currmorph = pos;

	// texturenum etc
	if( pos >= 0.5 ) currframenum = morphto;
	else currframenum = morphfrom;
}

}