/***************************************************************************
                          t3dmodel.cpp  -  model read/write code
                             -------------------
    begin                : Tue Jul 18 2000
    copyright            : (C) 2000 by Tero Kontkanen
    email                : tkontkanen@mail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "tgl_object.h"

namespace TGL{

/* print model information for debugging */
GLvoid tgl_object::debug_model( void )
{
	int i;

	if( tgl_debug >= 1 )
	fprintf( tglerr, "Model summary (%s):\n"
		"  Header: name='%s', frames=%d, vertices=%d, indices=%d\n"
		"  Params: hfre=%d, lighting=%d, coloring=%d, src_blendf=%d, dst_blendf=%d\n"
		"  Morphing modes: pos=%d, vec=%d, col=%d, tex=%d, cubic_strength=%.3f\n"
		"  Material: shiness=%.2f, specular=(%.2f,%.2f,%.2f,%.2f)\n"
		"            diffuse=(%.2f,%.2f,%.2f,%.2f)\n"
		"            ambient=(%.2f,%.2f,%.2f,%.2f)\n"
		"            emission=(%.2f,%.2f,%.2f,%.2f)\n",
		model_filename.c_str(),
		modelname, numframes, numvertices, numindices,
		hfre, lighting, coloring, src_blendfunc, dest_blendfunc,
		pos_morph, vec_morph, col_morph, tex_morph, cubic_strength,
		material.shiness,
		material.specular[0], material.specular[1], material.specular[2], material.specular[3],
		material.diffuse[0], material.diffuse[1], material.diffuse[2], material.diffuse[3],
		material.ambient[0], material.ambient[1], material.ambient[2], material.ambient[3],
		material.emission[0], material.emission[1], material.emission[2], material.emission[3] );

	if( tgl_debug >= 2 )
	{
		fprintf( tglerr, "  Textures:\n" );
		for( i=0; i<numframes; i++ )
			fprintf( tglerr, "    frame# %d '%s'\n", i, tgl_texturelist[textures[i]].name );
	}

	if( tgl_debug >= 4 )
	{
		fprintf( tglerr, "  Vertices:\n" );
		for( i=0; i< numvertices*numframes; i++ )
		{
			fprintf( tglerr, "    %d: %.3f %.3f %.3f  %.3f %.3f %.3f %.3f  %.3f %.3f\n", i,
				vertices[i].pos[0], vertices[i].pos[1], vertices[i].pos[2],
				vertices[i].col[0], vertices[i].col[1], vertices[i].col[2], vertices[i].col[3],
				vertices[i].tex[0], vertices[i].tex[1] );
		}

		fprintf( tglerr, "  Triangles:\n" );
		for( i=0; i< numindices/3; i++ )
		{
			fprintf( tglerr, "    %d: %d, %d, %d\n", i,
				indices[i*3], indices[i*3+1], indices[i*3+2] );
		}
	}			

}

/* Detects file format and tries to load it */
GLint tgl_object::load_model( const char *fname )
{
	FILE *f;
	char idstr[16];

	f = fopen( fname, "rb" );
	if( f == NULL )
	{
		tgl_file_not_found( "load_model", fname );
		return -1;
	}
	fread( idstr, 1, 16, f );

	// detect file format
	if( !strncmp( "T3Dmodel", idstr, 8 ) ) {
		load_t3dmodel( f, fname );
		filetype = TGL_T3D;
	}
	else if( !strncmp( "T3DModelBin", idstr, 11 ) ) {
		load_t3dmodelbin( f, fname );
		filetype = TGL_T3DBIN;
	}
	else
		tgl_msg( "load_model: usupported file format (%s)", TGL_ERROR, (void*)fname );

	if( f != NULL ) fclose( f );

	currframe = (TVERTEX*)tgl_alloc( "load_model", numvertices*sizeof( TVERTEX ) );
	vertexmorph = (TVERTEX*)tgl_alloc( "load_model", numvertices*sizeof( TVERTEX ) );

	calcnormals();
	
	calc_morphtable();
	memcpy( currframe, vertices, sizeof(TVERTEX)*numvertices ); // init first frame
//	calc_currframe();
	
	if( tgl_debug >= 1 ) debug_model();

	return 0;
}

/* T3Dmodel (text) format loader */
GLint tgl_object::load_t3dmodel( FILE *f, const char *fname )
{
	char idstr[10];
	int ver;

	// seek begin of file
	fseek( f, 0, SEEK_SET );

	//detect file version
	if( fscanf( f, "%s %d", &idstr, &ver ) !=2 )
		tgl_msg( "load_t3dmodel: corrupted file (%s)", TGL_ERROR, (void*)fname );

		
	// file loaders for different versions
	switch( ver )
	{
		case 100:
			return load_t3dmodel_100( f, fname );
			break;
		case 101:
			return load_t3dmodel_101( f, fname );
			break;
		default: // version not supported
			tgl_msg( "load_t3dmodel: Unsupported file version (%d)", TGL_ERROR, (void*)ver );
			return -1;
			break;
	}
}

/* T3DModelBin (binary) format loader */
GLint tgl_object::load_t3dmodelbin( FILE *f, const char *fname )
{
	int ver, reqver;

	// seek version number
	fseek( f, 11, SEEK_SET );

	//detect file version
	if( fread( &ver, 4, 1, f ) !=1 ) // current file version number
		tgl_msg( "load_t3dmodelbin: corrupted file (%s)", TGL_ERROR, (void*)fname );
	if( fread( &reqver, 4, 1, f ) !=1 ) // oldest supported loader version number
		tgl_msg( "load_t3dmodelbin: corrupted file (%s)", TGL_ERROR, (void*)fname );

	// file loaders for different versions
	switch( ver )
	{
		case 100:
			return load_t3dmodelbin_100( f, fname );
			break;
		case 101:
			return load_t3dmodelbin_101( f, fname );
			break;
		default: // version not supported
			if( reqver <= 101 )
			{
				tgl_msg( "load_t3dmodelbin: Trying to load file version %d as 101", TGL_INFO, (void*)ver );
				return load_t3dmodelbin_101( f, fname );
			}
			else
			{
				tgl_msg( "load_t3dmodelbin: Unsupported file version (%d)", TGL_ERROR, (void*)ver );
				return -1;
			}
			break;
	}
}


// support different file versions
GLint tgl_object::save_t3dmodel( const char *fname, int ver )
{
	if( ver == 100 ) return( save_t3dmodel_100( fname ) );
	if( ver == 101 ) return( save_t3dmodel_101( fname ) );
	tgl_msg( "save_t3dmodel: unsupported version (%d)", TGL_WARNING, (void*)ver );
}

// support different file versions
GLint tgl_object::save_t3dmodelbin( const char *fname, int ver )
{
	if( ver == 100 ) return( save_t3dmodelbin_100( fname ) );
	if( ver == 101 ) return( save_t3dmodelbin_101( fname ) );
	tgl_msg( "save_t3dmodelbin: unsupported version (%d)", TGL_WARNING, (void*)ver );
}

}